# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: caffe.proto
# PPL caffe proto, version Date: 2020/12/11

from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='caffe.proto',
  package='caffe',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=b'\n\x0b\x63\x61\x66\x66\x65.proto\x12\x05\x63\x61\x66\x66\x65\"\x1c\n\tBlobShape\x12\x0f\n\x03\x64im\x18\x01 \x03(\x03\x42\x02\x10\x01\"\xac\x01\n\tBlobProto\x12\x1f\n\x05shape\x18\x07 \x01(\x0b\x32\x10.caffe.BlobShape\x12\x10\n\x04\x64\x61ta\x18\x05 \x03(\x02\x42\x02\x10\x01\x12\x10\n\x04\x64iff\x18\x06 \x03(\x02\x42\x02\x10\x01\x12\x0e\n\x03num\x18\x01 \x01(\x05:\x01\x30\x12\x13\n\x08\x63hannels\x18\x02 \x01(\x05:\x01\x30\x12\x11\n\x06height\x18\x03 \x01(\x05:\x01\x30\x12\x10\n\x05width\x18\x04 \x01(\x05:\x01\x30\x12\x10\n\x05\x64\x65pth\x18\x08 \x01(\x05:\x01\x30\"2\n\x0f\x42lobProtoVector\x12\x1f\n\x05\x62lobs\x18\x01 \x03(\x0b\x32\x10.caffe.BlobProto\"\xa0\x01\n\x11QuantizeParameter\x12\x16\n\x04type\x18\x01 \x01(\t:\x08quantize\x12\x11\n\x04step\x18\x02 \x01(\x02:\x03\x30.1\x12\x14\n\trange_min\x18\x03 \x01(\x02:\x01\x30\x12\x16\n\trange_max\x18\x04 \x01(\x02:\x03\x32\x35\x35\x12\x15\n\nzero_point\x18\x05 \x01(\r:\x01\x30\x12\x1b\n\x0c\x61\x64just_range\x18\x06 \x01(\x08:\x05\x66\x61lse\"\x90\x01\n\x05\x44\x61tum\x12\x10\n\x08\x63hannels\x18\x01 \x01(\x05\x12\x0e\n\x06height\x18\x02 \x01(\x05\x12\r\n\x05width\x18\x03 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\r\n\x05label\x18\x05 \x01(\x05\x12\x12\n\nfloat_data\x18\x06 \x03(\x02\x12\x16\n\x07\x65ncoded\x18\x07 \x01(\x08:\x05\x66\x61lse\x12\r\n\x05\x64\x65pth\x18\x08 \x01(\x05\"\x90\x01\n\x0f\x46illerParameter\x12\x16\n\x04type\x18\x01 \x01(\t:\x08\x63onstant\x12\x10\n\x05value\x18\x02 \x01(\x02:\x01\x30\x12\x0e\n\x03min\x18\x03 \x01(\x02:\x01\x30\x12\x0e\n\x03max\x18\x04 \x01(\x02:\x01\x31\x12\x0f\n\x04mean\x18\x05 \x01(\x02:\x01\x30\x12\x0e\n\x03std\x18\x06 \x01(\x02:\x01\x31\x12\x12\n\x06sparse\x18\x07 \x01(\x05:\x02-1\"\x8e\x02\n\x0cNetParameter\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05input\x18\x03 \x03(\t\x12%\n\x0binput_shape\x18\x08 \x03(\x0b\x32\x10.caffe.BlobShape\x12\x11\n\tinput_dim\x18\x04 \x03(\x05\x12\x1d\n\x0e\x66orce_backward\x18\x05 \x01(\x08:\x05\x66\x61lse\x12\x1e\n\x05state\x18\x06 \x01(\x0b\x32\x0f.caffe.NetState\x12\x19\n\ndebug_info\x18\x07 \x01(\x08:\x05\x66\x61lse\x12$\n\x05layer\x18\x64 \x03(\x0b\x32\x15.caffe.LayerParameter\x12\'\n\x06layers\x18\x02 \x03(\x0b\x32\x17.caffe.V1LayerParameter\"\xa7\x08\n\x0fSolverParameter\x12\x0b\n\x03net\x18\x18 \x01(\t\x12&\n\tnet_param\x18\x19 \x01(\x0b\x32\x13.caffe.NetParameter\x12\x11\n\ttrain_net\x18\x01 \x01(\t\x12\x10\n\x08test_net\x18\x02 \x03(\t\x12,\n\x0ftrain_net_param\x18\x15 \x01(\x0b\x32\x13.caffe.NetParameter\x12+\n\x0etest_net_param\x18\x16 \x03(\x0b\x32\x13.caffe.NetParameter\x12$\n\x0btrain_state\x18\x1a \x01(\x0b\x32\x0f.caffe.NetState\x12#\n\ntest_state\x18\x1b \x03(\x0b\x32\x0f.caffe.NetState\x12\x11\n\ttest_iter\x18\x03 \x03(\x05\x12\x18\n\rtest_interval\x18\x04 \x01(\x05:\x01\x30\x12 \n\x11test_compute_loss\x18\x13 \x01(\x08:\x05\x66\x61lse\x12!\n\x13test_initialization\x18  \x01(\x08:\x04true\x12\x0f\n\x07\x62\x61se_lr\x18\x05 \x01(\x02\x12\x0f\n\x07\x64isplay\x18\x06 \x01(\x05\x12\x17\n\x0c\x61verage_loss\x18! \x01(\x05:\x01\x31\x12\x10\n\x08max_iter\x18\x07 \x01(\x05\x12\x11\n\tlr_policy\x18\x08 \x01(\t\x12\r\n\x05gamma\x18\t \x01(\x02\x12\r\n\x05power\x18\n \x01(\x02\x12\x10\n\x08momentum\x18\x0b \x01(\x02\x12\x14\n\x0cweight_decay\x18\x0c \x01(\x02\x12\x1f\n\x13regularization_type\x18\x1d \x01(\t:\x02L2\x12\x10\n\x08stepsize\x18\r \x01(\x05\x12\x11\n\tstepvalue\x18\" \x03(\x05\x12\x1a\n\x0e\x63lip_gradients\x18# \x01(\x02:\x02-1\x12\x13\n\x08snapshot\x18\x0e \x01(\x05:\x01\x30\x12\x17\n\x0fsnapshot_prefix\x18\x0f \x01(\t\x12\x1c\n\rsnapshot_diff\x18\x10 \x01(\x08:\x05\x66\x61lse\x12;\n\x0bsolver_mode\x18\x11 \x01(\x0e\x32!.caffe.SolverParameter.SolverMode:\x03GPU\x12\x14\n\tdevice_id\x18\x12 \x01(\x05:\x01\x30\x12\x17\n\x0brandom_seed\x18\x14 \x01(\x03:\x02-1\x12;\n\x0bsolver_type\x18\x1e \x01(\x0e\x32!.caffe.SolverParameter.SolverType:\x03SGD\x12\x14\n\x05\x64\x65lta\x18\x1f \x01(\x02:\x05\x31\x65-08\x12\x19\n\ndebug_info\x18\x17 \x01(\x08:\x05\x66\x61lse\x12\"\n\x14snapshot_after_train\x18\x1c \x01(\x08:\x04true\"\x1e\n\nSolverMode\x12\x07\n\x03\x43PU\x10\x00\x12\x07\n\x03GPU\x10\x01\"0\n\nSolverType\x12\x07\n\x03SGD\x10\x00\x12\x0c\n\x08NESTEROV\x10\x01\x12\x0b\n\x07\x41\x44\x41GRAD\x10\x02\"l\n\x0bSolverState\x12\x0c\n\x04iter\x18\x01 \x01(\x05\x12\x13\n\x0blearned_net\x18\x02 \x01(\t\x12!\n\x07history\x18\x03 \x03(\x0b\x32\x10.caffe.BlobProto\x12\x17\n\x0c\x63urrent_step\x18\x04 \x01(\x05:\x01\x30\"N\n\x08NetState\x12!\n\x05phase\x18\x01 \x01(\x0e\x32\x0c.caffe.Phase:\x04TEST\x12\x10\n\x05level\x18\x02 \x01(\x05:\x01\x30\x12\r\n\x05stage\x18\x03 \x03(\t\"s\n\x0cNetStateRule\x12\x1b\n\x05phase\x18\x01 \x01(\x0e\x32\x0c.caffe.Phase\x12\x11\n\tmin_level\x18\x02 \x01(\x05\x12\x11\n\tmax_level\x18\x03 \x01(\x05\x12\r\n\x05stage\x18\x04 \x03(\t\x12\x11\n\tnot_stage\x18\x05 \x03(\t\"\xa3\x01\n\tParamSpec\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x31\n\nshare_mode\x18\x02 \x01(\x0e\x32\x1d.caffe.ParamSpec.DimCheckMode\x12\x12\n\x07lr_mult\x18\x03 \x01(\x02:\x01\x31\x12\x15\n\ndecay_mult\x18\x04 \x01(\x02:\x01\x31\"*\n\x0c\x44imCheckMode\x12\n\n\x06STRICT\x10\x00\x12\x0e\n\nPERMISSIVE\x10\x01\"\x8c\x38\n\x0eLayerParameter\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x0e\n\x06\x62ottom\x18\x03 \x03(\t\x12\x0b\n\x03top\x18\x04 \x03(\t\x12\x1b\n\x05phase\x18\n \x01(\x0e\x32\x0c.caffe.Phase\x12\x13\n\x0bloss_weight\x18\x05 \x03(\x02\x12\x1f\n\x05param\x18\x06 \x03(\x0b\x32\x10.caffe.ParamSpec\x12\x1f\n\x05\x62lobs\x18\x07 \x03(\x0b\x32\x10.caffe.BlobProto\x12$\n\x07include\x18\x08 \x03(\x0b\x32\x13.caffe.NetStateRule\x12$\n\x07\x65xclude\x18\t \x03(\x0b\x32\x13.caffe.NetStateRule\x12\x1b\n\rtop_data_type\x18\x0b \x01(\t:\x04None\x12G\n\x11\x66orward_precision\x18\x0c \x01(\x0e\x32#.caffe.LayerParameter.PrecisionType:\x07\x44\x45\x46\x41ULT\x12\x37\n\x0ftransform_param\x18\x64 \x01(\x0b\x32\x1e.caffe.TransformationParameter\x12(\n\nloss_param\x18\x65 \x01(\x0b\x32\x14.caffe.LossParameter\x12\x30\n\x0e\x61\x63\x63uracy_param\x18\x66 \x01(\x0b\x32\x18.caffe.AccuracyParameter\x12,\n\x0c\x61rgmax_param\x18g \x01(\x0b\x32\x16.caffe.ArgMaxParameter\x12,\n\x0c\x63oncat_param\x18h \x01(\x0b\x32\x16.caffe.ConcatParameter\x12?\n\x16\x63ontrastive_loss_param\x18i \x01(\x0b\x32\x1f.caffe.ContrastiveLossParameter\x12\x36\n\x11\x63onvolution_param\x18j \x01(\x0b\x32\x1b.caffe.ConvolutionParameter\x12(\n\ndata_param\x18k \x01(\x0b\x32\x14.caffe.DataParameter\x12.\n\rdropout_param\x18l \x01(\x0b\x32\x17.caffe.DropoutParameter\x12\x33\n\x10\x64ummy_data_param\x18m \x01(\x0b\x32\x19.caffe.DummyDataParameter\x12.\n\reltwise_param\x18n \x01(\x0b\x32\x17.caffe.EltwiseParameter\x12&\n\texp_param\x18o \x01(\x0b\x32\x13.caffe.ExpParameter\x12\x31\n\x0fhdf5_data_param\x18p \x01(\x0b\x32\x18.caffe.HDF5DataParameter\x12\x35\n\x11hdf5_output_param\x18q \x01(\x0b\x32\x1a.caffe.HDF5OutputParameter\x12\x33\n\x10hinge_loss_param\x18r \x01(\x0b\x32\x19.caffe.HingeLossParameter\x12\x33\n\x10image_data_param\x18s \x01(\x0b\x32\x19.caffe.ImageDataParameter\x12\x39\n\x13infogain_loss_param\x18t \x01(\x0b\x32\x1c.caffe.InfogainLossParameter\x12\x39\n\x13inner_product_param\x18u \x01(\x0b\x32\x1c.caffe.InnerProductParameter\x12&\n\tlrn_param\x18v \x01(\x0b\x32\x13.caffe.LRNParameter\x12\x35\n\x11memory_data_param\x18w \x01(\x0b\x32\x1a.caffe.MemoryDataParameter\x12&\n\tmvn_param\x18x \x01(\x0b\x32\x13.caffe.MVNParameter\x12.\n\rpooling_param\x18y \x01(\x0b\x32\x17.caffe.PoolingParameter\x12@\n\x16pooling_specific_param\x18\xba\t \x01(\x0b\x32\x1f.caffe.PoolingSpecificParameter\x12*\n\x0bpower_param\x18z \x01(\x0b\x32\x15.caffe.PowerParameter\x12(\n\nrelu_param\x18{ \x01(\x0b\x32\x14.caffe.ReLUParameter\x12.\n\rsigmoid_param\x18| \x01(\x0b\x32\x17.caffe.SigmoidParameter\x12.\n\rsoftmax_param\x18} \x01(\x0b\x32\x17.caffe.SoftmaxParameter\x12*\n\x0bslice_param\x18~ \x01(\x0b\x32\x15.caffe.SliceParameter\x12(\n\ntanh_param\x18\x7f \x01(\x0b\x32\x14.caffe.TanHParameter\x12\x33\n\x0fthreshold_param\x18\x80\x01 \x01(\x0b\x32\x19.caffe.ThresholdParameter\x12\x36\n\x11window_data_param\x18\x81\x01 \x01(\x0b\x32\x1a.caffe.WindowDataParameter\x12-\n\x0cpython_param\x18\x82\x01 \x01(\x0b\x32\x16.caffe.PythonParameter\x12\'\n\tctc_param\x18\x83\x01 \x01(\x0b\x32\x13.caffe.CTCParameter\x12+\n\x0bprelu_param\x18\x84\x01 \x01(\x0b\x32\x15.caffe.PReLUParameter\x12/\n\rreshape_param\x18\x85\x01 \x01(\x0b\x32\x17.caffe.ReshapeParameter\x12\x36\n\x11roi_pooling_param\x18\x86\x01 \x01(\x0b\x32\x1a.caffe.ROIPoolingParameter\x12\x38\n\x12\x61\x66\x66ine_trans_param\x18\x87\x01 \x01(\x0b\x32\x1b.caffe.AffineTransParameter\x12\'\n\troi_param\x18\x88\x01 \x01(\x0b\x32\x13.caffe.ROIParameter\x12\x43\n\x18\x61\x66\x66ine_trans_point_param\x18\x89\x01 \x01(\x0b\x32 .caffe.AffineTransPointParameter\x12/\n\rpermute_param\x18\x8a\x01 \x01(\x0b\x32\x17.caffe.PermuteParameter\x12\x32\n\x0fprior_box_param\x18\x8b\x01 \x01(\x0b\x32\x18.caffe.PriorBoxParameter\x12=\n\x15\x63\x61lc_affine_mat_param\x18\x8c\x01 \x01(\x0b\x32\x1d.caffe.CalcAffineMatParameter\x12-\n\x0cinterp_param\x18\x8d\x01 \x01(\x0b\x32\x16.caffe.InterpParameter\x12@\n\x16\x64\x65tection_output_param\x18\x8e\x01 \x01(\x0b\x32\x1f.caffe.DetectionOutputParameter\x12+\n\x0brelu6_param\x18\x8f\x01 \x01(\x0b\x32\x15.caffe.ReLU6Parameter\x12)\n\nbias_param\x18\x90\x01 \x01(\x0b\x32\x14.caffe.BiasParameter\x12+\n\x0bzxybn_param\x18\x91\x01 \x01(\x0b\x32\x15.caffe.ZXYBNParameter\x12\x37\n\x11\x63orrelation_param\x18\x96\x01 \x01(\x0b\x32\x1b.caffe.CorrelationParameter\x12:\n\x13psroi_pooling_param\x18\x98\x01 \x01(\x0b\x32\x1c.caffe.PSROIPoolingParameter\x12\x36\n\x11nn_upsample_param\x18\xa3\x01 \x01(\x0b\x32\x1a.caffe.NNUpsampleParameter\x12\x32\n\x0froi_align_param\x18\x99\x01 \x01(\x0b\x32\x18.caffe.ROIAlignParameter\x12=\n\x12rpn_proposal_param\x18\xa6\x01 \x01(\x0b\x32 .caffe.RpnProposalLayerParameter\x12?\n\x16roi_mask_pooling_param\x18\xb0\x01 \x01(\x0b\x32\x1e.caffe.ROIMaskPoolingParameter\x12>\n\x15\x63hannel_shuffle_param\x18\xb1\x01 \x01(\x0b\x32\x1e.caffe.ChannelShuffleParameter\x12%\n\x08\x62n_param\x18\xb2\x01 \x01(\x0b\x32\x12.caffe.BNParameter\x12H\n\x1broi_align_pooling_pod_param\x18\xcd\x01 \x01(\x0b\x32\".caffe.ROIAlignPoolingPodParameter\x12\x41\n\x17roi_align_pooling_param\x18\xce\x01 \x01(\x0b\x32\x1f.caffe.ROIAlignPoolingParameter\x12\x45\n\x19psroi_align_pooling_param\x18\xcf\x01 \x01(\x0b\x32!.caffe.PSROIAlignPoolingParameter\x12?\n\x16\x61lignment_to_roi_param\x18\xda\x01 \x01(\x0b\x32\x1e.caffe.AlignmentToROIParameter\x12\x33\n\x0funpooling_param\x18\xf3\x01 \x01(\x0b\x32\x19.caffe.UnpoolingParameter\x12\'\n\tlog_param\x18\xf4\x01 \x01(\x0b\x32\x13.caffe.LogParameter\x12/\n\rreverse_param\x18\xad\x02 \x01(\x0b\x32\x17.caffe.ReverseParameter\x12)\n\nlstm_param\x18\xae\x02 \x01(\x0b\x32\x14.caffe.LSTMParameter\x12\x34\n\x10prior_vbox_param\x18\xb0\x02 \x01(\x0b\x32\x19.caffe.PriorVBoxParameter\x12\x39\n\x12\x62tcostvolume_param\x18\xe9\x07 \x01(\x0b\x32\x1c.caffe.BTCostVolumeParameter\x12\x34\n\x10moving_avg_param\x18\x87\x0e \x01(\x0b\x32\x19.caffe.MovingAvgParameter\x12/\n\rslicing_param\x18\xbc( \x01(\x0b\x32\x17.caffe.SlicingParameter\x12\x42\n\x17\x62ilateral_slicing_param\x18\xbd( \x01(\x0b\x32 .caffe.BilateralSlicingParameter\x12I\n\x1b\x62ilateral_slice_apply_param\x18\xbe( \x01(\x0b\x32#.caffe.BilateralSliceApplyParameter\x12Z\n$adaptive_bilateral_slice_apply_param\x18\xbf( \x01(\x0b\x32+.caffe.AdaptiveBilateralSliceApplyParameter\x12\x37\n\x11pointscurve_param\x18\xd0\x0f \x01(\x0b\x32\x1b.caffe.PointsCurveParameter\x12:\n\x13subpixel_down_param\x18\xd1\x0f \x01(\x0b\x32\x1c.caffe.SubpixelDownParameter\x12\x36\n\x11subpixel_up_param\x18\xd2\x0f \x01(\x0b\x32\x1a.caffe.SubpixelUpParameter\x12)\n\nclip_param\x18\xd3\x0f \x01(\x0b\x32\x14.caffe.ClipParameter\x12O\n\x1e\x65ltwise_affine_transform_param\x18\xd4\x0f \x01(\x0b\x32&.caffe.EltwiseAffineTransformParameter\x12)\n\ntile_param\x18\xd5\x0f \x01(\x0b\x32\x14.caffe.TileParameter\x12\'\n\tpad_param\x18\xd6\x0f \x01(\x0b\x32\x13.caffe.PadParameter\x12-\n\x0creduce_param\x18\xd7\x0f \x01(\x0b\x32\x16.caffe.ReduceParameter\x12;\n\x13\x63onvolution3d_param\x18\xd8\x0f \x01(\x0b\x32\x1d.caffe.Convolution3dParameter\x12?\n\x15\x64\x65\x63onvolution3d_param\x18\xd9\x0f \x01(\x0b\x32\x1f.caffe.Deconvolution3dParameter\x12-\n\x0crelu3d_param\x18\xda\x0f \x01(\x0b\x32\x16.caffe.ReLU3dParameter\x12-\n\rprelu3d_param\x18\xdb\x0f \x01(\x0b\x32\x15.caffe.PReLUParameter\x12\x33\n\x12leaky_relu3d_param\x18\xdc\x0f \x01(\x0b\x32\x16.caffe.ReLU3dParameter\x12,\n\x0b\x65lu3d_param\x18\xdd\x0f \x01(\x0b\x32\x16.caffe.ReLU3dParameter\x12\x37\n\x11groupnorm3d_param\x18\xde\x0f \x01(\x0b\x32\x1b.caffe.GroupNorm3dParameter\x12\x37\n\x11\x62\x61tchnorm3d_param\x18\xdf\x0f \x01(\x0b\x32\x1b.caffe.BatchNorm3dParameter\x12\x36\n\x12maxpooling3d_param\x18\xe0\x0f \x01(\x0b\x32\x19.caffe.Pooling3dParameter\x12\x36\n\x12\x61vepooling3d_param\x18\xe1\x0f \x01(\x0b\x32\x19.caffe.Pooling3dParameter\x12\x31\n\x0f\x64ropout3d_param\x18\xe2\x0f \x01(\x0b\x32\x17.caffe.DropoutParameter\x12\x31\n\x0einterp3d_param\x18\xe3\x0f \x01(\x0b\x32\x18.caffe.Interp3dParameter\x12\x33\n\x0fpooling3d_param\x18\xe4\x0f \x01(\x0b\x32\x19.caffe.Pooling3dParameter\x12\x31\n\x0freshape3d_param\x18\xe6\x0f \x01(\x0b\x32\x17.caffe.ReshapeParameter\x12\x31\n\x0fsoftmax3d_param\x18\xe7\x0f \x01(\x0b\x32\x17.caffe.SoftmaxParameter\x12/\n\x0e\x61rgmax3d_param\x18\xe8\x0f \x01(\x0b\x32\x16.caffe.ArgMaxParameter\x12<\n\x15inner_product3d_param\x18\xe9\x0f \x01(\x0b\x32\x1c.caffe.InnerProductParameter\x12-\n\rslice3d_param\x18\xea\x0f \x01(\x0b\x32\x15.caffe.SliceParameter\x12/\n\x0e\x63oncat3d_param\x18\xeb\x0f \x01(\x0b\x32\x16.caffe.ConcatParameter\x12\'\n\nbn3d_param\x18\xec\x0f \x01(\x0b\x32\x12.caffe.BNParameter\x12/\n\x0ematmul3d_param\x18\xed\x0f \x01(\x0b\x32\x16.caffe.MatMulParameter\x12\x35\n\x11transpose3d_param\x18\xee\x0f \x01(\x0b\x32\x19.caffe.TransposeParameter\x12+\n\x0bscale_param\x18\x89\' \x01(\x0b\x32\x15.caffe.ScaleParameter\x12)\n\ncrop_param\x18\x8a\' \x01(\x0b\x32\x14.caffe.CropParameter\x12\x33\n\x0frecurrent_param\x18\x8c\' \x01(\x0b\x32\x19.caffe.RecurrentParameter\x12\x34\n\x10\x62\x61tch_norm_param\x18\x8d\' \x01(\x0b\x32\x19.caffe.BatchNormParameter\x12/\n\rflatten_param\x18\x8e\' \x01(\x0b\x32\x17.caffe.FlattenParameter\x12-\n\x0cscales_param\x18\x8f\' \x01(\x0b\x32\x16.caffe.ScalesParameter\x12\x33\n\x0ftranspose_param\x18\x90\' \x01(\x0b\x32\x19.caffe.TransposeParameter\x12\x35\n\rheatmap_param\x18\x91\' \x01(\x0b\x32\x1d.caffe.HeatMap2CoordParameter\x12\x33\n\x0fnormalize_param\x18\x92\' \x01(\x0b\x32\x19.caffe.NormalizeParameter\x12;\n\x13\x63orrelation2d_param\x18\x93\' \x01(\x0b\x32\x1d.caffe.Correlation2DParameter\x12-\n\x0cmatmul_param\x18\x94\' \x01(\x0b\x32\x16.caffe.MatMulParameter\x12\x33\n\x0fparameter_param\x18\x95\' \x01(\x0b\x32\x19.caffe.ParameterParameter\x12\x39\n\x12pixelshuffle_param\x18\x96\' \x01(\x0b\x32\x1c.caffe.PixelShuffleParameter\x12:\n\x13roi_transform_param\x18\x97\' \x01(\x0b\x32\x1c.caffe.ROITransformParameter\x12:\n\x13instance_norm_param\x18\x98\' \x01(\x0b\x32\x1c.caffe.InstanceNormParameter\x12\x36\n\x11grid_sample_param\x18\x9a\' \x01(\x0b\x32\x1a.caffe.GridSampleParameter\x12\x31\n\x0ereducel2_param\x18\x99\' \x01(\x0b\x32\x18.caffe.ReduceL2Parameter\x12)\n\nmean_param\x18\x9b\' \x01(\x0b\x32\x14.caffe.MeanParameter\x12\x31\n\x0evariance_param\x18\x9c\' \x01(\x0b\x32\x18.caffe.VarianceParameter\x12;\n\x14grid_sample_3d_param\x18\x9d\' \x01(\x0b\x32\x1c.caffe.GridSample3DParameter\x12\'\n\tgru_param\x18\x9e\' \x01(\x0b\x32\x13.caffe.GRUParameter\x12=\n\x14\x63orrelationmig_param\x18\x9f\' \x01(\x0b\x32\x1e.caffe.CorrelationMigParameter\x12/\n\rargsort_param\x18\xa0\' \x01(\x0b\x32\x17.caffe.ArgSortParameter\x12/\n\rcumprod_param\x18\xa1\' \x01(\x0b\x32\x17.caffe.CumProdParameter\x12<\n\x14main_transform_param\x18\xa2\' \x01(\x0b\x32\x1d.caffe.MainTransformParameter\x12\x31\n\x0equantize_param\x18\xf0. \x03(\x0b\x32\x18.caffe.QuantizeParameter\x12\x19\n\rquantize_bits\x18\xf1. \x01(\r:\x01\x30\"b\n\rPrecisionType\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\x08\n\x04INT8\x10\x01\x12\x0b\n\x07\x46LOAT16\x10\x02\x12\x0b\n\x07\x46LOAT32\x10\x03\x12\t\n\x05INT4B\x10\x04\x12\t\n\x05UINT8\x10\x05\x12\n\n\x06UINT16\x10\x06\"\xfb\x02\n\x16\x43onvolution3dParameter\x12\x12\n\nnum_output\x18\x01 \x01(\r\x12\x17\n\tbias_term\x18\x02 \x01(\x08:\x04true\x12\x0e\n\x03pad\x18\x03 \x01(\r:\x01\x30\x12\x10\n\x05pad_d\x18\x17 \x01(\r:\x01\x30\x12\x10\n\x05pad_h\x18\t \x01(\r:\x01\x30\x12\x10\n\x05pad_w\x18\n \x01(\r:\x01\x30\x12\x13\n\x0bkernel_size\x18\x04 \x01(\r\x12\x10\n\x08kernel_d\x18\x16 \x01(\r\x12\x10\n\x08kernel_h\x18\x0b \x01(\r\x12\x10\n\x08kernel_w\x18\x0c \x01(\r\x12\x10\n\x05group\x18\x05 \x01(\r:\x01\x31\x12\x11\n\x06stride\x18\x06 \x01(\r:\x01\x31\x12\x10\n\x08stride_d\x18\x15 \x01(\r\x12\x10\n\x08stride_h\x18\r \x01(\r\x12\x10\n\x08stride_w\x18\x0e \x01(\r\x12\x0f\n\x04hole\x18\x11 \x01(\r:\x01\x31\x12\x11\n\x06hole_d\x18\x13 \x01(\r:\x01\x31\x12\x11\n\x06hole_h\x18\x12 \x01(\r:\x01\x31\x12\x11\n\x06hole_w\x18\x14 \x01(\r:\x01\x31\"\xa8\x02\n\x1f\x43onvolutionTranspose3dParameter\x12\x12\n\nnum_output\x18\x01 \x01(\r\x12\x17\n\tbias_term\x18\x02 \x01(\x08:\x04true\x12\x0e\n\x03pad\x18\x03 \x01(\r:\x01\x30\x12\x10\n\x05pad_h\x18\t \x01(\r:\x01\x30\x12\x10\n\x05pad_w\x18\n \x01(\r:\x01\x30\x12\x13\n\x0bkernel_size\x18\x04 \x01(\r\x12\x10\n\x08kernel_h\x18\x0b \x01(\r\x12\x10\n\x08kernel_w\x18\x0c \x01(\r\x12\x10\n\x05group\x18\x05 \x01(\r:\x01\x31\x12\x11\n\x06stride\x18\x06 \x01(\r:\x01\x31\x12\x10\n\x08stride_h\x18\r \x01(\r\x12\x10\n\x08stride_w\x18\x0e \x01(\r\x12\x0f\n\x04hole\x18\x11 \x01(\r:\x01\x31\x12\x11\n\x06hole_h\x18\x12 \x01(\r:\x01\x31\"\xd3\x03\n\x18\x44\x65\x63onvolution3dParameter\x12\x12\n\nnum_output\x18\x01 \x01(\r\x12\x17\n\tbias_term\x18\x02 \x01(\x08:\x04true\x12\x0e\n\x03pad\x18\x03 \x01(\r:\x01\x30\x12\x10\n\x05pad_d\x18\x17 \x01(\r:\x01\x30\x12\x10\n\x05pad_h\x18\t \x01(\r:\x01\x30\x12\x10\n\x05pad_w\x18\n \x01(\r:\x01\x30\x12\x13\n\x0bkernel_size\x18\x04 \x01(\r\x12\x10\n\x08kernel_d\x18\x16 \x01(\r\x12\x10\n\x08kernel_h\x18\x0b \x01(\r\x12\x10\n\x08kernel_w\x18\x0c \x01(\r\x12\x10\n\x05group\x18\x05 \x01(\r:\x01\x31\x12\x11\n\x06stride\x18\x06 \x01(\r:\x01\x31\x12\x10\n\x08stride_d\x18\x15 \x01(\r\x12\x10\n\x08stride_h\x18\r \x01(\r\x12\x10\n\x08stride_w\x18\x0e \x01(\r\x12\x0f\n\x04hole\x18\x11 \x01(\r:\x01\x31\x12\x11\n\x06hole_d\x18\x13 \x01(\r:\x01\x31\x12\x11\n\x06hole_h\x18\x12 \x01(\r:\x01\x31\x12\x11\n\x06hole_w\x18\x14 \x01(\r:\x01\x31\x12\x12\n\x07out_pad\x18\x18 \x01(\r:\x01\x30\x12\x14\n\tout_pad_d\x18\x19 \x01(\r:\x01\x30\x12\x14\n\tout_pad_h\x18\x1a \x01(\r:\x01\x30\x12\x14\n\tout_pad_w\x18\x1b \x01(\r:\x01\x30\"K\n\x0fReLU3dParameter\x12\x19\n\x0enegative_slope\x18\x01 \x01(\x02:\x01\x30\x12\x1d\n\x0e\x63hannel_shared\x18\x02 \x01(\x08:\x05\x66\x61lse\"\x83\x01\n\x14GroupNorm3dParameter\x12\x10\n\x05group\x18\x01 \x01(\r:\x01\x31\x12$\n\x17moving_average_fraction\x18\x04 \x01(\x02:\x03\x30.1\x12\x1f\n\x10use_global_stats\x18\x02 \x01(\x08:\x05\x66\x61lse\x12\x12\n\x03\x65ps\x18\x03 \x01(\x02:\x05\x31\x65-06\"q\n\x14\x42\x61tchNorm3dParameter\x12$\n\x17moving_average_fraction\x18\x03 \x01(\x02:\x03\x30.1\x12\x1f\n\x10use_global_stats\x18\x01 \x01(\x08:\x05\x66\x61lse\x12\x12\n\x03\x65ps\x18\x02 \x01(\x02:\x05\x31\x65-06\"\xff\x02\n\x12Pooling3dParameter\x12\x37\n\x04pool\x18\x01 \x01(\x0e\x32$.caffe.Pooling3dParameter.PoolMethod:\x03MAX\x12\x0e\n\x03pad\x18\x04 \x01(\r:\x01\x30\x12\x10\n\x05pad_d\x18\x0b \x01(\r:\x01\x30\x12\x10\n\x05pad_h\x18\t \x01(\r:\x01\x30\x12\x10\n\x05pad_w\x18\n \x01(\r:\x01\x30\x12\x13\n\x0bkernel_size\x18\x02 \x01(\r\x12\x10\n\x08kernel_d\x18\x0e \x01(\r\x12\x10\n\x08kernel_h\x18\x05 \x01(\r\x12\x10\n\x08kernel_w\x18\x06 \x01(\r\x12\x11\n\x06stride\x18\x03 \x01(\r:\x01\x31\x12\x10\n\x08stride_d\x18\x0f \x01(\r\x12\x10\n\x08stride_h\x18\x07 \x01(\r\x12\x10\n\x08stride_w\x18\x08 \x01(\r\x12\x1d\n\x0eglobal_pooling\x18\x0c \x01(\x08:\x05\x66\x61lse\x12\x17\n\tceil_mode\x18\r \x01(\x08:\x04true\"\x1e\n\nPoolMethod\x12\x07\n\x03MAX\x10\x00\x12\x07\n\x03\x41VE\x10\x01\"\x86\x01\n\x12NormalizeParameter\x12\x10\n\x05\x63oeff\x18\x01 \x01(\x02:\x01\x30\x12;\n\x04mode\x18\x02 \x01(\x0e\x32).caffe.NormalizeParameter.NormalizeMethod:\x02L2\"!\n\x0fNormalizeMethod\x12\x06\n\x02L1\x10\x01\x12\x06\n\x02L2\x10\x02\"a\n\x16HeatMap2CoordParameter\x12\x12\n\x07\x63oord_h\x18\x01 \x01(\x05:\x01\x30\x12\x12\n\x07\x63oord_w\x18\x02 \x01(\x05:\x01\x30\x12\x1f\n\x10\x63oord_reposition\x18\x03 \x01(\x08:\x05\x66\x61lse\"]\n\rBiasParameter\x12\x0f\n\x04\x61xis\x18\x01 \x01(\x05:\x01\x31\x12\x13\n\x08num_axes\x18\x02 \x01(\x05:\x01\x31\x12&\n\x06\x66iller\x18\x03 \x01(\x0b\x32\x16.caffe.FillerParameter\"x\n\x17\x41lignmentToROIParameter\x12\x16\n\x0bmouth_scale\x18\x01 \x01(\x02:\x01\x31\x12\x14\n\teye_scale\x18\x02 \x01(\x02:\x01\x31\x12\x16\n\tmax_width\x18\x03 \x01(\x05:\x03\x32\x35\x36\x12\x17\n\nmax_height\x18\x04 \x01(\x05:\x03\x32\x35\x36\"u\n\x18ROIAlignPoolingParameter\x12\x13\n\x08pooled_h\x18\x01 \x01(\r:\x01\x30\x12\x13\n\x08pooled_w\x18\x02 \x01(\r:\x01\x30\x12\x18\n\rspatial_scale\x18\x03 \x01(\x02:\x01\x31\x12\x15\n\nsample_num\x18\x04 \x01(\x05:\x01\x31\"x\n\x1bROIAlignPoolingPodParameter\x12\x13\n\x08pooled_h\x18\x01 \x01(\r:\x01\x30\x12\x13\n\x08pooled_w\x18\x02 \x01(\r:\x01\x30\x12\x18\n\rspatial_scale\x18\x03 \x01(\x02:\x01\x31\x12\x15\n\nsample_num\x18\x04 \x01(\x05:\x01\x31\"r\n\x1aPSROIAlignPoolingParameter\x12\x15\n\rspatial_scale\x18\x01 \x02(\x02\x12\x12\n\noutput_dim\x18\x02 \x02(\x05\x12\x12\n\ngroup_size\x18\x03 \x02(\x05\x12\x15\n\nsample_num\x18\x04 \x01(\x05:\x01\x31\"\xb7\x03\n\x12UnpoolingParameter\x12;\n\x06unpool\x18\x01 \x01(\x0e\x32&.caffe.UnpoolingParameter.UnpoolMethod:\x03MAX\x12\x0e\n\x03pad\x18\x04 \x01(\r:\x01\x30\x12\x10\n\x05pad_h\x18\t \x01(\r:\x01\x30\x12\x10\n\x05pad_w\x18\n \x01(\r:\x01\x30\x12\x13\n\x0bkernel_size\x18\x02 \x01(\r\x12\x10\n\x08kernel_h\x18\x05 \x01(\r\x12\x10\n\x08kernel_w\x18\x06 \x01(\r\x12\x11\n\x06stride\x18\x03 \x01(\r:\x01\x31\x12\x10\n\x08stride_h\x18\x07 \x01(\r\x12\x10\n\x08stride_w\x18\x08 \x01(\r\x12\x13\n\x0bunpool_size\x18\x0c \x01(\r\x12\x10\n\x08unpool_h\x18\r \x01(\r\x12\x10\n\x08unpool_w\x18\x0e \x01(\r\x12\x39\n\x06\x65ngine\x18\x0b \x01(\x0e\x32 .caffe.UnpoolingParameter.Engine:\x07\x44\x45\x46\x41ULT\"*\n\x0cUnpoolMethod\x12\x07\n\x03MAX\x10\x00\x12\x07\n\x03\x41VE\x10\x01\x12\x08\n\x04TILE\x10\x02\" \n\x06\x45ngine\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\t\n\x05\x43\x41\x46\x46\x45\x10\x01\"#\n\x10ReverseParameter\x12\x0f\n\x04\x61xis\x18\x01 \x01(\x05:\x01\x30\"\xb5\x01\n\rLSTMParameter\x12\x12\n\nnum_output\x18\x01 \x01(\r\x12\x1d\n\x12\x63lipping_threshold\x18\x02 \x01(\x02:\x01\x30\x12-\n\rweight_filler\x18\x03 \x01(\x0b\x32\x16.caffe.FillerParameter\x12+\n\x0b\x62ias_filler\x18\x04 \x01(\x0b\x32\x16.caffe.FillerParameter\x12\x15\n\nbatch_size\x18\x05 \x01(\r:\x01\x31\"\xd1\x01\n\x12PriorVBoxParameter\x12\x0e\n\x06height\x18\x01 \x03(\x02\x12\x11\n\x05width\x18\x02 \x01(\x02:\x02\x31\x36\x12\x13\n\x04\x63lip\x18\x05 \x01(\x08:\x05\x66\x61lse\x12\x10\n\x08variance\x18\x06 \x03(\x02\x12\x10\n\x08img_size\x18\x07 \x01(\r\x12\r\n\x05img_h\x18\x08 \x01(\r\x12\r\n\x05img_w\x18\t \x01(\r\x12\x0c\n\x04step\x18\n \x01(\x02\x12\x0e\n\x06step_h\x18\x0b \x01(\x02\x12\x0e\n\x06step_w\x18\x0c \x01(\x02\x12\x13\n\x06offset\x18\r \x01(\x02:\x03\x30.5\"(\n\x13NNUpsampleParameter\x12\x11\n\x06resize\x18\x01 \x01(\r:\x01\x32\"\xc1\x03\n\x15\x42TCostVolumeParameter\x12\x18\n\rmin_disparity\x18\x01 \x01(\x05:\x01\x30\x12\x19\n\rdisparity_num\x18\x02 \x01(\r:\x02\x36\x34\x12\x1d\n\x0esave_disparity\x18\x03 \x01(\x08:\x05\x66\x61lse\x12\x11\n\x06step_w\x18\x04 \x01(\r:\x01\x31\x12\x11\n\x06step_h\x18\x05 \x01(\r:\x01\x31\x12\x14\n\tstep_mode\x18\x06 \x01(\r:\x01\x30\x12V\n\x10\x63ost_domain_type\x18\x07 \x01(\x0e\x32+.caffe.BTCostVolumeParameter.CostDomainType:\x0fSOBEL_AND_IMAGE\x12L\n\x0eimplement_mode\x18\x08 \x01(\x0e\x32*.caffe.BTCostVolumeParameter.ImplementMode:\x08GPU_MODE\"E\n\x0e\x43ostDomainType\x12\x0e\n\nSOBEL_ONLY\x10\x00\x12\x0e\n\nIMAGE_ONLY\x10\x01\x12\x13\n\x0fSOBEL_AND_IMAGE\x10\x02\"+\n\rImplementMode\x12\x0c\n\x08GPU_MODE\x10\x00\x12\x0c\n\x08\x43PU_MODE\x10\x01\"\xdd\x02\n\x14\x43orrelationParameter\x12\x0e\n\x03pad\x18\x02 \x01(\r:\x01\x30\x12\x13\n\x0bkernel_size\x18\x03 \x01(\r\x12\x18\n\x10max_displacement\x18\x04 \x01(\r\x12\x13\n\x08stride_1\x18\x05 \x01(\r:\x01\x31\x12\x13\n\x08stride_2\x18\x06 \x01(\r:\x01\x31\x12\x1b\n\x10single_direction\x18\x08 \x01(\x05:\x01\x30\x12\x15\n\x06\x64o_abs\x18\x07 \x01(\x08:\x05\x66\x61lse\x12O\n\x10\x63orrelation_type\x18\x0f \x01(\x0e\x32+.caffe.CorrelationParameter.CorrelationType:\x08MULTIPLY\x12\x14\n\tpad_shift\x18\x10 \x01(\x05:\x01\x30\x12\x12\n\x03mvn\x18\t \x01(\x08:\x05\x66\x61lse\"-\n\x0f\x43orrelationType\x12\x0c\n\x08MULTIPLY\x10\x00\x12\x0c\n\x08SUBTRACT\x10\x01\"l\n\x15PSROIPoolingParameter\x12\x15\n\rspatial_scale\x18\x01 \x02(\x02\x12\x12\n\noutput_dim\x18\x02 \x02(\x05\x12\x12\n\ngroup_size\x18\x03 \x02(\x05\x12\x14\n\troi_scale\x18\x04 \x01(\x02:\x01\x31\"\xa5\x01\n\x0eScaleParameter\x12\x0f\n\x04\x61xis\x18\x01 \x01(\x05:\x01\x31\x12\x13\n\x08num_axes\x18\x02 \x01(\x05:\x01\x31\x12&\n\x06\x66iller\x18\x03 \x01(\x0b\x32\x16.caffe.FillerParameter\x12\x18\n\tbias_term\x18\x04 \x01(\x08:\x05\x66\x61lse\x12+\n\x0b\x62ias_filler\x18\x05 \x01(\x0b\x32\x16.caffe.FillerParameter\"4\n\x0fScalesParameter\x12\x10\n\x05\x61lpha\x18\x01 \x01(\x02:\x01\x31\x12\x0f\n\x04\x62\x65ta\x18\x02 \x01(\x02:\x01\x30\"!\n\x12TransposeParameter\x12\x0b\n\x03\x64im\x18\x01 \x03(\x05\"W\n\x11ROIAlignParameter\x12\x13\n\x08pooled_h\x18\x01 \x01(\r:\x01\x30\x12\x13\n\x08pooled_w\x18\x02 \x01(\r:\x01\x30\x12\x18\n\rspatial_scale\x18\x03 \x01(\x02:\x01\x31\"\xe8\x05\n\x19RpnProposalLayerParameter\x12\x17\n\x0b\x66\x65\x61t_stride\x18\x01 \x01(\x05:\x02\x31\x36\x12\x19\n\x0e\x61llowed_border\x18\x02 \x01(\x05:\x01\x30\x12G\n\x0ftest_desc_param\x18\x03 \x01(\x0b\x32..caffe.RpnProposalLayerParameter.TestDescParam\x12I\n\x10train_desc_param\x18\x04 \x01(\x0b\x32/.caffe.RpnProposalLayerParameter.TrainDescParam\x12\x15\n\ranchor_scales\x18\x05 \x03(\x05\x12\x15\n\ranchor_ratios\x18\x06 \x03(\x02\x12\x10\n\x05score\x18\x07 \x01(\x02:\x01\x30\x12\x1d\n\x0ewithOutSoftmax\x18\x08 \x01(\x08:\x05\x66\x61lse\x12\x1c\n\x11\x62\x62ox_reg_xy_limit\x18\t \x01(\x02:\x01\x35\x12\x1b\n\x10\x62\x62ox_reg_hw_temp\x18\n \x01(\x02:\x01\x31\x12&\n\x1b\x62\x62ox_reg_hw_ratio_thresh_hi\x18\x0b \x01(\x02:\x01\x35\x12(\n\x1b\x62\x62ox_reg_hw_ratio_thresh_lo\x18\x0c \x01(\x02:\x03\x30.2\x1a\x88\x01\n\rTestDescParam\x12\x1f\n\x11rpn_pre_nms_top_n\x18\x01 \x01(\x05:\x04\x36\x30\x30\x30\x12\x1f\n\x12rpn_post_nms_top_n\x18\x02 \x01(\x05:\x03\x33\x30\x30\x12\x18\n\x0crpn_min_size\x18\x03 \x01(\x05:\x02\x31\x36\x12\x1b\n\x0erpn_nms_thresh\x18\x04 \x01(\x02:\x03\x30.7\x1a\x8b\x01\n\x0eTrainDescParam\x12 \n\x11rpn_pre_nms_top_n\x18\x01 \x01(\x05:\x05\x31\x32\x30\x30\x30\x12 \n\x12rpn_post_nms_top_n\x18\x02 \x01(\x05:\x04\x32\x30\x30\x30\x12\x18\n\x0crpn_min_size\x18\x03 \x01(\x05:\x02\x31\x36\x12\x1b\n\x0erpn_nms_thresh\x18\x04 \x01(\x02:\x03\x30.7\"\xa0\x01\n\x17ROIMaskPoolingParameter\x12\x13\n\x08pooled_h\x18\x01 \x01(\x05:\x01\x30\x12\x13\n\x08pooled_w\x18\x02 \x01(\x05:\x01\x30\x12\x18\n\rspatial_scale\x18\x03 \x01(\x02:\x01\x31\x12\x14\n\thalf_part\x18\x04 \x01(\x05:\x01\x30\x12\x14\n\troi_scale\x18\x05 \x01(\x02:\x01\x31\x12\x15\n\nmask_scale\x18\x06 \x01(\x02:\x01\x30\"\x91\x02\n\x0eZXYBNParameter\x12,\n\x0cslope_filler\x18\x01 \x01(\x0b\x32\x16.caffe.FillerParameter\x12+\n\x0b\x62ias_filler\x18\x02 \x01(\x0b\x32\x16.caffe.FillerParameter\x12\x15\n\x08momentum\x18\x03 \x01(\x02:\x03\x30.9\x12\x12\n\x03\x65ps\x18\x04 \x01(\x02:\x05\x31\x65-05\x12\x15\n\x06\x66rozen\x18\x05 \x01(\x08:\x05\x66\x61lse\x12\x35\n\x06\x65ngine\x18\x06 \x01(\x0e\x32\x1c.caffe.ZXYBNParameter.Engine:\x07\x44\x45\x46\x41ULT\"+\n\x06\x45ngine\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\t\n\x05\x43\x41\x46\x46\x45\x10\x01\x12\t\n\x05\x43UDNN\x10\x02\"\x7f\n\x17TransformationParameter\x12\x10\n\x05scale\x18\x01 \x01(\x02:\x01\x31\x12\x15\n\x06mirror\x18\x02 \x01(\x08:\x05\x66\x61lse\x12\x14\n\tcrop_size\x18\x03 \x01(\r:\x01\x30\x12\x11\n\tmean_file\x18\x04 \x01(\t\x12\x12\n\nmean_value\x18\x05 \x03(\x02\">\n\rLossParameter\x12\x14\n\x0cignore_label\x18\x01 \x01(\x05\x12\x17\n\tnormalize\x18\x02 \x01(\x08:\x04true\"%\n\x11\x41\x63\x63uracyParameter\x12\x10\n\x05top_k\x18\x01 \x01(\r:\x01\x31\"M\n\x0f\x41rgMaxParameter\x12\x1a\n\x0bout_max_val\x18\x01 \x01(\x08:\x05\x66\x61lse\x12\x10\n\x05top_k\x18\x02 \x01(\r:\x01\x31\x12\x0c\n\x04\x61xis\x18\x03 \x01(\x05\"9\n\x0f\x43oncatParameter\x12\x0f\n\x04\x61xis\x18\x02 \x01(\x05:\x01\x31\x12\x15\n\nconcat_dim\x18\x01 \x01(\r:\x01\x31\"-\n\x18\x43ontrastiveLossParameter\x12\x11\n\x06margin\x18\x01 \x01(\x02:\x01\x31\"\xec\x0b\n\x14\x43onvolutionParameter\x12\x12\n\nnum_output\x18\x01 \x01(\r\x12\x17\n\tbias_term\x18\x02 \x01(\x08:\x04true\x12\x0e\n\x03pad\x18\x03 \x01(\r:\x01\x30\x12\x10\n\x05pad_h\x18\t \x01(\r:\x01\x30\x12\x10\n\x05pad_w\x18\n \x01(\r:\x01\x30\x12\x13\n\x0bkernel_size\x18\x04 \x01(\r\x12\x10\n\x08kernel_h\x18\x0b \x01(\r\x12\x10\n\x08kernel_w\x18\x0c \x01(\r\x12\x10\n\x05group\x18\x05 \x01(\r:\x01\x31\x12\x11\n\x06stride\x18\x06 \x01(\r:\x01\x31\x12\x10\n\x08stride_h\x18\r \x01(\r\x12\x10\n\x08stride_w\x18\x0e \x01(\r\x12\x0f\n\x04hole\x18\x11 \x01(\r:\x01\x31\x12\x11\n\x06hole_h\x18\x12 \x01(\r:\x01\x31\x12\x11\n\x06hole_w\x18\x13 \x01(\r:\x01\x31\x12-\n\rweight_filler\x18\x07 \x01(\x0b\x32\x16.caffe.FillerParameter\x12+\n\x0b\x62ias_filler\x18\x08 \x01(\x0b\x32\x16.caffe.FillerParameter\x12;\n\x06\x65ngine\x18\x0f \x01(\x0e\x32\".caffe.ConvolutionParameter.Engine:\x07\x44\x45\x46\x41ULT\x12o\n\x0b\x46orwardAlgo\x18\x14 \x01(\x0e\x32\x37.caffe.ConvolutionParameter.PPLConvolutionForwardAlgo_t:!PPL_CONVOLUTION_FORWARD_ALGO_NONE\x12\x16\n\x0bntile_width\x18H \x01(\x05:\x01\x31\x12\x17\n\x0cntile_height\x18I \x01(\x05:\x01\x31\x12\x30\n\x0equantize_param\x18J \x01(\x0b\x32\x18.caffe.QuantizeParameter\x12\x16\n\x07is_relu\x18M \x01(\x08:\x05\x66\x61lse\x12;\n\x19perchannel_quantize_param\x18K \x03(\x0b\x32\x18.caffe.QuantizeParameter\x12\x12\n\x07out_pad\x18N \x01(\r:\x01\x30\x12\x14\n\tout_pad_h\x18O \x01(\r:\x01\x30\x12\x14\n\tout_pad_w\x18P \x01(\r:\x01\x30\"+\n\x06\x45ngine\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\t\n\x05\x43\x41\x46\x46\x45\x10\x01\x12\t\n\x05\x43UDNN\x10\x02\"\x8a\x05\n\x1bPPLConvolutionForwardAlgo_t\x12.\n!PPL_CONVOLUTION_FORWARD_ALGO_NONE\x10\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x12.\n*PPL_CONVOLUTION_FORWARD_ALGO_IMPLICIT_GEMM\x10\x00\x12\x36\n2PPL_CONVOLUTION_FORWARD_ALGO_IMPLICIT_PRECOMP_GEMM\x10\x01\x12%\n!PPL_CONVOLUTION_FORWARD_ALGO_GEMM\x10\x02\x12\'\n#PPL_CONVOLUTION_FORWARD_ALGO_DIRECT\x10\x03\x12$\n PPL_CONVOLUTION_FORWARD_ALGO_FFT\x10\x04\x12+\n\'PPL_CONVOLUTION_FORWARD_ALGO_FFT_TILING\x10\x05\x12\x30\n,PPL_CONVOLUTION_FORWARD_ALGO_WINOGRAD_BLK2X2\x10\x06\x12\x30\n,PPL_CONVOLUTION_FORWARD_ALGO_WINOGRAD_BLK4X4\x10\x07\x12\x30\n,PPL_CONVOLUTION_FORWARD_ALGO_WINOGRAD_BLK6X6\x10\x08\x12\x37\n3PPL_CONVOLUTION_FORWARD_ALGO_WINOGRAD_BLK4x4_FLT3x3\x10\t\x12\x37\n3PPL_CONVOLUTION_FORWARD_ALGO_WINOGRAD_BLK4x4_FLT5X5\x10\n\x12(\n$PPL_CONVOLUTION_FORWARD_ALGO_INVGEMM\x10\x0b\"\x8f\x02\n\rDataParameter\x12\x0e\n\x06source\x18\x01 \x01(\t\x12\x12\n\nbatch_size\x18\x04 \x01(\r\x12\x14\n\trand_skip\x18\x07 \x01(\r:\x01\x30\x12\x31\n\x07\x62\x61\x63kend\x18\x08 \x01(\x0e\x32\x17.caffe.DataParameter.DB:\x07LEVELDB\x12\x10\n\x05scale\x18\x02 \x01(\x02:\x01\x31\x12\x11\n\tmean_file\x18\x03 \x01(\t\x12\x14\n\tcrop_size\x18\x05 \x01(\r:\x01\x30\x12\x15\n\x06mirror\x18\x06 \x01(\x08:\x05\x66\x61lse\x12\"\n\x13\x66orce_encoded_color\x18\t \x01(\x08:\x05\x66\x61lse\"\x1b\n\x02\x44\x42\x12\x0b\n\x07LEVELDB\x10\x00\x12\x08\n\x04LMDB\x10\x01\"K\n\x1eNonMaximumSuppressionParameter\x12\x1a\n\rnms_threshold\x18\x01 \x01(\x02:\x03\x30.3\x12\r\n\x05top_k\x18\x02 \x01(\x05\"\xaa\x01\n\x13SaveOutputParameter\x12\x18\n\x10output_directory\x18\x01 \x01(\t\x12\x1a\n\x12output_name_prefix\x18\x02 \x01(\t\x12\x15\n\routput_format\x18\x03 \x01(\t\x12\x16\n\x0elabel_map_file\x18\x04 \x01(\t\x12\x16\n\x0ename_size_file\x18\x05 \x01(\t\x12\x16\n\x0enum_test_image\x18\x06 \x01(\r\"\xb4\x03\n\x18\x44\x65tectionOutputParameter\x12\x13\n\x0bnum_classes\x18\x01 \x01(\r\x12\x1c\n\x0eshare_location\x18\x02 \x01(\x08:\x04true\x12\x1e\n\x13\x62\x61\x63kground_label_id\x18\x03 \x01(\x05:\x01\x30\x12\x38\n\tnms_param\x18\x04 \x01(\x0b\x32%.caffe.NonMaximumSuppressionParameter\x12\x35\n\x11save_output_param\x18\x05 \x01(\x0b\x32\x1a.caffe.SaveOutputParameter\x12<\n\tcode_type\x18\x06 \x01(\x0e\x32!.caffe.PriorBoxParameter.CodeType:\x06\x43ORNER\x12)\n\x1avariance_encoded_in_target\x18\x08 \x01(\x08:\x05\x66\x61lse\x12\x16\n\nkeep_top_k\x18\x07 \x01(\x05:\x02-1\x12\x1c\n\x14\x63onfidence_threshold\x18\t \x01(\x02\x12\x18\n\tvisualize\x18\n \x01(\x08:\x05\x66\x61lse\x12\x1b\n\x13visualize_threshold\x18\x0b \x01(\x02\".\n\x10\x44ropoutParameter\x12\x1a\n\rdropout_ratio\x18\x01 \x01(\x02:\x03\x30.5\"\xa0\x01\n\x12\x44ummyDataParameter\x12+\n\x0b\x64\x61ta_filler\x18\x01 \x03(\x0b\x32\x16.caffe.FillerParameter\x12\x1f\n\x05shape\x18\x06 \x03(\x0b\x32\x10.caffe.BlobShape\x12\x0b\n\x03num\x18\x02 \x03(\r\x12\x10\n\x08\x63hannels\x18\x03 \x03(\r\x12\x0e\n\x06height\x18\x04 \x03(\r\x12\r\n\x05width\x18\x05 \x03(\r\"\xa5\x01\n\x10\x45ltwiseParameter\x12\x39\n\toperation\x18\x01 \x01(\x0e\x32!.caffe.EltwiseParameter.EltwiseOp:\x03SUM\x12\r\n\x05\x63oeff\x18\x02 \x03(\x02\x12\x1e\n\x10stable_prod_grad\x18\x03 \x01(\x08:\x04true\"\'\n\tEltwiseOp\x12\x08\n\x04PROD\x10\x00\x12\x07\n\x03SUM\x10\x01\x12\x07\n\x03MAX\x10\x02\"D\n\x0c\x45xpParameter\x12\x10\n\x04\x62\x61se\x18\x01 \x01(\x02:\x02-1\x12\x10\n\x05scale\x18\x02 \x01(\x02:\x01\x31\x12\x10\n\x05shift\x18\x03 \x01(\x02:\x01\x30\"D\n\x0cLogParameter\x12\x10\n\x04\x62\x61se\x18\x01 \x01(\x02:\x02-1\x12\x10\n\x05scale\x18\x02 \x01(\x02:\x01\x31\x12\x10\n\x05shift\x18\x03 \x01(\x02:\x01\x30\"7\n\x11HDF5DataParameter\x12\x0e\n\x06source\x18\x01 \x01(\t\x12\x12\n\nbatch_size\x18\x02 \x01(\r\"(\n\x13HDF5OutputParameter\x12\x11\n\tfile_name\x18\x01 \x01(\t\"^\n\x12HingeLossParameter\x12\x30\n\x04norm\x18\x01 \x01(\x0e\x32\x1e.caffe.HingeLossParameter.Norm:\x02L1\"\x16\n\x04Norm\x12\x06\n\x02L1\x10\x01\x12\x06\n\x02L2\x10\x02\"\x94\x02\n\x12ImageDataParameter\x12\x0e\n\x06source\x18\x01 \x01(\t\x12\x12\n\nbatch_size\x18\x04 \x01(\r\x12\x14\n\trand_skip\x18\x07 \x01(\r:\x01\x30\x12\x16\n\x07shuffle\x18\x08 \x01(\x08:\x05\x66\x61lse\x12\x15\n\nnew_height\x18\t \x01(\r:\x01\x30\x12\x14\n\tnew_width\x18\n \x01(\r:\x01\x30\x12\x16\n\x08is_color\x18\x0b \x01(\x08:\x04true\x12\x10\n\x05scale\x18\x02 \x01(\x02:\x01\x31\x12\x11\n\tmean_file\x18\x03 \x01(\t\x12\x14\n\tcrop_size\x18\x05 \x01(\r:\x01\x30\x12\x15\n\x06mirror\x18\x06 \x01(\x08:\x05\x66\x61lse\x12\x15\n\x0broot_folder\x18\x0c \x01(\t:\x00\"\'\n\x15InfogainLossParameter\x12\x0e\n\x06source\x18\x01 \x01(\t\"\xa0\x02\n\x15InnerProductParameter\x12\x12\n\nnum_output\x18\x01 \x01(\r\x12\x17\n\tbias_term\x18\x02 \x01(\x08:\x04true\x12-\n\rweight_filler\x18\x03 \x01(\x0b\x32\x16.caffe.FillerParameter\x12+\n\x0b\x62ias_filler\x18\x04 \x01(\x0b\x32\x16.caffe.FillerParameter\x12\x0f\n\x04\x61xis\x18\x05 \x01(\x05:\x01\x31\x12\x30\n\x0equantize_param\x18\x06 \x01(\x0b\x32\x18.caffe.QuantizeParameter\x12;\n\x19perchannel_quantize_param\x18\x07 \x03(\x0b\x32\x18.caffe.QuantizeParameter\"\xd6\x01\n\x0cLRNParameter\x12\x15\n\nlocal_size\x18\x01 \x01(\r:\x01\x35\x12\x10\n\x05\x61lpha\x18\x02 \x01(\x02:\x01\x31\x12\x12\n\x04\x62\x65ta\x18\x03 \x01(\x02:\x04\x30.75\x12\x44\n\x0bnorm_region\x18\x04 \x01(\x0e\x32\x1e.caffe.LRNParameter.NormRegion:\x0f\x41\x43ROSS_CHANNELS\x12\x0c\n\x01k\x18\x05 \x01(\x02:\x01\x31\"5\n\nNormRegion\x12\x13\n\x0f\x41\x43ROSS_CHANNELS\x10\x00\x12\x12\n\x0eWITHIN_CHANNEL\x10\x01\"Z\n\x13MemoryDataParameter\x12\x12\n\nbatch_size\x18\x01 \x01(\r\x12\x10\n\x08\x63hannels\x18\x02 \x01(\r\x12\x0e\n\x06height\x18\x03 \x01(\r\x12\r\n\x05width\x18\x04 \x01(\r\"P\n\x0cMVNParameter\x12 \n\x12normalize_variance\x18\x01 \x01(\x08:\x04true\x12\x1e\n\x0f\x61\x63ross_channels\x18\x02 \x01(\x08:\x05\x66\x61lse\"!\n\x10PermuteParameter\x12\r\n\x05order\x18\x01 \x03(\r\"\xf9\x03\n\x10PoolingParameter\x12\x35\n\x04pool\x18\x01 \x01(\x0e\x32\".caffe.PoolingParameter.PoolMethod:\x03MAX\x12\x0e\n\x03pad\x18\x04 \x01(\r:\x01\x30\x12\x10\n\x05pad_h\x18\t \x01(\r:\x01\x30\x12\x10\n\x05pad_w\x18\n \x01(\r:\x01\x30\x12\x13\n\x0bkernel_size\x18\x02 \x01(\r\x12\x10\n\x08kernel_h\x18\x05 \x01(\r\x12\x10\n\x08kernel_w\x18\x06 \x01(\r\x12\x11\n\x06stride\x18\x03 \x01(\r:\x01\x31\x12\x10\n\x08stride_h\x18\x07 \x01(\r\x12\x10\n\x08stride_w\x18\x08 \x01(\r\x12\x37\n\x06\x65ngine\x18\x0b \x01(\x0e\x32\x1e.caffe.PoolingParameter.Engine:\x07\x44\x45\x46\x41ULT\x12\x1d\n\x0eglobal_pooling\x18\x0c \x01(\x08:\x05\x66\x61lse\x12\x17\n\tceil_mode\x18\r \x01(\x08:\x04true\"l\n\nPoolMethod\x12\x07\n\x03MAX\x10\x00\x12\x07\n\x03\x41VE\x10\x01\x12\x0e\n\nSTOCHASTIC\x10\x02\x12\x1d\n\x19\x41VE_COUNT_INCLUDE_PADDING\x10\x03\x12\x1d\n\x19\x41VE_COUNT_EXCLUDE_PADDING\x10\x04\"+\n\x06\x45ngine\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\t\n\x05\x43\x41\x46\x46\x45\x10\x01\x12\t\n\x05\x43UDNN\x10\x02\"\xa2\x01\n\x18PoolingSpecificParameter\x12@\n\x04pool\x18\x01 \x02(\x0e\x32\x32.caffe.PoolingSpecificParameter.PoolSpecificMethod\x12\r\n\x05spe_w\x18\x02 \x02(\r\x12\r\n\x05spe_h\x18\x03 \x02(\r\"&\n\x12PoolSpecificMethod\x12\x07\n\x03MAX\x10\x00\x12\x07\n\x03\x41VE\x10\x01\"F\n\x0ePowerParameter\x12\x10\n\x05power\x18\x01 \x01(\x02:\x01\x31\x12\x10\n\x05scale\x18\x02 \x01(\x02:\x01\x31\x12\x10\n\x05shift\x18\x03 \x01(\x02:\x01\x30\"\xb0\x01\n\x11PriorBoxParameter\x12\x10\n\x08min_size\x18\x01 \x01(\x02\x12\x10\n\x08max_size\x18\x02 \x01(\x02\x12\x14\n\x0c\x61spect_ratio\x18\x03 \x03(\x02\x12\x12\n\x04\x66lip\x18\x04 \x01(\x08:\x04true\x12\x12\n\x04\x63lip\x18\x05 \x01(\x08:\x04true\x12\x10\n\x08variance\x18\x06 \x03(\x02\"\'\n\x08\x43odeType\x12\n\n\x06\x43ORNER\x10\x01\x12\x0f\n\x0b\x43\x45NTER_SIZE\x10\x02\"0\n\x0fPythonParameter\x12\x0e\n\x06module\x18\x01 \x01(\t\x12\r\n\x05layer\x18\x02 \x01(\t\"\x8d\x01\n\rReLUParameter\x12\x19\n\x0enegative_slope\x18\x01 \x01(\x02:\x01\x30\x12\x34\n\x06\x65ngine\x18\x02 \x01(\x0e\x32\x1b.caffe.ReLUParameter.Engine:\x07\x44\x45\x46\x41ULT\"+\n\x06\x45ngine\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\t\n\x05\x43\x41\x46\x46\x45\x10\x01\x12\t\n\x05\x43UDNN\x10\x02\"Y\n\x13ROIPoolingParameter\x12\x13\n\x08pooled_h\x18\x01 \x01(\r:\x01\x30\x12\x13\n\x08pooled_w\x18\x02 \x01(\r:\x01\x30\x12\x18\n\rspatial_scale\x18\x03 \x01(\x02:\x01\x31\"Z\n\x10ReshapeParameter\x12\x1f\n\x05shape\x18\x01 \x01(\x0b\x32\x10.caffe.BlobShape\x12\x0f\n\x04\x61xis\x18\x02 \x01(\x05:\x01\x30\x12\x14\n\x08num_axes\x18\x03 \x01(\x05:\x02-1\"x\n\x10SigmoidParameter\x12\x37\n\x06\x65ngine\x18\x01 \x01(\x0e\x32\x1e.caffe.SigmoidParameter.Engine:\x07\x44\x45\x46\x41ULT\"+\n\x06\x45ngine\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\t\n\x05\x43\x41\x46\x46\x45\x10\x01\x12\t\n\x05\x43UDNN\x10\x02\"L\n\x0eSliceParameter\x12\x0f\n\x04\x61xis\x18\x03 \x01(\x05:\x01\x31\x12\x13\n\x0bslice_point\x18\x02 \x03(\r\x12\x14\n\tslice_dim\x18\x01 \x01(\r:\x01\x31\"\x89\x01\n\x10SoftmaxParameter\x12\x37\n\x06\x65ngine\x18\x01 \x01(\x0e\x32\x1e.caffe.SoftmaxParameter.Engine:\x07\x44\x45\x46\x41ULT\x12\x0f\n\x04\x61xis\x18\x02 \x01(\x05:\x01\x31\"+\n\x06\x45ngine\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\t\n\x05\x43\x41\x46\x46\x45\x10\x01\x12\t\n\x05\x43UDNN\x10\x02\"r\n\rTanHParameter\x12\x34\n\x06\x65ngine\x18\x01 \x01(\x0e\x32\x1b.caffe.TanHParameter.Engine:\x07\x44\x45\x46\x41ULT\"+\n\x06\x45ngine\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x00\x12\t\n\x05\x43\x41\x46\x46\x45\x10\x01\x12\t\n\x05\x43UDNN\x10\x02\"*\n\x12ThresholdParameter\x12\x14\n\tthreshold\x18\x01 \x01(\x02:\x01\x30\"\xc1\x02\n\x13WindowDataParameter\x12\x0e\n\x06source\x18\x01 \x01(\t\x12\x10\n\x05scale\x18\x02 \x01(\x02:\x01\x31\x12\x11\n\tmean_file\x18\x03 \x01(\t\x12\x12\n\nbatch_size\x18\x04 \x01(\r\x12\x14\n\tcrop_size\x18\x05 \x01(\r:\x01\x30\x12\x15\n\x06mirror\x18\x06 \x01(\x08:\x05\x66\x61lse\x12\x19\n\x0c\x66g_threshold\x18\x07 \x01(\x02:\x03\x30.5\x12\x19\n\x0c\x62g_threshold\x18\x08 \x01(\x02:\x03\x30.5\x12\x19\n\x0b\x66g_fraction\x18\t \x01(\x02:\x04\x30.25\x12\x16\n\x0b\x63ontext_pad\x18\n \x01(\r:\x01\x30\x12\x17\n\tcrop_mode\x18\x0b \x01(\t:\x04warp\x12\x1b\n\x0c\x63\x61\x63he_images\x18\x0c \x01(\x08:\x05\x66\x61lse\x12\x15\n\x0broot_folder\x18\r \x01(\t:\x00\"\xb5\x01\n\x0b\x42NParameter\x12,\n\x0cscale_filler\x18\x01 \x01(\x0b\x32\x16.caffe.FillerParameter\x12,\n\x0cshift_filler\x18\x02 \x01(\x0b\x32\x16.caffe.FillerParameter\x12\x16\n\x07var_eps\x18\x03 \x01(\x02:\x05\x31\x65-10\x12\x1d\n\x0emoving_average\x18\x04 \x01(\x08:\x05\x66\x61lse\x12\x13\n\x05\x64\x65\x63\x61y\x18\x05 \x01(\x02:\x04\x30.05\"\xa5\x01\n\x0c\x43TCParameter\x12\x16\n\tthreshold\x18\x01 \x01(\x02:\x03\x30.7\x12;\n\x0b\x64\x65\x63ode_type\x18\x02 \x01(\x0e\x32\x1b.caffe.CTCParameter.Decoder:\tbest_path\"@\n\x07\x44\x65\x63oder\x12\r\n\tbest_path\x10\x00\x12\x13\n\x0f\x62\x65st_path_thres\x10\x01\x12\x11\n\rprefix_search\x10\x02\"\x89\x01\n\x0ePReLUParameter\x12&\n\x06\x66iller\x18\x01 \x01(\x0b\x32\x16.caffe.FillerParameter\x12\x1d\n\x0e\x63hannel_shared\x18\x02 \x01(\x08:\x05\x66\x61lse\x12\x30\n\x0equantize_param\x18\x03 \x01(\x0b\x32\x18.caffe.QuantizeParameter\"\xa1\x01\n\rCropParameter\x12/\n\x04type\x18\x01 \x01(\x0e\x32\x19.caffe.CropParameter.Type:\x06\x43\x45NTER\x12\x11\n\x06\x63rop_w\x18\x02 \x01(\r:\x01\x30\x12\x11\n\x06\x63rop_h\x18\x03 \x01(\r:\x01\x30\x12\x19\n\nprint_info\x18\x04 \x01(\x08:\x05\x66\x61lse\"\x1e\n\x04Type\x12\n\n\x06\x43\x45NTER\x10\x01\x12\n\n\x06RANDOM\x10\x02\"\xaf\x01\n\x14\x41\x66\x66ineTransParameter\x12\x13\n\x08output_h\x18\x01 \x01(\r:\x01\x30\x12\x13\n\x08output_w\x18\x02 \x01(\r:\x01\x30\x12\x17\n\x0c\x62order_value\x18\x03 \x01(\x02:\x01\x30\x12\x12\n\naffine_mat\x18\x04 \x03(\x02\x12\x10\n\x05scale\x18\x05 \x01(\x02:\x01\x31\x12\x16\n\x0btranslate_x\x18\x06 \x01(\x02:\x01\x30\x12\x16\n\x0btranslate_y\x18\x07 \x01(\x02:\x01\x30\"\x89\x01\n\x19\x41\x66\x66ineTransPointParameter\x12\x16\n\x07inverse\x18\x01 \x01(\x08:\x05\x66\x61lse\x12\x16\n\x0btranslate_x\x18\x02 \x01(\x02:\x01\x30\x12\x16\n\x0btranslate_y\x18\x03 \x01(\x02:\x01\x30\x12\x12\n\naffine_mat\x18\x04 \x03(\x02\x12\x10\n\x05scale\x18\x05 \x01(\x02:\x01\x31\"@\n\x16\x43\x61lcAffineMatParameter\x12\x12\n\nlandmark_x\x18\x01 \x03(\x02\x12\x12\n\nlandmark_y\x18\x02 \x03(\x02\"\x84\x01\n\x0cROIParameter\x12\x0e\n\x06\x63rop_h\x18\x01 \x02(\r\x12\x0e\n\x06\x63rop_w\x18\x02 \x02(\r\x12\x10\n\x08\x63\x65nter_x\x18\x03 \x03(\x02\x12\x10\n\x08\x63\x65nter_y\x18\x04 \x03(\x02\x12\x30\n\x04type\x18\x05 \x01(\x0e\x32\x11.caffe.InterpType:\x0fINTERP_BILINEAR\"\xad\x01\n\x0fInterpParameter\x12\x11\n\x06height\x18\x01 \x01(\x05:\x01\x30\x12\x10\n\x05width\x18\x02 \x01(\x05:\x01\x30\x12\x16\n\x0bzoom_factor\x18\x03 \x01(\x05:\x01\x31\x12\x18\n\rshrink_factor\x18\x04 \x01(\x05:\x01\x31\x12\x12\n\x07pad_beg\x18\x05 \x01(\x05:\x01\x30\x12\x12\n\x07pad_end\x18\x06 \x01(\x05:\x01\x30\x12\x1b\n\ralign_corners\x18\x07 \x01(\x08:\x04true\"\xa2\x01\n\x12RecurrentParameter\x12\x15\n\nnum_output\x18\x01 \x01(\r:\x01\x30\x12-\n\rweight_filler\x18\x02 \x01(\x0b\x32\x16.caffe.FillerParameter\x12+\n\x0b\x62ias_filler\x18\x03 \x01(\x0b\x32\x16.caffe.FillerParameter\x12\x19\n\ndebug_info\x18\x04 \x01(\x08:\x05\x66\x61lse\"9\n\x10\x46lattenParameter\x12\x0f\n\x04\x61xis\x18\x01 \x01(\x05:\x01\x31\x12\x14\n\x08\x65nd_axis\x18\x02 \x01(\x05:\x02-1\"j\n\x12\x42\x61tchNormParameter\x12\x18\n\x10use_global_stats\x18\x01 \x01(\x08\x12&\n\x17moving_average_fraction\x18\x02 \x01(\x02:\x05\x30.999\x12\x12\n\x03\x65ps\x18\x03 \x01(\x02:\x05\x31\x65-05\"\xe0\x13\n\x10V1LayerParameter\x12\x0e\n\x06\x62ottom\x18\x02 \x03(\t\x12\x0b\n\x03top\x18\x03 \x03(\t\x12\x0c\n\x04name\x18\x04 \x01(\t\x12$\n\x07include\x18  \x03(\x0b\x32\x13.caffe.NetStateRule\x12$\n\x07\x65xclude\x18! \x03(\x0b\x32\x13.caffe.NetStateRule\x12/\n\x04type\x18\x05 \x01(\x0e\x32!.caffe.V1LayerParameter.LayerType\x12\x1f\n\x05\x62lobs\x18\x06 \x03(\x0b\x32\x10.caffe.BlobProto\x12\x0e\n\x05param\x18\xe9\x07 \x03(\t\x12>\n\x0f\x62lob_share_mode\x18\xea\x07 \x03(\x0e\x32$.caffe.V1LayerParameter.DimCheckMode\x12\x10\n\x08\x62lobs_lr\x18\x07 \x03(\x02\x12\x14\n\x0cweight_decay\x18\x08 \x03(\x02\x12\x13\n\x0bloss_weight\x18# \x03(\x02\x12\x30\n\x0e\x61\x63\x63uracy_param\x18\x1b \x01(\x0b\x32\x18.caffe.AccuracyParameter\x12,\n\x0c\x61rgmax_param\x18\x17 \x01(\x0b\x32\x16.caffe.ArgMaxParameter\x12,\n\x0c\x63oncat_param\x18\t \x01(\x0b\x32\x16.caffe.ConcatParameter\x12?\n\x16\x63ontrastive_loss_param\x18( \x01(\x0b\x32\x1f.caffe.ContrastiveLossParameter\x12\x36\n\x11\x63onvolution_param\x18\n \x01(\x0b\x32\x1b.caffe.ConvolutionParameter\x12(\n\ndata_param\x18\x0b \x01(\x0b\x32\x14.caffe.DataParameter\x12.\n\rdropout_param\x18\x0c \x01(\x0b\x32\x17.caffe.DropoutParameter\x12\x33\n\x10\x64ummy_data_param\x18\x1a \x01(\x0b\x32\x19.caffe.DummyDataParameter\x12.\n\reltwise_param\x18\x18 \x01(\x0b\x32\x17.caffe.EltwiseParameter\x12&\n\texp_param\x18) \x01(\x0b\x32\x13.caffe.ExpParameter\x12\x31\n\x0fhdf5_data_param\x18\r \x01(\x0b\x32\x18.caffe.HDF5DataParameter\x12\x35\n\x11hdf5_output_param\x18\x0e \x01(\x0b\x32\x1a.caffe.HDF5OutputParameter\x12\x33\n\x10hinge_loss_param\x18\x1d \x01(\x0b\x32\x19.caffe.HingeLossParameter\x12\x33\n\x10image_data_param\x18\x0f \x01(\x0b\x32\x19.caffe.ImageDataParameter\x12\x39\n\x13infogain_loss_param\x18\x10 \x01(\x0b\x32\x1c.caffe.InfogainLossParameter\x12\x39\n\x13inner_product_param\x18\x11 \x01(\x0b\x32\x1c.caffe.InnerProductParameter\x12&\n\tlrn_param\x18\x12 \x01(\x0b\x32\x13.caffe.LRNParameter\x12\x35\n\x11memory_data_param\x18\x16 \x01(\x0b\x32\x1a.caffe.MemoryDataParameter\x12&\n\tmvn_param\x18\" \x01(\x0b\x32\x13.caffe.MVNParameter\x12.\n\rpooling_param\x18\x13 \x01(\x0b\x32\x17.caffe.PoolingParameter\x12*\n\x0bpower_param\x18\x15 \x01(\x0b\x32\x15.caffe.PowerParameter\x12(\n\nrelu_param\x18\x1e \x01(\x0b\x32\x14.caffe.ReLUParameter\x12.\n\rsigmoid_param\x18& \x01(\x0b\x32\x17.caffe.SigmoidParameter\x12.\n\rsoftmax_param\x18\' \x01(\x0b\x32\x17.caffe.SoftmaxParameter\x12*\n\x0bslice_param\x18\x1f \x01(\x0b\x32\x15.caffe.SliceParameter\x12(\n\ntanh_param\x18% \x01(\x0b\x32\x14.caffe.TanHParameter\x12\x32\n\x0fthreshold_param\x18\x19 \x01(\x0b\x32\x19.caffe.ThresholdParameter\x12\x35\n\x11window_data_param\x18\x14 \x01(\x0b\x32\x1a.caffe.WindowDataParameter\x12\x37\n\x0ftransform_param\x18$ \x01(\x0b\x32\x1e.caffe.TransformationParameter\x12(\n\nloss_param\x18* \x01(\x0b\x32\x14.caffe.LossParameter\x12&\n\x05layer\x18\x01 \x01(\x0b\x32\x17.caffe.V0LayerParameter\"\xd8\x04\n\tLayerType\x12\x08\n\x04NONE\x10\x00\x12\n\n\x06\x41\x42SVAL\x10#\x12\x0c\n\x08\x41\x43\x43URACY\x10\x01\x12\n\n\x06\x41RGMAX\x10\x1e\x12\x08\n\x04\x42NLL\x10\x02\x12\n\n\x06\x43ONCAT\x10\x03\x12\x14\n\x10\x43ONTRASTIVE_LOSS\x10%\x12\x0f\n\x0b\x43ONVOLUTION\x10\x04\x12\x08\n\x04\x44\x41TA\x10\x05\x12\x11\n\rDECONVOLUTION\x10\'\x12\x0b\n\x07\x44ROPOUT\x10\x06\x12\x0e\n\nDUMMY_DATA\x10 \x12\x12\n\x0e\x45UCLIDEAN_LOSS\x10\x07\x12\x0b\n\x07\x45LTWISE\x10\x19\x12\x07\n\x03\x45XP\x10&\x12\x0b\n\x07\x46LATTEN\x10\x08\x12\r\n\tHDF5_DATA\x10\t\x12\x0f\n\x0bHDF5_OUTPUT\x10\n\x12\x0e\n\nHINGE_LOSS\x10\x1c\x12\n\n\x06IM2COL\x10\x0b\x12\x0e\n\nIMAGE_DATA\x10\x0c\x12\x11\n\rINFOGAIN_LOSS\x10\r\x12\x11\n\rINNER_PRODUCT\x10\x0e\x12\x07\n\x03LRN\x10\x0f\x12\x0f\n\x0bMEMORY_DATA\x10\x1d\x12\x1d\n\x19MULTINOMIAL_LOGISTIC_LOSS\x10\x10\x12\x07\n\x03MVN\x10\"\x12\x0b\n\x07POOLING\x10\x11\x12\t\n\x05POWER\x10\x1a\x12\x08\n\x04RELU\x10\x12\x12\x0b\n\x07SIGMOID\x10\x13\x12\x1e\n\x1aSIGMOID_CROSS_ENTROPY_LOSS\x10\x1b\x12\x0b\n\x07SILENCE\x10$\x12\x0b\n\x07SOFTMAX\x10\x14\x12\x10\n\x0cSOFTMAX_LOSS\x10\x15\x12\t\n\x05SPLIT\x10\x16\x12\t\n\x05SLICE\x10!\x12\x08\n\x04TANH\x10\x17\x12\x0f\n\x0bWINDOW_DATA\x10\x18\x12\r\n\tTHRESHOLD\x10\x1f\"*\n\x0c\x44imCheckMode\x12\n\n\x06STRICT\x10\x00\x12\x0e\n\nPERMISSIVE\x10\x01\"\xae\x08\n\x10V0LayerParameter\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x12\n\nnum_output\x18\x03 \x01(\r\x12\x16\n\x08\x62iasterm\x18\x04 \x01(\x08:\x04true\x12-\n\rweight_filler\x18\x05 \x01(\x0b\x32\x16.caffe.FillerParameter\x12+\n\x0b\x62ias_filler\x18\x06 \x01(\x0b\x32\x16.caffe.FillerParameter\x12\x0e\n\x03pad\x18\x07 \x01(\r:\x01\x30\x12\x12\n\nkernelsize\x18\x08 \x01(\r\x12\x10\n\x05group\x18\t \x01(\r:\x01\x31\x12\x11\n\x06stride\x18\n \x01(\r:\x01\x31\x12\x35\n\x04pool\x18\x0b \x01(\x0e\x32\".caffe.V0LayerParameter.PoolMethod:\x03MAX\x12\x1a\n\rdropout_ratio\x18\x0c \x01(\x02:\x03\x30.5\x12\x15\n\nlocal_size\x18\r \x01(\r:\x01\x35\x12\x10\n\x05\x61lpha\x18\x0e \x01(\x02:\x01\x31\x12\x12\n\x04\x62\x65ta\x18\x0f \x01(\x02:\x04\x30.75\x12\x0c\n\x01k\x18\x16 \x01(\x02:\x01\x31\x12\x0e\n\x06source\x18\x10 \x01(\t\x12\x10\n\x05scale\x18\x11 \x01(\x02:\x01\x31\x12\x10\n\x08meanfile\x18\x12 \x01(\t\x12\x11\n\tbatchsize\x18\x13 \x01(\r\x12\x13\n\x08\x63ropsize\x18\x14 \x01(\r:\x01\x30\x12\x15\n\x06mirror\x18\x15 \x01(\x08:\x05\x66\x61lse\x12\x1f\n\x05\x62lobs\x18\x32 \x03(\x0b\x32\x10.caffe.BlobProto\x12\x10\n\x08\x62lobs_lr\x18\x33 \x03(\x02\x12\x14\n\x0cweight_decay\x18\x34 \x03(\x02\x12\x14\n\trand_skip\x18\x35 \x01(\r:\x01\x30\x12\x1d\n\x10\x64\x65t_fg_threshold\x18\x36 \x01(\x02:\x03\x30.5\x12\x1d\n\x10\x64\x65t_bg_threshold\x18\x37 \x01(\x02:\x03\x30.5\x12\x1d\n\x0f\x64\x65t_fg_fraction\x18\x38 \x01(\x02:\x04\x30.25\x12\x1a\n\x0f\x64\x65t_context_pad\x18: \x01(\r:\x01\x30\x12\x1b\n\rdet_crop_mode\x18; \x01(\t:\x04warp\x12\x12\n\x07new_num\x18< \x01(\x05:\x01\x30\x12\x17\n\x0cnew_channels\x18= \x01(\x05:\x01\x30\x12\x15\n\nnew_height\x18> \x01(\x05:\x01\x30\x12\x14\n\tnew_width\x18? \x01(\x05:\x01\x30\x12\x1d\n\x0eshuffle_images\x18@ \x01(\x08:\x05\x66\x61lse\x12\x15\n\nconcat_dim\x18\x41 \x01(\r:\x01\x31\x12\x16\n\x0bntile_width\x18H \x01(\x05:\x01\x31\x12\x17\n\x0cntile_height\x18I \x01(\x05:\x01\x31\x12\x36\n\x11hdf5_output_param\x18\xe9\x07 \x01(\x0b\x32\x1a.caffe.HDF5OutputParameter\".\n\nPoolMethod\x12\x07\n\x03MAX\x10\x00\x12\x07\n\x03\x41VE\x10\x01\x12\x0e\n\nSTOCHASTIC\x10\x02\"+\n\x17\x43hannelShuffleParameter\x12\x10\n\x05group\x18\x01 \x02(\r:\x01\x31\"&\n\x12MovingAvgParameter\x12\x10\n\x05\x64\x65\x63\x61y\x18\x01 \x02(\x02:\x01\x30\"\xac\x01\n\x10SlicingParameter\x12\x1a\n\x0f\x63oefficient_len\x18\x01 \x01(\r:\x01\x31\x12\x13\n\x07\x64\x65pth_d\x18\x02 \x01(\x05:\x02-1\x12\x13\n\x07scale_x\x18\x03 \x01(\x02:\x02-1\x12\x13\n\x07scale_y\x18\x04 \x01(\x02:\x02-1\x12\x13\n\x08offset_x\x18\x05 \x01(\x05:\x01\x30\x12\x13\n\x08offset_y\x18\x06 \x01(\x05:\x01\x30\x12\x13\n\x08offset_z\x18\x07 \x01(\x05:\x01\x30\"7\n\x19\x42ilateralSlicingParameter\x12\x1a\n\x0f\x63oefficient_len\x18\x01 \x02(\r:\x01\x31\"T\n\x1c\x42ilateralSliceApplyParameter\x12\x1a\n\x0f\x63oefficient_len\x18\x01 \x02(\r:\x01\x32\x12\x18\n\nhas_offset\x18\x02 \x01(\x08:\x04true\"p\n$AdaptiveBilateralSliceApplyParameter\x12\x1a\n\x0f\x63oefficient_len\x18\x01 \x01(\r:\x01\x32\x12\x18\n\nhas_offset\x18\x02 \x01(\x08:\x04true\x12\x12\n\x07\x64\x65pth_d\x18\x03 \x01(\x05:\x01\x31\"\x88\x01\n\x14PointsCurveParameter\x12\x15\n\tnumpoints\x18\x01 \x01(\r:\x02\x31\x36\x12,\n\x0cslope_filler\x18\x02 \x01(\x0b\x32\x16.caffe.FillerParameter\x12+\n\x0b\x62ias_filler\x18\x03 \x01(\x0b\x32\x16.caffe.FillerParameter\".\n\x15SubpixelDownParameter\x12\x15\n\ndownsample\x18\x01 \x01(\r:\x01\x31\"\xf5\x01\n\x13SubpixelUpParameter\x12\x13\n\x08upsample\x18\x01 \x01(\r:\x01\x31\x12\x45\n\x0foutput_datatype\x18\x02 \x01(\x0e\x32#.caffe.SubpixelUpParameter.DataType:\x07\x46LOAT32\x12:\n\x04mode\x18\x03 \x01(\x0e\x32#.caffe.SubpixelUpParameter.ModeType:\x07pytorch\"\"\n\x08\x44\x61taType\x12\x0b\n\x07\x46LOAT32\x10\x00\x12\t\n\x05UINT8\x10\x01\"\"\n\x08ModeType\x12\t\n\x05\x63\x61\x66\x66\x65\x10\x00\x12\x0b\n\x07pytorch\x10\x01\")\n\rClipParameter\x12\x0b\n\x03min\x18\x01 \x01(\x02\x12\x0b\n\x03max\x18\x02 \x01(\x02\"!\n\x1f\x45ltwiseAffineTransformParameter\"2\n\rTileParameter\x12\x0f\n\x04\x61xis\x18\x01 \x01(\x05:\x01\x31\x12\x10\n\x05tiles\x18\x02 \x02(\x05:\x01\x31\"I\n\x0fReduceParameter\x12\x0f\n\x04\x61xis\x18\x01 \x01(\x05:\x01\x31\x12\x0f\n\x04mode\x18\x02 \x01(\x05:\x01\x30\"\x14\n\x08ReduceOp\x12\x08\n\x04MEAN\x10\x00\"\x94\x01\n\x0cPadParameter\x12\x0e\n\x03pad\x18\x01 \x01(\x05:\x01\x30\x12\x35\n\x04mode\x18\x02 \x01(\x0e\x32\x1b.caffe.PadParameter.PadMode:\nREFLECTION\x12\x10\n\x05pad_w\x18\x03 \x01(\x05:\x01\x30\x12\x10\n\x05pad_h\x18\x04 \x01(\x05:\x01\x30\"\x19\n\x07PadMode\x12\x0e\n\nREFLECTION\x10\x01\"+\n\x0eReLU6Parameter\x12\x19\n\x0enegative_slope\x18\x01 \x01(\x02:\x01\x30\"+\n\x16\x43orrelation2DParameter\x12\x11\n\x06groups\x18\x01 \x01(\r:\x01\x31\"t\n\x0fMatMulParameter\x12\x33\n\x04mode\x18\x01 \x01(\x0e\x32!.caffe.MatMulParameter.MatMulMode:\x02NN\",\n\nMatMulMode\x12\x06\n\x02NN\x10\x01\x12\x06\n\x02NT\x10\x02\x12\x06\n\x02TN\x10\x03\x12\x06\n\x02TT\x10\x04\"\x80\x01\n\x12ParameterParameter\x12\x10\n\x05\x62\x61tch\x18\x01 \x01(\x05:\x01\x31\x12\r\n\x01m\x18\x02 \x01(\x05:\x02-1\x12\r\n\x01n\x18\x03 \x01(\x05:\x02-1\x12\x13\n\x07\x63hannel\x18\x04 \x01(\x05:\x02-1\x12\x12\n\x06height\x18\x05 \x01(\x05:\x02-1\x12\x11\n\x05width\x18\x06 \x01(\x05:\x02-1\"2\n\x15PixelShuffleParameter\x12\x19\n\x0eupscale_factor\x18\x01 \x02(\x05:\x01\x31\"C\n\x15ROITransformParameter\x12\x14\n\x0croi_pooled_h\x18\x01 \x02(\x05\x12\x14\n\x0croi_pooled_w\x18\x02 \x02(\x05\"D\n\x16MainTransformParameter\x12\x14\n\x0croi_pooled_h\x18\x01 \x02(\x05\x12\x14\n\x0croi_pooled_w\x18\x02 \x02(\x05\"X\n\x15InstanceNormParameter\x12\x14\n\x0cnum_features\x18\x01 \x01(\x05\x12\x12\n\x03\x65ps\x18\x02 \x01(\x02:\x05\x31\x65-05\x12\x15\n\x06\x61\x66\x66ine\x18\x03 \x01(\x08:\x05\x66\x61lse\"\x80\x01\n\x13GridSampleParameter\x12\x43\n\x0cpadding_mode\x18\x01 \x01(\x0e\x32&.caffe.GridSampleParameter.PaddingMode:\x05ZEROS\"$\n\x0bPaddingMode\x12\t\n\x05ZEROS\x10\x00\x12\n\n\x06\x42ORDER\x10\x01\"\xba\x01\n\x15GridSample3DParameter\x12\x45\n\x0cpadding_mode\x18\x01 \x01(\x0e\x32(.caffe.GridSample3DParameter.PaddingMode:\x05ZEROS\x12\x1c\n\ralign_corners\x18\x02 \x01(\x08:\x05\x66\x61lse\x12\x16\n\x0eoutput_channel\x18\x03 \x02(\x05\"$\n\x0bPaddingMode\x12\t\n\x05ZEROS\x10\x00\x12\n\n\x06\x42ORDER\x10\x01\"9\n\x11ReduceL2Parameter\x12\x0f\n\x04\x61xes\x18\x01 \x01(\x05:\x01\x34\x12\x13\n\x08keepdims\x18\x02 \x01(\x05:\x01\x31\"7\n\x11VarianceParameter\x12\x0b\n\x03\x64im\x18\x01 \x03(\x05\x12\x15\n\x07keepdim\x18\x02 \x01(\x08:\x04true\"3\n\rMeanParameter\x12\x0b\n\x03\x64im\x18\x01 \x03(\x05\x12\x15\n\x07keepdim\x18\x02 \x01(\x08:\x04true\"\xc1\x01\n\x11Interp3dParameter\x12\x10\n\x05\x64\x65pth\x18\x08 \x01(\x05:\x01\x30\x12\x11\n\x06height\x18\x01 \x01(\x05:\x01\x30\x12\x10\n\x05width\x18\x02 \x01(\x05:\x01\x30\x12\x16\n\x0bzoom_factor\x18\x03 \x01(\x05:\x01\x31\x12\x18\n\rshrink_factor\x18\x04 \x01(\x05:\x01\x31\x12\x12\n\x07pad_beg\x18\x05 \x01(\x05:\x01\x30\x12\x12\n\x07pad_end\x18\x06 \x01(\x05:\x01\x30\x12\x1b\n\ralign_corners\x18\x07 \x01(\x08:\x04true\"A\n\x0cGRUParameter\x12\x13\n\x0bhidden_size\x18\x01 \x02(\x05\x12\x1c\n\rbidirectional\x18\x02 \x01(\x08:\x05\x66\x61lse\"0\n\x17\x43orrelationMigParameter\x12\x15\n\nshift_step\x18\x01 \x01(\x05:\x01\x39\";\n\x10\x41rgSortParameter\x12\x19\n\ndescending\x18\x01 \x01(\x08:\x05\x66\x61lse\x12\x0c\n\x04\x61xis\x18\x02 \x01(\x05\" \n\x10\x43umProdParameter\x12\x0c\n\x04\x61xis\x18\x01 \x01(\x05*\x1c\n\x05Phase\x12\t\n\x05TRAIN\x10\x00\x12\x08\n\x04TEST\x10\x01*5\n\nInterpType\x12\x13\n\x0fINTERP_BILINEAR\x10\x00\x12\x12\n\x0eINTERP_NEAREST\x10\x01'
)

_PHASE = _descriptor.EnumDescriptor(
  name='Phase',
  full_name='caffe.Phase',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TRAIN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=30654,
  serialized_end=30682,
)
_sym_db.RegisterEnumDescriptor(_PHASE)

Phase = enum_type_wrapper.EnumTypeWrapper(_PHASE)
_INTERPTYPE = _descriptor.EnumDescriptor(
  name='InterpType',
  full_name='caffe.InterpType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INTERP_BILINEAR', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INTERP_NEAREST', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=30684,
  serialized_end=30737,
)
_sym_db.RegisterEnumDescriptor(_INTERPTYPE)

InterpType = enum_type_wrapper.EnumTypeWrapper(_INTERPTYPE)
TRAIN = 0
TEST = 1
INTERP_BILINEAR = 0
INTERP_NEAREST = 1


_SOLVERPARAMETER_SOLVERMODE = _descriptor.EnumDescriptor(
  name='SolverMode',
  full_name='caffe.SolverParameter.SolverMode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CPU', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GPU', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1993,
  serialized_end=2023,
)
_sym_db.RegisterEnumDescriptor(_SOLVERPARAMETER_SOLVERMODE)

_SOLVERPARAMETER_SOLVERTYPE = _descriptor.EnumDescriptor(
  name='SolverType',
  full_name='caffe.SolverParameter.SolverType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SGD', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NESTEROV', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ADAGRAD', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2025,
  serialized_end=2073,
)
_sym_db.RegisterEnumDescriptor(_SOLVERPARAMETER_SOLVERTYPE)

_PARAMSPEC_DIMCHECKMODE = _descriptor.EnumDescriptor(
  name='DimCheckMode',
  full_name='caffe.ParamSpec.DimCheckMode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STRICT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PERMISSIVE', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2504,
  serialized_end=2546,
)
_sym_db.RegisterEnumDescriptor(_PARAMSPEC_DIMCHECKMODE)

_LAYERPARAMETER_PRECISIONTYPE = _descriptor.EnumDescriptor(
  name='PrecisionType',
  full_name='caffe.LayerParameter.PrecisionType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEFAULT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INT8', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT16', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT32', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INT4B', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UINT8', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UINT16', index=6, number=6,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=9631,
  serialized_end=9729,
)
_sym_db.RegisterEnumDescriptor(_LAYERPARAMETER_PRECISIONTYPE)

_POOLING3DPARAMETER_POOLMETHOD = _descriptor.EnumDescriptor(
  name='PoolMethod',
  full_name='caffe.Pooling3dParameter.PoolMethod',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MAX', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AVE', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=11562,
  serialized_end=11592,
)
_sym_db.RegisterEnumDescriptor(_POOLING3DPARAMETER_POOLMETHOD)

_NORMALIZEPARAMETER_NORMALIZEMETHOD = _descriptor.EnumDescriptor(
  name='NormalizeMethod',
  full_name='caffe.NormalizeParameter.NormalizeMethod',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='L1', index=0, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='L2', index=1, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=11696,
  serialized_end=11729,
)
_sym_db.RegisterEnumDescriptor(_NORMALIZEPARAMETER_NORMALIZEMETHOD)

_UNPOOLINGPARAMETER_UNPOOLMETHOD = _descriptor.EnumDescriptor(
  name='UnpoolMethod',
  full_name='caffe.UnpoolingParameter.UnpoolMethod',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MAX', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AVE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TILE', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=12768,
  serialized_end=12810,
)
_sym_db.RegisterEnumDescriptor(_UNPOOLINGPARAMETER_UNPOOLMETHOD)

_UNPOOLINGPARAMETER_ENGINE = _descriptor.EnumDescriptor(
  name='Engine',
  full_name='caffe.UnpoolingParameter.Engine',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEFAULT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CAFFE', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=12812,
  serialized_end=12844,
)
_sym_db.RegisterEnumDescriptor(_UNPOOLINGPARAMETER_ENGINE)

_BTCOSTVOLUMEPARAMETER_COSTDOMAINTYPE = _descriptor.EnumDescriptor(
  name='CostDomainType',
  full_name='caffe.BTCostVolumeParameter.CostDomainType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SOBEL_ONLY', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IMAGE_ONLY', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SOBEL_AND_IMAGE', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=13657,
  serialized_end=13726,
)
_sym_db.RegisterEnumDescriptor(_BTCOSTVOLUMEPARAMETER_COSTDOMAINTYPE)

_BTCOSTVOLUMEPARAMETER_IMPLEMENTMODE = _descriptor.EnumDescriptor(
  name='ImplementMode',
  full_name='caffe.BTCostVolumeParameter.ImplementMode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GPU_MODE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CPU_MODE', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=13728,
  serialized_end=13771,
)
_sym_db.RegisterEnumDescriptor(_BTCOSTVOLUMEPARAMETER_IMPLEMENTMODE)

_CORRELATIONPARAMETER_CORRELATIONTYPE = _descriptor.EnumDescriptor(
  name='CorrelationType',
  full_name='caffe.CorrelationParameter.CorrelationType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MULTIPLY', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUBTRACT', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=14078,
  serialized_end=14123,
)
_sym_db.RegisterEnumDescriptor(_CORRELATIONPARAMETER_CORRELATIONTYPE)

_ZXYBNPARAMETER_ENGINE = _descriptor.EnumDescriptor(
  name='Engine',
  full_name='caffe.ZXYBNParameter.Engine',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEFAULT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CAFFE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUDNN', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15722,
  serialized_end=15765,
)
_sym_db.RegisterEnumDescriptor(_ZXYBNPARAMETER_ENGINE)

_CONVOLUTIONPARAMETER_ENGINE = _descriptor.EnumDescriptor(
  name='Engine',
  full_name='caffe.ConvolutionParameter.Engine',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEFAULT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CAFFE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUDNN', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15722,
  serialized_end=15765,
)
_sym_db.RegisterEnumDescriptor(_CONVOLUTIONPARAMETER_ENGINE)

_CONVOLUTIONPARAMETER_PPLCONVOLUTIONFORWARDALGO_T = _descriptor.EnumDescriptor(
  name='PPLConvolutionForwardAlgo_t',
  full_name='caffe.ConvolutionParameter.PPLConvolutionForwardAlgo_t',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_NONE', index=0, number=-1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_IMPLICIT_GEMM', index=1, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_IMPLICIT_PRECOMP_GEMM', index=2, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_GEMM', index=3, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_DIRECT', index=4, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_FFT', index=5, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_FFT_TILING', index=6, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_WINOGRAD_BLK2X2', index=7, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_WINOGRAD_BLK4X4', index=8, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_WINOGRAD_BLK6X6', index=9, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_WINOGRAD_BLK4x4_FLT3x3', index=10, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_WINOGRAD_BLK4x4_FLT5X5', index=11, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PPL_CONVOLUTION_FORWARD_ALGO_INVGEMM', index=12, number=11,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=17051,
  serialized_end=17701,
)
_sym_db.RegisterEnumDescriptor(_CONVOLUTIONPARAMETER_PPLCONVOLUTIONFORWARDALGO_T)

_DATAPARAMETER_DB = _descriptor.EnumDescriptor(
  name='DB',
  full_name='caffe.DataParameter.DB',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LEVELDB', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LMDB', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=17948,
  serialized_end=17975,
)
_sym_db.RegisterEnumDescriptor(_DATAPARAMETER_DB)

_ELTWISEPARAMETER_ELTWISEOP = _descriptor.EnumDescriptor(
  name='EltwiseOp',
  full_name='caffe.EltwiseParameter.EltwiseOp',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PROD', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUM', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MAX', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=19004,
  serialized_end=19043,
)
_sym_db.RegisterEnumDescriptor(_ELTWISEPARAMETER_ELTWISEOP)

_HINGELOSSPARAMETER_NORM = _descriptor.EnumDescriptor(
  name='Norm',
  full_name='caffe.HingeLossParameter.Norm',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='L1', index=0, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='L2', index=1, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=19356,
  serialized_end=19378,
)
_sym_db.RegisterEnumDescriptor(_HINGELOSSPARAMETER_NORM)

_LRNPARAMETER_NORMREGION = _descriptor.EnumDescriptor(
  name='NormRegion',
  full_name='caffe.LRNParameter.NormRegion',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ACROSS_CHANNELS', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WITHIN_CHANNEL', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=20153,
  serialized_end=20206,
)
_sym_db.RegisterEnumDescriptor(_LRNPARAMETER_NORMREGION)

_POOLINGPARAMETER_POOLMETHOD = _descriptor.EnumDescriptor(
  name='PoolMethod',
  full_name='caffe.PoolingParameter.PoolMethod',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MAX', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AVE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STOCHASTIC', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AVE_COUNT_INCLUDE_PADDING', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AVE_COUNT_EXCLUDE_PADDING', index=4, number=4,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=20770,
  serialized_end=20878,
)
_sym_db.RegisterEnumDescriptor(_POOLINGPARAMETER_POOLMETHOD)

_POOLINGPARAMETER_ENGINE = _descriptor.EnumDescriptor(
  name='Engine',
  full_name='caffe.PoolingParameter.Engine',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEFAULT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CAFFE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUDNN', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15722,
  serialized_end=15765,
)
_sym_db.RegisterEnumDescriptor(_POOLINGPARAMETER_ENGINE)

_POOLINGSPECIFICPARAMETER_POOLSPECIFICMETHOD = _descriptor.EnumDescriptor(
  name='PoolSpecificMethod',
  full_name='caffe.PoolingSpecificParameter.PoolSpecificMethod',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MAX', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AVE', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=21050,
  serialized_end=21088,
)
_sym_db.RegisterEnumDescriptor(_POOLINGSPECIFICPARAMETER_POOLSPECIFICMETHOD)

_PRIORBOXPARAMETER_CODETYPE = _descriptor.EnumDescriptor(
  name='CodeType',
  full_name='caffe.PriorBoxParameter.CodeType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CORNER', index=0, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CENTER_SIZE', index=1, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=21300,
  serialized_end=21339,
)
_sym_db.RegisterEnumDescriptor(_PRIORBOXPARAMETER_CODETYPE)

_RELUPARAMETER_ENGINE = _descriptor.EnumDescriptor(
  name='Engine',
  full_name='caffe.ReLUParameter.Engine',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEFAULT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CAFFE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUDNN', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15722,
  serialized_end=15765,
)
_sym_db.RegisterEnumDescriptor(_RELUPARAMETER_ENGINE)

_SIGMOIDPARAMETER_ENGINE = _descriptor.EnumDescriptor(
  name='Engine',
  full_name='caffe.SigmoidParameter.Engine',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEFAULT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CAFFE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUDNN', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15722,
  serialized_end=15765,
)
_sym_db.RegisterEnumDescriptor(_SIGMOIDPARAMETER_ENGINE)

_SOFTMAXPARAMETER_ENGINE = _descriptor.EnumDescriptor(
  name='Engine',
  full_name='caffe.SoftmaxParameter.Engine',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEFAULT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CAFFE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUDNN', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15722,
  serialized_end=15765,
)
_sym_db.RegisterEnumDescriptor(_SOFTMAXPARAMETER_ENGINE)

_TANHPARAMETER_ENGINE = _descriptor.EnumDescriptor(
  name='Engine',
  full_name='caffe.TanHParameter.Engine',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEFAULT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CAFFE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUDNN', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=15722,
  serialized_end=15765,
)
_sym_db.RegisterEnumDescriptor(_TANHPARAMETER_ENGINE)

_CTCPARAMETER_DECODER = _descriptor.EnumDescriptor(
  name='Decoder',
  full_name='caffe.CTCParameter.Decoder',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='best_path', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='best_path_thres', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='prefix_search', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=22828,
  serialized_end=22892,
)
_sym_db.RegisterEnumDescriptor(_CTCPARAMETER_DECODER)

_CROPPARAMETER_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='caffe.CropParameter.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CENTER', index=0, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RANDOM', index=1, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=23166,
  serialized_end=23196,
)
_sym_db.RegisterEnumDescriptor(_CROPPARAMETER_TYPE)

_V1LAYERPARAMETER_LAYERTYPE = _descriptor.EnumDescriptor(
  name='LayerType',
  full_name='caffe.V1LayerParameter.LayerType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ABSVAL', index=1, number=35,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ACCURACY', index=2, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMAX', index=3, number=30,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BNLL', index=4, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONCAT', index=5, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONTRASTIVE_LOSS', index=6, number=37,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONVOLUTION', index=7, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATA', index=8, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DECONVOLUTION', index=9, number=39,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DROPOUT', index=10, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DUMMY_DATA', index=11, number=32,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EUCLIDEAN_LOSS', index=12, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ELTWISE', index=13, number=25,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXP', index=14, number=38,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLATTEN', index=15, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HDF5_DATA', index=16, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HDF5_OUTPUT', index=17, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HINGE_LOSS', index=18, number=28,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IM2COL', index=19, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IMAGE_DATA', index=20, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INFOGAIN_LOSS', index=21, number=13,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INNER_PRODUCT', index=22, number=14,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LRN', index=23, number=15,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEMORY_DATA', index=24, number=29,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MULTINOMIAL_LOGISTIC_LOSS', index=25, number=16,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MVN', index=26, number=34,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POOLING', index=27, number=17,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POWER', index=28, number=26,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RELU', index=29, number=18,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIGMOID', index=30, number=19,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIGMOID_CROSS_ENTROPY_LOSS', index=31, number=27,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SILENCE', index=32, number=36,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SOFTMAX', index=33, number=20,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SOFTMAX_LOSS', index=34, number=21,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SPLIT', index=35, number=22,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SLICE', index=36, number=33,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TANH', index=37, number=23,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WINDOW_DATA', index=38, number=24,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='THRESHOLD', index=39, number=31,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=26110,
  serialized_end=26710,
)
_sym_db.RegisterEnumDescriptor(_V1LAYERPARAMETER_LAYERTYPE)

_V1LAYERPARAMETER_DIMCHECKMODE = _descriptor.EnumDescriptor(
  name='DimCheckMode',
  full_name='caffe.V1LayerParameter.DimCheckMode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STRICT', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PERMISSIVE', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2504,
  serialized_end=2546,
)
_sym_db.RegisterEnumDescriptor(_V1LAYERPARAMETER_DIMCHECKMODE)

_V0LAYERPARAMETER_POOLMETHOD = _descriptor.EnumDescriptor(
  name='PoolMethod',
  full_name='caffe.V0LayerParameter.PoolMethod',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MAX', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AVE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STOCHASTIC', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=20770,
  serialized_end=20816,
)
_sym_db.RegisterEnumDescriptor(_V0LAYERPARAMETER_POOLMETHOD)

_SUBPIXELUPPARAMETER_DATATYPE = _descriptor.EnumDescriptor(
  name='DataType',
  full_name='caffe.SubpixelUpParameter.DataType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FLOAT32', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UINT8', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=28709,
  serialized_end=28743,
)
_sym_db.RegisterEnumDescriptor(_SUBPIXELUPPARAMETER_DATATYPE)

_SUBPIXELUPPARAMETER_MODETYPE = _descriptor.EnumDescriptor(
  name='ModeType',
  full_name='caffe.SubpixelUpParameter.ModeType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='caffe', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='pytorch', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=28745,
  serialized_end=28779,
)
_sym_db.RegisterEnumDescriptor(_SUBPIXELUPPARAMETER_MODETYPE)

_REDUCEPARAMETER_REDUCEOP = _descriptor.EnumDescriptor(
  name='ReduceOp',
  full_name='caffe.ReduceParameter.ReduceOp',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='MEAN', index=0, number=0,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=28964,
  serialized_end=28984,
)
_sym_db.RegisterEnumDescriptor(_REDUCEPARAMETER_REDUCEOP)

_PADPARAMETER_PADMODE = _descriptor.EnumDescriptor(
  name='PadMode',
  full_name='caffe.PadParameter.PadMode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REFLECTION', index=0, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=29110,
  serialized_end=29135,
)
_sym_db.RegisterEnumDescriptor(_PADPARAMETER_PADMODE)

_MATMULPARAMETER_MATMULMODE = _descriptor.EnumDescriptor(
  name='MatMulMode',
  full_name='caffe.MatMulParameter.MatMulMode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NN', index=0, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NT', index=1, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TN', index=2, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TT', index=3, number=4,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=29299,
  serialized_end=29343,
)
_sym_db.RegisterEnumDescriptor(_MATMULPARAMETER_MATMULMODE)

_GRIDSAMPLEPARAMETER_PADDINGMODE = _descriptor.EnumDescriptor(
  name='PaddingMode',
  full_name='caffe.GridSampleParameter.PaddingMode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ZEROS', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BORDER', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=29850,
  serialized_end=29886,
)
_sym_db.RegisterEnumDescriptor(_GRIDSAMPLEPARAMETER_PADDINGMODE)

_GRIDSAMPLE3DPARAMETER_PADDINGMODE = _descriptor.EnumDescriptor(
  name='PaddingMode',
  full_name='caffe.GridSample3DParameter.PaddingMode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ZEROS', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BORDER', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=29850,
  serialized_end=29886,
)
_sym_db.RegisterEnumDescriptor(_GRIDSAMPLE3DPARAMETER_PADDINGMODE)


_BLOBSHAPE = _descriptor.Descriptor(
  name='BlobShape',
  full_name='caffe.BlobShape',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dim', full_name='caffe.BlobShape.dim', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=50,
)


_BLOBPROTO = _descriptor.Descriptor(
  name='BlobProto',
  full_name='caffe.BlobProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='shape', full_name='caffe.BlobProto.shape', index=0,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='caffe.BlobProto.data', index=1,
      number=5, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='diff', full_name='caffe.BlobProto.diff', index=2,
      number=6, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num', full_name='caffe.BlobProto.num', index=3,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='channels', full_name='caffe.BlobProto.channels', index=4,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='caffe.BlobProto.height', index=5,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='caffe.BlobProto.width', index=6,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='depth', full_name='caffe.BlobProto.depth', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=225,
)


_BLOBPROTOVECTOR = _descriptor.Descriptor(
  name='BlobProtoVector',
  full_name='caffe.BlobProtoVector',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='blobs', full_name='caffe.BlobProtoVector.blobs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=227,
  serialized_end=277,
)


_QUANTIZEPARAMETER = _descriptor.Descriptor(
  name='QuantizeParameter',
  full_name='caffe.QuantizeParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='caffe.QuantizeParameter.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b'quantize'.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step', full_name='caffe.QuantizeParameter.step', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='range_min', full_name='caffe.QuantizeParameter.range_min', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='range_max', full_name='caffe.QuantizeParameter.range_max', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(255),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='zero_point', full_name='caffe.QuantizeParameter.zero_point', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adjust_range', full_name='caffe.QuantizeParameter.adjust_range', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=280,
  serialized_end=440,
)


_DATUM = _descriptor.Descriptor(
  name='Datum',
  full_name='caffe.Datum',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='channels', full_name='caffe.Datum.channels', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='caffe.Datum.height', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='caffe.Datum.width', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='caffe.Datum.data', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b'',
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='label', full_name='caffe.Datum.label', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='float_data', full_name='caffe.Datum.float_data', index=5,
      number=6, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='encoded', full_name='caffe.Datum.encoded', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='depth', full_name='caffe.Datum.depth', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=443,
  serialized_end=587,
)


_FILLERPARAMETER = _descriptor.Descriptor(
  name='FillerParameter',
  full_name='caffe.FillerParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='caffe.FillerParameter.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b'constant'.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='caffe.FillerParameter.value', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='min', full_name='caffe.FillerParameter.min', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max', full_name='caffe.FillerParameter.max', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mean', full_name='caffe.FillerParameter.mean', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='std', full_name='caffe.FillerParameter.std', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sparse', full_name='caffe.FillerParameter.sparse', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=590,
  serialized_end=734,
)


_NETPARAMETER = _descriptor.Descriptor(
  name='NetParameter',
  full_name='caffe.NetParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='caffe.NetParameter.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='input', full_name='caffe.NetParameter.input', index=1,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='input_shape', full_name='caffe.NetParameter.input_shape', index=2,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='input_dim', full_name='caffe.NetParameter.input_dim', index=3,
      number=4, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='force_backward', full_name='caffe.NetParameter.force_backward', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state', full_name='caffe.NetParameter.state', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='debug_info', full_name='caffe.NetParameter.debug_info', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='layer', full_name='caffe.NetParameter.layer', index=7,
      number=100, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='layers', full_name='caffe.NetParameter.layers', index=8,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=737,
  serialized_end=1007,
)


_SOLVERPARAMETER = _descriptor.Descriptor(
  name='SolverParameter',
  full_name='caffe.SolverParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='net', full_name='caffe.SolverParameter.net', index=0,
      number=24, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='net_param', full_name='caffe.SolverParameter.net_param', index=1,
      number=25, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='train_net', full_name='caffe.SolverParameter.train_net', index=2,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_net', full_name='caffe.SolverParameter.test_net', index=3,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='train_net_param', full_name='caffe.SolverParameter.train_net_param', index=4,
      number=21, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_net_param', full_name='caffe.SolverParameter.test_net_param', index=5,
      number=22, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='train_state', full_name='caffe.SolverParameter.train_state', index=6,
      number=26, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_state', full_name='caffe.SolverParameter.test_state', index=7,
      number=27, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_iter', full_name='caffe.SolverParameter.test_iter', index=8,
      number=3, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_interval', full_name='caffe.SolverParameter.test_interval', index=9,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_compute_loss', full_name='caffe.SolverParameter.test_compute_loss', index=10,
      number=19, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_initialization', full_name='caffe.SolverParameter.test_initialization', index=11,
      number=32, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='base_lr', full_name='caffe.SolverParameter.base_lr', index=12,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='display', full_name='caffe.SolverParameter.display', index=13,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='average_loss', full_name='caffe.SolverParameter.average_loss', index=14,
      number=33, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_iter', full_name='caffe.SolverParameter.max_iter', index=15,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lr_policy', full_name='caffe.SolverParameter.lr_policy', index=16,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gamma', full_name='caffe.SolverParameter.gamma', index=17,
      number=9, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='power', full_name='caffe.SolverParameter.power', index=18,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='momentum', full_name='caffe.SolverParameter.momentum', index=19,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_decay', full_name='caffe.SolverParameter.weight_decay', index=20,
      number=12, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='regularization_type', full_name='caffe.SolverParameter.regularization_type', index=21,
      number=29, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b'L2'.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stepsize', full_name='caffe.SolverParameter.stepsize', index=22,
      number=13, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stepvalue', full_name='caffe.SolverParameter.stepvalue', index=23,
      number=34, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='clip_gradients', full_name='caffe.SolverParameter.clip_gradients', index=24,
      number=35, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(-1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='snapshot', full_name='caffe.SolverParameter.snapshot', index=25,
      number=14, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='snapshot_prefix', full_name='caffe.SolverParameter.snapshot_prefix', index=26,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='snapshot_diff', full_name='caffe.SolverParameter.snapshot_diff', index=27,
      number=16, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='solver_mode', full_name='caffe.SolverParameter.solver_mode', index=28,
      number=17, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='device_id', full_name='caffe.SolverParameter.device_id', index=29,
      number=18, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='random_seed', full_name='caffe.SolverParameter.random_seed', index=30,
      number=20, type=3, cpp_type=2, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='solver_type', full_name='caffe.SolverParameter.solver_type', index=31,
      number=30, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='delta', full_name='caffe.SolverParameter.delta', index=32,
      number=31, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1e-08),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='debug_info', full_name='caffe.SolverParameter.debug_info', index=33,
      number=23, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='snapshot_after_train', full_name='caffe.SolverParameter.snapshot_after_train', index=34,
      number=28, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SOLVERPARAMETER_SOLVERMODE,
    _SOLVERPARAMETER_SOLVERTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1010,
  serialized_end=2073,
)


_SOLVERSTATE = _descriptor.Descriptor(
  name='SolverState',
  full_name='caffe.SolverState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='iter', full_name='caffe.SolverState.iter', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='learned_net', full_name='caffe.SolverState.learned_net', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='history', full_name='caffe.SolverState.history', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='current_step', full_name='caffe.SolverState.current_step', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2075,
  serialized_end=2183,
)


_NETSTATE = _descriptor.Descriptor(
  name='NetState',
  full_name='caffe.NetState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='phase', full_name='caffe.NetState.phase', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='level', full_name='caffe.NetState.level', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stage', full_name='caffe.NetState.stage', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2185,
  serialized_end=2263,
)


_NETSTATERULE = _descriptor.Descriptor(
  name='NetStateRule',
  full_name='caffe.NetStateRule',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='phase', full_name='caffe.NetStateRule.phase', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='min_level', full_name='caffe.NetStateRule.min_level', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_level', full_name='caffe.NetStateRule.max_level', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stage', full_name='caffe.NetStateRule.stage', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='not_stage', full_name='caffe.NetStateRule.not_stage', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2265,
  serialized_end=2380,
)


_PARAMSPEC = _descriptor.Descriptor(
  name='ParamSpec',
  full_name='caffe.ParamSpec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='caffe.ParamSpec.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='share_mode', full_name='caffe.ParamSpec.share_mode', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lr_mult', full_name='caffe.ParamSpec.lr_mult', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='decay_mult', full_name='caffe.ParamSpec.decay_mult', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PARAMSPEC_DIMCHECKMODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2383,
  serialized_end=2546,
)


_LAYERPARAMETER = _descriptor.Descriptor(
  name='LayerParameter',
  full_name='caffe.LayerParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='caffe.LayerParameter.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='caffe.LayerParameter.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bottom', full_name='caffe.LayerParameter.bottom', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='top', full_name='caffe.LayerParameter.top', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='phase', full_name='caffe.LayerParameter.phase', index=4,
      number=10, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='loss_weight', full_name='caffe.LayerParameter.loss_weight', index=5,
      number=5, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='param', full_name='caffe.LayerParameter.param', index=6,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blobs', full_name='caffe.LayerParameter.blobs', index=7,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='include', full_name='caffe.LayerParameter.include', index=8,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exclude', full_name='caffe.LayerParameter.exclude', index=9,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='top_data_type', full_name='caffe.LayerParameter.top_data_type', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b'None'.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='forward_precision', full_name='caffe.LayerParameter.forward_precision', index=11,
      number=12, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transform_param', full_name='caffe.LayerParameter.transform_param', index=12,
      number=100, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='loss_param', full_name='caffe.LayerParameter.loss_param', index=13,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='accuracy_param', full_name='caffe.LayerParameter.accuracy_param', index=14,
      number=102, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='argmax_param', full_name='caffe.LayerParameter.argmax_param', index=15,
      number=103, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='concat_param', full_name='caffe.LayerParameter.concat_param', index=16,
      number=104, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contrastive_loss_param', full_name='caffe.LayerParameter.contrastive_loss_param', index=17,
      number=105, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='convolution_param', full_name='caffe.LayerParameter.convolution_param', index=18,
      number=106, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_param', full_name='caffe.LayerParameter.data_param', index=19,
      number=107, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dropout_param', full_name='caffe.LayerParameter.dropout_param', index=20,
      number=108, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dummy_data_param', full_name='caffe.LayerParameter.dummy_data_param', index=21,
      number=109, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eltwise_param', full_name='caffe.LayerParameter.eltwise_param', index=22,
      number=110, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exp_param', full_name='caffe.LayerParameter.exp_param', index=23,
      number=111, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hdf5_data_param', full_name='caffe.LayerParameter.hdf5_data_param', index=24,
      number=112, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hdf5_output_param', full_name='caffe.LayerParameter.hdf5_output_param', index=25,
      number=113, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hinge_loss_param', full_name='caffe.LayerParameter.hinge_loss_param', index=26,
      number=114, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='image_data_param', full_name='caffe.LayerParameter.image_data_param', index=27,
      number=115, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='infogain_loss_param', full_name='caffe.LayerParameter.infogain_loss_param', index=28,
      number=116, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inner_product_param', full_name='caffe.LayerParameter.inner_product_param', index=29,
      number=117, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lrn_param', full_name='caffe.LayerParameter.lrn_param', index=30,
      number=118, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='memory_data_param', full_name='caffe.LayerParameter.memory_data_param', index=31,
      number=119, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mvn_param', full_name='caffe.LayerParameter.mvn_param', index=32,
      number=120, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pooling_param', full_name='caffe.LayerParameter.pooling_param', index=33,
      number=121, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pooling_specific_param', full_name='caffe.LayerParameter.pooling_specific_param', index=34,
      number=1210, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='power_param', full_name='caffe.LayerParameter.power_param', index=35,
      number=122, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='relu_param', full_name='caffe.LayerParameter.relu_param', index=36,
      number=123, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sigmoid_param', full_name='caffe.LayerParameter.sigmoid_param', index=37,
      number=124, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='softmax_param', full_name='caffe.LayerParameter.softmax_param', index=38,
      number=125, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slice_param', full_name='caffe.LayerParameter.slice_param', index=39,
      number=126, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tanh_param', full_name='caffe.LayerParameter.tanh_param', index=40,
      number=127, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='threshold_param', full_name='caffe.LayerParameter.threshold_param', index=41,
      number=128, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='window_data_param', full_name='caffe.LayerParameter.window_data_param', index=42,
      number=129, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='python_param', full_name='caffe.LayerParameter.python_param', index=43,
      number=130, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ctc_param', full_name='caffe.LayerParameter.ctc_param', index=44,
      number=131, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='prelu_param', full_name='caffe.LayerParameter.prelu_param', index=45,
      number=132, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reshape_param', full_name='caffe.LayerParameter.reshape_param', index=46,
      number=133, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_pooling_param', full_name='caffe.LayerParameter.roi_pooling_param', index=47,
      number=134, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='affine_trans_param', full_name='caffe.LayerParameter.affine_trans_param', index=48,
      number=135, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_param', full_name='caffe.LayerParameter.roi_param', index=49,
      number=136, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='affine_trans_point_param', full_name='caffe.LayerParameter.affine_trans_point_param', index=50,
      number=137, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='permute_param', full_name='caffe.LayerParameter.permute_param', index=51,
      number=138, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='prior_box_param', full_name='caffe.LayerParameter.prior_box_param', index=52,
      number=139, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='calc_affine_mat_param', full_name='caffe.LayerParameter.calc_affine_mat_param', index=53,
      number=140, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='interp_param', full_name='caffe.LayerParameter.interp_param', index=54,
      number=141, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='detection_output_param', full_name='caffe.LayerParameter.detection_output_param', index=55,
      number=142, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='relu6_param', full_name='caffe.LayerParameter.relu6_param', index=56,
      number=143, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_param', full_name='caffe.LayerParameter.bias_param', index=57,
      number=144, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='zxybn_param', full_name='caffe.LayerParameter.zxybn_param', index=58,
      number=145, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='correlation_param', full_name='caffe.LayerParameter.correlation_param', index=59,
      number=150, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='psroi_pooling_param', full_name='caffe.LayerParameter.psroi_pooling_param', index=60,
      number=152, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nn_upsample_param', full_name='caffe.LayerParameter.nn_upsample_param', index=61,
      number=163, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_align_param', full_name='caffe.LayerParameter.roi_align_param', index=62,
      number=153, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rpn_proposal_param', full_name='caffe.LayerParameter.rpn_proposal_param', index=63,
      number=166, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_mask_pooling_param', full_name='caffe.LayerParameter.roi_mask_pooling_param', index=64,
      number=176, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='channel_shuffle_param', full_name='caffe.LayerParameter.channel_shuffle_param', index=65,
      number=177, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bn_param', full_name='caffe.LayerParameter.bn_param', index=66,
      number=178, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_align_pooling_pod_param', full_name='caffe.LayerParameter.roi_align_pooling_pod_param', index=67,
      number=205, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_align_pooling_param', full_name='caffe.LayerParameter.roi_align_pooling_param', index=68,
      number=206, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='psroi_align_pooling_param', full_name='caffe.LayerParameter.psroi_align_pooling_param', index=69,
      number=207, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='alignment_to_roi_param', full_name='caffe.LayerParameter.alignment_to_roi_param', index=70,
      number=218, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unpooling_param', full_name='caffe.LayerParameter.unpooling_param', index=71,
      number=243, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='log_param', full_name='caffe.LayerParameter.log_param', index=72,
      number=244, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reverse_param', full_name='caffe.LayerParameter.reverse_param', index=73,
      number=301, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lstm_param', full_name='caffe.LayerParameter.lstm_param', index=74,
      number=302, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='prior_vbox_param', full_name='caffe.LayerParameter.prior_vbox_param', index=75,
      number=304, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='btcostvolume_param', full_name='caffe.LayerParameter.btcostvolume_param', index=76,
      number=1001, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='moving_avg_param', full_name='caffe.LayerParameter.moving_avg_param', index=77,
      number=1799, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slicing_param', full_name='caffe.LayerParameter.slicing_param', index=78,
      number=5180, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bilateral_slicing_param', full_name='caffe.LayerParameter.bilateral_slicing_param', index=79,
      number=5181, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bilateral_slice_apply_param', full_name='caffe.LayerParameter.bilateral_slice_apply_param', index=80,
      number=5182, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adaptive_bilateral_slice_apply_param', full_name='caffe.LayerParameter.adaptive_bilateral_slice_apply_param', index=81,
      number=5183, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pointscurve_param', full_name='caffe.LayerParameter.pointscurve_param', index=82,
      number=2000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='subpixel_down_param', full_name='caffe.LayerParameter.subpixel_down_param', index=83,
      number=2001, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='subpixel_up_param', full_name='caffe.LayerParameter.subpixel_up_param', index=84,
      number=2002, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='clip_param', full_name='caffe.LayerParameter.clip_param', index=85,
      number=2003, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eltwise_affine_transform_param', full_name='caffe.LayerParameter.eltwise_affine_transform_param', index=86,
      number=2004, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tile_param', full_name='caffe.LayerParameter.tile_param', index=87,
      number=2005, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_param', full_name='caffe.LayerParameter.pad_param', index=88,
      number=2006, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reduce_param', full_name='caffe.LayerParameter.reduce_param', index=89,
      number=2007, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='convolution3d_param', full_name='caffe.LayerParameter.convolution3d_param', index=90,
      number=2008, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deconvolution3d_param', full_name='caffe.LayerParameter.deconvolution3d_param', index=91,
      number=2009, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='relu3d_param', full_name='caffe.LayerParameter.relu3d_param', index=92,
      number=2010, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='prelu3d_param', full_name='caffe.LayerParameter.prelu3d_param', index=93,
      number=2011, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='leaky_relu3d_param', full_name='caffe.LayerParameter.leaky_relu3d_param', index=94,
      number=2012, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='elu3d_param', full_name='caffe.LayerParameter.elu3d_param', index=95,
      number=2013, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='groupnorm3d_param', full_name='caffe.LayerParameter.groupnorm3d_param', index=96,
      number=2014, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='batchnorm3d_param', full_name='caffe.LayerParameter.batchnorm3d_param', index=97,
      number=2015, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='maxpooling3d_param', full_name='caffe.LayerParameter.maxpooling3d_param', index=98,
      number=2016, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='avepooling3d_param', full_name='caffe.LayerParameter.avepooling3d_param', index=99,
      number=2017, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dropout3d_param', full_name='caffe.LayerParameter.dropout3d_param', index=100,
      number=2018, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='interp3d_param', full_name='caffe.LayerParameter.interp3d_param', index=101,
      number=2019, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pooling3d_param', full_name='caffe.LayerParameter.pooling3d_param', index=102,
      number=2020, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reshape3d_param', full_name='caffe.LayerParameter.reshape3d_param', index=103,
      number=2022, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='softmax3d_param', full_name='caffe.LayerParameter.softmax3d_param', index=104,
      number=2023, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='argmax3d_param', full_name='caffe.LayerParameter.argmax3d_param', index=105,
      number=2024, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inner_product3d_param', full_name='caffe.LayerParameter.inner_product3d_param', index=106,
      number=2025, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slice3d_param', full_name='caffe.LayerParameter.slice3d_param', index=107,
      number=2026, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='concat3d_param', full_name='caffe.LayerParameter.concat3d_param', index=108,
      number=2027, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bn3d_param', full_name='caffe.LayerParameter.bn3d_param', index=109,
      number=2028, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='matmul3d_param', full_name='caffe.LayerParameter.matmul3d_param', index=110,
      number=2029, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transpose3d_param', full_name='caffe.LayerParameter.transpose3d_param', index=111,
      number=2030, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale_param', full_name='caffe.LayerParameter.scale_param', index=112,
      number=5001, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='crop_param', full_name='caffe.LayerParameter.crop_param', index=113,
      number=5002, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='recurrent_param', full_name='caffe.LayerParameter.recurrent_param', index=114,
      number=5004, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='batch_norm_param', full_name='caffe.LayerParameter.batch_norm_param', index=115,
      number=5005, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='flatten_param', full_name='caffe.LayerParameter.flatten_param', index=116,
      number=5006, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scales_param', full_name='caffe.LayerParameter.scales_param', index=117,
      number=5007, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transpose_param', full_name='caffe.LayerParameter.transpose_param', index=118,
      number=5008, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='heatmap_param', full_name='caffe.LayerParameter.heatmap_param', index=119,
      number=5009, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='normalize_param', full_name='caffe.LayerParameter.normalize_param', index=120,
      number=5010, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='correlation2d_param', full_name='caffe.LayerParameter.correlation2d_param', index=121,
      number=5011, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='matmul_param', full_name='caffe.LayerParameter.matmul_param', index=122,
      number=5012, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parameter_param', full_name='caffe.LayerParameter.parameter_param', index=123,
      number=5013, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pixelshuffle_param', full_name='caffe.LayerParameter.pixelshuffle_param', index=124,
      number=5014, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_transform_param', full_name='caffe.LayerParameter.roi_transform_param', index=125,
      number=5015, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='instance_norm_param', full_name='caffe.LayerParameter.instance_norm_param', index=126,
      number=5016, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='grid_sample_param', full_name='caffe.LayerParameter.grid_sample_param', index=127,
      number=5018, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reducel2_param', full_name='caffe.LayerParameter.reducel2_param', index=128,
      number=5017, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mean_param', full_name='caffe.LayerParameter.mean_param', index=129,
      number=5019, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='variance_param', full_name='caffe.LayerParameter.variance_param', index=130,
      number=5020, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='grid_sample_3d_param', full_name='caffe.LayerParameter.grid_sample_3d_param', index=131,
      number=5021, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gru_param', full_name='caffe.LayerParameter.gru_param', index=132,
      number=5022, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='correlationmig_param', full_name='caffe.LayerParameter.correlationmig_param', index=133,
      number=5023, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='argsort_param', full_name='caffe.LayerParameter.argsort_param', index=134,
      number=5024, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cumprod_param', full_name='caffe.LayerParameter.cumprod_param', index=135,
      number=5025, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='main_transform_param', full_name='caffe.LayerParameter.main_transform_param', index=136,
      number=5026, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quantize_param', full_name='caffe.LayerParameter.quantize_param', index=137,
      number=6000, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quantize_bits', full_name='caffe.LayerParameter.quantize_bits', index=138,
      number=6001, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LAYERPARAMETER_PRECISIONTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2549,
  serialized_end=9729,
)


_CONVOLUTION3DPARAMETER = _descriptor.Descriptor(
  name='Convolution3dParameter',
  full_name='caffe.Convolution3dParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_output', full_name='caffe.Convolution3dParameter.num_output', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_term', full_name='caffe.Convolution3dParameter.bias_term', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad', full_name='caffe.Convolution3dParameter.pad', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_d', full_name='caffe.Convolution3dParameter.pad_d', index=3,
      number=23, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_h', full_name='caffe.Convolution3dParameter.pad_h', index=4,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_w', full_name='caffe.Convolution3dParameter.pad_w', index=5,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_size', full_name='caffe.Convolution3dParameter.kernel_size', index=6,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_d', full_name='caffe.Convolution3dParameter.kernel_d', index=7,
      number=22, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_h', full_name='caffe.Convolution3dParameter.kernel_h', index=8,
      number=11, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_w', full_name='caffe.Convolution3dParameter.kernel_w', index=9,
      number=12, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='group', full_name='caffe.Convolution3dParameter.group', index=10,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride', full_name='caffe.Convolution3dParameter.stride', index=11,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_d', full_name='caffe.Convolution3dParameter.stride_d', index=12,
      number=21, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_h', full_name='caffe.Convolution3dParameter.stride_h', index=13,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_w', full_name='caffe.Convolution3dParameter.stride_w', index=14,
      number=14, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole', full_name='caffe.Convolution3dParameter.hole', index=15,
      number=17, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole_d', full_name='caffe.Convolution3dParameter.hole_d', index=16,
      number=19, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole_h', full_name='caffe.Convolution3dParameter.hole_h', index=17,
      number=18, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole_w', full_name='caffe.Convolution3dParameter.hole_w', index=18,
      number=20, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9732,
  serialized_end=10111,
)


_CONVOLUTIONTRANSPOSE3DPARAMETER = _descriptor.Descriptor(
  name='ConvolutionTranspose3dParameter',
  full_name='caffe.ConvolutionTranspose3dParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_output', full_name='caffe.ConvolutionTranspose3dParameter.num_output', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_term', full_name='caffe.ConvolutionTranspose3dParameter.bias_term', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad', full_name='caffe.ConvolutionTranspose3dParameter.pad', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_h', full_name='caffe.ConvolutionTranspose3dParameter.pad_h', index=3,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_w', full_name='caffe.ConvolutionTranspose3dParameter.pad_w', index=4,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_size', full_name='caffe.ConvolutionTranspose3dParameter.kernel_size', index=5,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_h', full_name='caffe.ConvolutionTranspose3dParameter.kernel_h', index=6,
      number=11, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_w', full_name='caffe.ConvolutionTranspose3dParameter.kernel_w', index=7,
      number=12, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='group', full_name='caffe.ConvolutionTranspose3dParameter.group', index=8,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride', full_name='caffe.ConvolutionTranspose3dParameter.stride', index=9,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_h', full_name='caffe.ConvolutionTranspose3dParameter.stride_h', index=10,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_w', full_name='caffe.ConvolutionTranspose3dParameter.stride_w', index=11,
      number=14, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole', full_name='caffe.ConvolutionTranspose3dParameter.hole', index=12,
      number=17, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole_h', full_name='caffe.ConvolutionTranspose3dParameter.hole_h', index=13,
      number=18, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10114,
  serialized_end=10410,
)


_DECONVOLUTION3DPARAMETER = _descriptor.Descriptor(
  name='Deconvolution3dParameter',
  full_name='caffe.Deconvolution3dParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_output', full_name='caffe.Deconvolution3dParameter.num_output', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_term', full_name='caffe.Deconvolution3dParameter.bias_term', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad', full_name='caffe.Deconvolution3dParameter.pad', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_d', full_name='caffe.Deconvolution3dParameter.pad_d', index=3,
      number=23, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_h', full_name='caffe.Deconvolution3dParameter.pad_h', index=4,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_w', full_name='caffe.Deconvolution3dParameter.pad_w', index=5,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_size', full_name='caffe.Deconvolution3dParameter.kernel_size', index=6,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_d', full_name='caffe.Deconvolution3dParameter.kernel_d', index=7,
      number=22, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_h', full_name='caffe.Deconvolution3dParameter.kernel_h', index=8,
      number=11, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_w', full_name='caffe.Deconvolution3dParameter.kernel_w', index=9,
      number=12, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='group', full_name='caffe.Deconvolution3dParameter.group', index=10,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride', full_name='caffe.Deconvolution3dParameter.stride', index=11,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_d', full_name='caffe.Deconvolution3dParameter.stride_d', index=12,
      number=21, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_h', full_name='caffe.Deconvolution3dParameter.stride_h', index=13,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_w', full_name='caffe.Deconvolution3dParameter.stride_w', index=14,
      number=14, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole', full_name='caffe.Deconvolution3dParameter.hole', index=15,
      number=17, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole_d', full_name='caffe.Deconvolution3dParameter.hole_d', index=16,
      number=19, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole_h', full_name='caffe.Deconvolution3dParameter.hole_h', index=17,
      number=18, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole_w', full_name='caffe.Deconvolution3dParameter.hole_w', index=18,
      number=20, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_pad', full_name='caffe.Deconvolution3dParameter.out_pad', index=19,
      number=24, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_pad_d', full_name='caffe.Deconvolution3dParameter.out_pad_d', index=20,
      number=25, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_pad_h', full_name='caffe.Deconvolution3dParameter.out_pad_h', index=21,
      number=26, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_pad_w', full_name='caffe.Deconvolution3dParameter.out_pad_w', index=22,
      number=27, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10413,
  serialized_end=10880,
)


_RELU3DPARAMETER = _descriptor.Descriptor(
  name='ReLU3dParameter',
  full_name='caffe.ReLU3dParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='negative_slope', full_name='caffe.ReLU3dParameter.negative_slope', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='channel_shared', full_name='caffe.ReLU3dParameter.channel_shared', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10882,
  serialized_end=10957,
)


_GROUPNORM3DPARAMETER = _descriptor.Descriptor(
  name='GroupNorm3dParameter',
  full_name='caffe.GroupNorm3dParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='group', full_name='caffe.GroupNorm3dParameter.group', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='moving_average_fraction', full_name='caffe.GroupNorm3dParameter.moving_average_fraction', index=1,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='use_global_stats', full_name='caffe.GroupNorm3dParameter.use_global_stats', index=2,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eps', full_name='caffe.GroupNorm3dParameter.eps', index=3,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1e-06),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10960,
  serialized_end=11091,
)


_BATCHNORM3DPARAMETER = _descriptor.Descriptor(
  name='BatchNorm3dParameter',
  full_name='caffe.BatchNorm3dParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='moving_average_fraction', full_name='caffe.BatchNorm3dParameter.moving_average_fraction', index=0,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='use_global_stats', full_name='caffe.BatchNorm3dParameter.use_global_stats', index=1,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eps', full_name='caffe.BatchNorm3dParameter.eps', index=2,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1e-06),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11093,
  serialized_end=11206,
)


_POOLING3DPARAMETER = _descriptor.Descriptor(
  name='Pooling3dParameter',
  full_name='caffe.Pooling3dParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pool', full_name='caffe.Pooling3dParameter.pool', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad', full_name='caffe.Pooling3dParameter.pad', index=1,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_d', full_name='caffe.Pooling3dParameter.pad_d', index=2,
      number=11, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_h', full_name='caffe.Pooling3dParameter.pad_h', index=3,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_w', full_name='caffe.Pooling3dParameter.pad_w', index=4,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_size', full_name='caffe.Pooling3dParameter.kernel_size', index=5,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_d', full_name='caffe.Pooling3dParameter.kernel_d', index=6,
      number=14, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_h', full_name='caffe.Pooling3dParameter.kernel_h', index=7,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_w', full_name='caffe.Pooling3dParameter.kernel_w', index=8,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride', full_name='caffe.Pooling3dParameter.stride', index=9,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_d', full_name='caffe.Pooling3dParameter.stride_d', index=10,
      number=15, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_h', full_name='caffe.Pooling3dParameter.stride_h', index=11,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_w', full_name='caffe.Pooling3dParameter.stride_w', index=12,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='global_pooling', full_name='caffe.Pooling3dParameter.global_pooling', index=13,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ceil_mode', full_name='caffe.Pooling3dParameter.ceil_mode', index=14,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _POOLING3DPARAMETER_POOLMETHOD,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11209,
  serialized_end=11592,
)


_NORMALIZEPARAMETER = _descriptor.Descriptor(
  name='NormalizeParameter',
  full_name='caffe.NormalizeParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='coeff', full_name='caffe.NormalizeParameter.coeff', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mode', full_name='caffe.NormalizeParameter.mode', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _NORMALIZEPARAMETER_NORMALIZEMETHOD,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11595,
  serialized_end=11729,
)


_HEATMAP2COORDPARAMETER = _descriptor.Descriptor(
  name='HeatMap2CoordParameter',
  full_name='caffe.HeatMap2CoordParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='coord_h', full_name='caffe.HeatMap2CoordParameter.coord_h', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='coord_w', full_name='caffe.HeatMap2CoordParameter.coord_w', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='coord_reposition', full_name='caffe.HeatMap2CoordParameter.coord_reposition', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11731,
  serialized_end=11828,
)


_BIASPARAMETER = _descriptor.Descriptor(
  name='BiasParameter',
  full_name='caffe.BiasParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.BiasParameter.axis', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num_axes', full_name='caffe.BiasParameter.num_axes', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='filler', full_name='caffe.BiasParameter.filler', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11830,
  serialized_end=11923,
)


_ALIGNMENTTOROIPARAMETER = _descriptor.Descriptor(
  name='AlignmentToROIParameter',
  full_name='caffe.AlignmentToROIParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mouth_scale', full_name='caffe.AlignmentToROIParameter.mouth_scale', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eye_scale', full_name='caffe.AlignmentToROIParameter.eye_scale', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_width', full_name='caffe.AlignmentToROIParameter.max_width', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=256,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_height', full_name='caffe.AlignmentToROIParameter.max_height', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=256,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11925,
  serialized_end=12045,
)


_ROIALIGNPOOLINGPARAMETER = _descriptor.Descriptor(
  name='ROIAlignPoolingParameter',
  full_name='caffe.ROIAlignPoolingParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pooled_h', full_name='caffe.ROIAlignPoolingParameter.pooled_h', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pooled_w', full_name='caffe.ROIAlignPoolingParameter.pooled_w', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spatial_scale', full_name='caffe.ROIAlignPoolingParameter.spatial_scale', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sample_num', full_name='caffe.ROIAlignPoolingParameter.sample_num', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12047,
  serialized_end=12164,
)


_ROIALIGNPOOLINGPODPARAMETER = _descriptor.Descriptor(
  name='ROIAlignPoolingPodParameter',
  full_name='caffe.ROIAlignPoolingPodParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pooled_h', full_name='caffe.ROIAlignPoolingPodParameter.pooled_h', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pooled_w', full_name='caffe.ROIAlignPoolingPodParameter.pooled_w', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spatial_scale', full_name='caffe.ROIAlignPoolingPodParameter.spatial_scale', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sample_num', full_name='caffe.ROIAlignPoolingPodParameter.sample_num', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12166,
  serialized_end=12286,
)


_PSROIALIGNPOOLINGPARAMETER = _descriptor.Descriptor(
  name='PSROIAlignPoolingParameter',
  full_name='caffe.PSROIAlignPoolingParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='spatial_scale', full_name='caffe.PSROIAlignPoolingParameter.spatial_scale', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='output_dim', full_name='caffe.PSROIAlignPoolingParameter.output_dim', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='group_size', full_name='caffe.PSROIAlignPoolingParameter.group_size', index=2,
      number=3, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sample_num', full_name='caffe.PSROIAlignPoolingParameter.sample_num', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12288,
  serialized_end=12402,
)


_UNPOOLINGPARAMETER = _descriptor.Descriptor(
  name='UnpoolingParameter',
  full_name='caffe.UnpoolingParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='unpool', full_name='caffe.UnpoolingParameter.unpool', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad', full_name='caffe.UnpoolingParameter.pad', index=1,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_h', full_name='caffe.UnpoolingParameter.pad_h', index=2,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_w', full_name='caffe.UnpoolingParameter.pad_w', index=3,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_size', full_name='caffe.UnpoolingParameter.kernel_size', index=4,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_h', full_name='caffe.UnpoolingParameter.kernel_h', index=5,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_w', full_name='caffe.UnpoolingParameter.kernel_w', index=6,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride', full_name='caffe.UnpoolingParameter.stride', index=7,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_h', full_name='caffe.UnpoolingParameter.stride_h', index=8,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_w', full_name='caffe.UnpoolingParameter.stride_w', index=9,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unpool_size', full_name='caffe.UnpoolingParameter.unpool_size', index=10,
      number=12, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unpool_h', full_name='caffe.UnpoolingParameter.unpool_h', index=11,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unpool_w', full_name='caffe.UnpoolingParameter.unpool_w', index=12,
      number=14, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='engine', full_name='caffe.UnpoolingParameter.engine', index=13,
      number=11, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _UNPOOLINGPARAMETER_UNPOOLMETHOD,
    _UNPOOLINGPARAMETER_ENGINE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12405,
  serialized_end=12844,
)


_REVERSEPARAMETER = _descriptor.Descriptor(
  name='ReverseParameter',
  full_name='caffe.ReverseParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.ReverseParameter.axis', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12846,
  serialized_end=12881,
)


_LSTMPARAMETER = _descriptor.Descriptor(
  name='LSTMParameter',
  full_name='caffe.LSTMParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_output', full_name='caffe.LSTMParameter.num_output', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='clipping_threshold', full_name='caffe.LSTMParameter.clipping_threshold', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_filler', full_name='caffe.LSTMParameter.weight_filler', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_filler', full_name='caffe.LSTMParameter.bias_filler', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='batch_size', full_name='caffe.LSTMParameter.batch_size', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12884,
  serialized_end=13065,
)


_PRIORVBOXPARAMETER = _descriptor.Descriptor(
  name='PriorVBoxParameter',
  full_name='caffe.PriorVBoxParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='height', full_name='caffe.PriorVBoxParameter.height', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='caffe.PriorVBoxParameter.width', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(16),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='clip', full_name='caffe.PriorVBoxParameter.clip', index=2,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='variance', full_name='caffe.PriorVBoxParameter.variance', index=3,
      number=6, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='img_size', full_name='caffe.PriorVBoxParameter.img_size', index=4,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='img_h', full_name='caffe.PriorVBoxParameter.img_h', index=5,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='img_w', full_name='caffe.PriorVBoxParameter.img_w', index=6,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step', full_name='caffe.PriorVBoxParameter.step', index=7,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step_h', full_name='caffe.PriorVBoxParameter.step_h', index=8,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step_w', full_name='caffe.PriorVBoxParameter.step_w', index=9,
      number=12, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offset', full_name='caffe.PriorVBoxParameter.offset', index=10,
      number=13, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13068,
  serialized_end=13277,
)


_NNUPSAMPLEPARAMETER = _descriptor.Descriptor(
  name='NNUpsampleParameter',
  full_name='caffe.NNUpsampleParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='resize', full_name='caffe.NNUpsampleParameter.resize', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13279,
  serialized_end=13319,
)


_BTCOSTVOLUMEPARAMETER = _descriptor.Descriptor(
  name='BTCostVolumeParameter',
  full_name='caffe.BTCostVolumeParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='min_disparity', full_name='caffe.BTCostVolumeParameter.min_disparity', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='disparity_num', full_name='caffe.BTCostVolumeParameter.disparity_num', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=64,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='save_disparity', full_name='caffe.BTCostVolumeParameter.save_disparity', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step_w', full_name='caffe.BTCostVolumeParameter.step_w', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step_h', full_name='caffe.BTCostVolumeParameter.step_h', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step_mode', full_name='caffe.BTCostVolumeParameter.step_mode', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cost_domain_type', full_name='caffe.BTCostVolumeParameter.cost_domain_type', index=6,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='implement_mode', full_name='caffe.BTCostVolumeParameter.implement_mode', index=7,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _BTCOSTVOLUMEPARAMETER_COSTDOMAINTYPE,
    _BTCOSTVOLUMEPARAMETER_IMPLEMENTMODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13322,
  serialized_end=13771,
)


_CORRELATIONPARAMETER = _descriptor.Descriptor(
  name='CorrelationParameter',
  full_name='caffe.CorrelationParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pad', full_name='caffe.CorrelationParameter.pad', index=0,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_size', full_name='caffe.CorrelationParameter.kernel_size', index=1,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_displacement', full_name='caffe.CorrelationParameter.max_displacement', index=2,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_1', full_name='caffe.CorrelationParameter.stride_1', index=3,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_2', full_name='caffe.CorrelationParameter.stride_2', index=4,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='single_direction', full_name='caffe.CorrelationParameter.single_direction', index=5,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='do_abs', full_name='caffe.CorrelationParameter.do_abs', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='correlation_type', full_name='caffe.CorrelationParameter.correlation_type', index=7,
      number=15, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_shift', full_name='caffe.CorrelationParameter.pad_shift', index=8,
      number=16, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mvn', full_name='caffe.CorrelationParameter.mvn', index=9,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CORRELATIONPARAMETER_CORRELATIONTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13774,
  serialized_end=14123,
)


_PSROIPOOLINGPARAMETER = _descriptor.Descriptor(
  name='PSROIPoolingParameter',
  full_name='caffe.PSROIPoolingParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='spatial_scale', full_name='caffe.PSROIPoolingParameter.spatial_scale', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='output_dim', full_name='caffe.PSROIPoolingParameter.output_dim', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='group_size', full_name='caffe.PSROIPoolingParameter.group_size', index=2,
      number=3, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_scale', full_name='caffe.PSROIPoolingParameter.roi_scale', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14125,
  serialized_end=14233,
)


_SCALEPARAMETER = _descriptor.Descriptor(
  name='ScaleParameter',
  full_name='caffe.ScaleParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.ScaleParameter.axis', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num_axes', full_name='caffe.ScaleParameter.num_axes', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='filler', full_name='caffe.ScaleParameter.filler', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_term', full_name='caffe.ScaleParameter.bias_term', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_filler', full_name='caffe.ScaleParameter.bias_filler', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14236,
  serialized_end=14401,
)


_SCALESPARAMETER = _descriptor.Descriptor(
  name='ScalesParameter',
  full_name='caffe.ScalesParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='alpha', full_name='caffe.ScalesParameter.alpha', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beta', full_name='caffe.ScalesParameter.beta', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14403,
  serialized_end=14455,
)


_TRANSPOSEPARAMETER = _descriptor.Descriptor(
  name='TransposeParameter',
  full_name='caffe.TransposeParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dim', full_name='caffe.TransposeParameter.dim', index=0,
      number=1, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14457,
  serialized_end=14490,
)


_ROIALIGNPARAMETER = _descriptor.Descriptor(
  name='ROIAlignParameter',
  full_name='caffe.ROIAlignParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pooled_h', full_name='caffe.ROIAlignParameter.pooled_h', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pooled_w', full_name='caffe.ROIAlignParameter.pooled_w', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spatial_scale', full_name='caffe.ROIAlignParameter.spatial_scale', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14492,
  serialized_end=14579,
)


_RPNPROPOSALLAYERPARAMETER_TESTDESCPARAM = _descriptor.Descriptor(
  name='TestDescParam',
  full_name='caffe.RpnProposalLayerParameter.TestDescParam',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rpn_pre_nms_top_n', full_name='caffe.RpnProposalLayerParameter.TestDescParam.rpn_pre_nms_top_n', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=6000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rpn_post_nms_top_n', full_name='caffe.RpnProposalLayerParameter.TestDescParam.rpn_post_nms_top_n', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=300,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rpn_min_size', full_name='caffe.RpnProposalLayerParameter.TestDescParam.rpn_min_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=16,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rpn_nms_thresh', full_name='caffe.RpnProposalLayerParameter.TestDescParam.rpn_nms_thresh', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.7),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15048,
  serialized_end=15184,
)

_RPNPROPOSALLAYERPARAMETER_TRAINDESCPARAM = _descriptor.Descriptor(
  name='TrainDescParam',
  full_name='caffe.RpnProposalLayerParameter.TrainDescParam',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rpn_pre_nms_top_n', full_name='caffe.RpnProposalLayerParameter.TrainDescParam.rpn_pre_nms_top_n', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=12000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rpn_post_nms_top_n', full_name='caffe.RpnProposalLayerParameter.TrainDescParam.rpn_post_nms_top_n', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=2000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rpn_min_size', full_name='caffe.RpnProposalLayerParameter.TrainDescParam.rpn_min_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=16,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rpn_nms_thresh', full_name='caffe.RpnProposalLayerParameter.TrainDescParam.rpn_nms_thresh', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.7),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15187,
  serialized_end=15326,
)

_RPNPROPOSALLAYERPARAMETER = _descriptor.Descriptor(
  name='RpnProposalLayerParameter',
  full_name='caffe.RpnProposalLayerParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='feat_stride', full_name='caffe.RpnProposalLayerParameter.feat_stride', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=16,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='allowed_border', full_name='caffe.RpnProposalLayerParameter.allowed_border', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_desc_param', full_name='caffe.RpnProposalLayerParameter.test_desc_param', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='train_desc_param', full_name='caffe.RpnProposalLayerParameter.train_desc_param', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='anchor_scales', full_name='caffe.RpnProposalLayerParameter.anchor_scales', index=4,
      number=5, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='anchor_ratios', full_name='caffe.RpnProposalLayerParameter.anchor_ratios', index=5,
      number=6, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='score', full_name='caffe.RpnProposalLayerParameter.score', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='withOutSoftmax', full_name='caffe.RpnProposalLayerParameter.withOutSoftmax', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bbox_reg_xy_limit', full_name='caffe.RpnProposalLayerParameter.bbox_reg_xy_limit', index=8,
      number=9, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bbox_reg_hw_temp', full_name='caffe.RpnProposalLayerParameter.bbox_reg_hw_temp', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bbox_reg_hw_ratio_thresh_hi', full_name='caffe.RpnProposalLayerParameter.bbox_reg_hw_ratio_thresh_hi', index=10,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bbox_reg_hw_ratio_thresh_lo', full_name='caffe.RpnProposalLayerParameter.bbox_reg_hw_ratio_thresh_lo', index=11,
      number=12, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.2),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_RPNPROPOSALLAYERPARAMETER_TESTDESCPARAM, _RPNPROPOSALLAYERPARAMETER_TRAINDESCPARAM, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14582,
  serialized_end=15326,
)


_ROIMASKPOOLINGPARAMETER = _descriptor.Descriptor(
  name='ROIMaskPoolingParameter',
  full_name='caffe.ROIMaskPoolingParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pooled_h', full_name='caffe.ROIMaskPoolingParameter.pooled_h', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pooled_w', full_name='caffe.ROIMaskPoolingParameter.pooled_w', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spatial_scale', full_name='caffe.ROIMaskPoolingParameter.spatial_scale', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='half_part', full_name='caffe.ROIMaskPoolingParameter.half_part', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_scale', full_name='caffe.ROIMaskPoolingParameter.roi_scale', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mask_scale', full_name='caffe.ROIMaskPoolingParameter.mask_scale', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15329,
  serialized_end=15489,
)


_ZXYBNPARAMETER = _descriptor.Descriptor(
  name='ZXYBNParameter',
  full_name='caffe.ZXYBNParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='slope_filler', full_name='caffe.ZXYBNParameter.slope_filler', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_filler', full_name='caffe.ZXYBNParameter.bias_filler', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='momentum', full_name='caffe.ZXYBNParameter.momentum', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.9),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eps', full_name='caffe.ZXYBNParameter.eps', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1e-05),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='frozen', full_name='caffe.ZXYBNParameter.frozen', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='engine', full_name='caffe.ZXYBNParameter.engine', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ZXYBNPARAMETER_ENGINE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15492,
  serialized_end=15765,
)


_TRANSFORMATIONPARAMETER = _descriptor.Descriptor(
  name='TransformationParameter',
  full_name='caffe.TransformationParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='scale', full_name='caffe.TransformationParameter.scale', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mirror', full_name='caffe.TransformationParameter.mirror', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='crop_size', full_name='caffe.TransformationParameter.crop_size', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mean_file', full_name='caffe.TransformationParameter.mean_file', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mean_value', full_name='caffe.TransformationParameter.mean_value', index=4,
      number=5, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15767,
  serialized_end=15894,
)


_LOSSPARAMETER = _descriptor.Descriptor(
  name='LossParameter',
  full_name='caffe.LossParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ignore_label', full_name='caffe.LossParameter.ignore_label', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='normalize', full_name='caffe.LossParameter.normalize', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15896,
  serialized_end=15958,
)


_ACCURACYPARAMETER = _descriptor.Descriptor(
  name='AccuracyParameter',
  full_name='caffe.AccuracyParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='top_k', full_name='caffe.AccuracyParameter.top_k', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15960,
  serialized_end=15997,
)


_ARGMAXPARAMETER = _descriptor.Descriptor(
  name='ArgMaxParameter',
  full_name='caffe.ArgMaxParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='out_max_val', full_name='caffe.ArgMaxParameter.out_max_val', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='top_k', full_name='caffe.ArgMaxParameter.top_k', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.ArgMaxParameter.axis', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15999,
  serialized_end=16076,
)


_CONCATPARAMETER = _descriptor.Descriptor(
  name='ConcatParameter',
  full_name='caffe.ConcatParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.ConcatParameter.axis', index=0,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='concat_dim', full_name='caffe.ConcatParameter.concat_dim', index=1,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16078,
  serialized_end=16135,
)


_CONTRASTIVELOSSPARAMETER = _descriptor.Descriptor(
  name='ContrastiveLossParameter',
  full_name='caffe.ContrastiveLossParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='margin', full_name='caffe.ContrastiveLossParameter.margin', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16137,
  serialized_end=16182,
)


_CONVOLUTIONPARAMETER = _descriptor.Descriptor(
  name='ConvolutionParameter',
  full_name='caffe.ConvolutionParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_output', full_name='caffe.ConvolutionParameter.num_output', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_term', full_name='caffe.ConvolutionParameter.bias_term', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad', full_name='caffe.ConvolutionParameter.pad', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_h', full_name='caffe.ConvolutionParameter.pad_h', index=3,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_w', full_name='caffe.ConvolutionParameter.pad_w', index=4,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_size', full_name='caffe.ConvolutionParameter.kernel_size', index=5,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_h', full_name='caffe.ConvolutionParameter.kernel_h', index=6,
      number=11, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_w', full_name='caffe.ConvolutionParameter.kernel_w', index=7,
      number=12, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='group', full_name='caffe.ConvolutionParameter.group', index=8,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride', full_name='caffe.ConvolutionParameter.stride', index=9,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_h', full_name='caffe.ConvolutionParameter.stride_h', index=10,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_w', full_name='caffe.ConvolutionParameter.stride_w', index=11,
      number=14, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole', full_name='caffe.ConvolutionParameter.hole', index=12,
      number=17, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole_h', full_name='caffe.ConvolutionParameter.hole_h', index=13,
      number=18, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hole_w', full_name='caffe.ConvolutionParameter.hole_w', index=14,
      number=19, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_filler', full_name='caffe.ConvolutionParameter.weight_filler', index=15,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_filler', full_name='caffe.ConvolutionParameter.bias_filler', index=16,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='engine', full_name='caffe.ConvolutionParameter.engine', index=17,
      number=15, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ForwardAlgo', full_name='caffe.ConvolutionParameter.ForwardAlgo', index=18,
      number=20, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ntile_width', full_name='caffe.ConvolutionParameter.ntile_width', index=19,
      number=72, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ntile_height', full_name='caffe.ConvolutionParameter.ntile_height', index=20,
      number=73, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quantize_param', full_name='caffe.ConvolutionParameter.quantize_param', index=21,
      number=74, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='is_relu', full_name='caffe.ConvolutionParameter.is_relu', index=22,
      number=77, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='perchannel_quantize_param', full_name='caffe.ConvolutionParameter.perchannel_quantize_param', index=23,
      number=75, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_pad', full_name='caffe.ConvolutionParameter.out_pad', index=24,
      number=78, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_pad_h', full_name='caffe.ConvolutionParameter.out_pad_h', index=25,
      number=79, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_pad_w', full_name='caffe.ConvolutionParameter.out_pad_w', index=26,
      number=80, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CONVOLUTIONPARAMETER_ENGINE,
    _CONVOLUTIONPARAMETER_PPLCONVOLUTIONFORWARDALGO_T,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16185,
  serialized_end=17701,
)


_DATAPARAMETER = _descriptor.Descriptor(
  name='DataParameter',
  full_name='caffe.DataParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source', full_name='caffe.DataParameter.source', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='batch_size', full_name='caffe.DataParameter.batch_size', index=1,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rand_skip', full_name='caffe.DataParameter.rand_skip', index=2,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='backend', full_name='caffe.DataParameter.backend', index=3,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale', full_name='caffe.DataParameter.scale', index=4,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mean_file', full_name='caffe.DataParameter.mean_file', index=5,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='crop_size', full_name='caffe.DataParameter.crop_size', index=6,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mirror', full_name='caffe.DataParameter.mirror', index=7,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='force_encoded_color', full_name='caffe.DataParameter.force_encoded_color', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _DATAPARAMETER_DB,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17704,
  serialized_end=17975,
)


_NONMAXIMUMSUPPRESSIONPARAMETER = _descriptor.Descriptor(
  name='NonMaximumSuppressionParameter',
  full_name='caffe.NonMaximumSuppressionParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nms_threshold', full_name='caffe.NonMaximumSuppressionParameter.nms_threshold', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.3),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='top_k', full_name='caffe.NonMaximumSuppressionParameter.top_k', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17977,
  serialized_end=18052,
)


_SAVEOUTPUTPARAMETER = _descriptor.Descriptor(
  name='SaveOutputParameter',
  full_name='caffe.SaveOutputParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='output_directory', full_name='caffe.SaveOutputParameter.output_directory', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='output_name_prefix', full_name='caffe.SaveOutputParameter.output_name_prefix', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='output_format', full_name='caffe.SaveOutputParameter.output_format', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='label_map_file', full_name='caffe.SaveOutputParameter.label_map_file', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name_size_file', full_name='caffe.SaveOutputParameter.name_size_file', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num_test_image', full_name='caffe.SaveOutputParameter.num_test_image', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18055,
  serialized_end=18225,
)


_DETECTIONOUTPUTPARAMETER = _descriptor.Descriptor(
  name='DetectionOutputParameter',
  full_name='caffe.DetectionOutputParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_classes', full_name='caffe.DetectionOutputParameter.num_classes', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='share_location', full_name='caffe.DetectionOutputParameter.share_location', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='background_label_id', full_name='caffe.DetectionOutputParameter.background_label_id', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nms_param', full_name='caffe.DetectionOutputParameter.nms_param', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='save_output_param', full_name='caffe.DetectionOutputParameter.save_output_param', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='code_type', full_name='caffe.DetectionOutputParameter.code_type', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='variance_encoded_in_target', full_name='caffe.DetectionOutputParameter.variance_encoded_in_target', index=6,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keep_top_k', full_name='caffe.DetectionOutputParameter.keep_top_k', index=7,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='confidence_threshold', full_name='caffe.DetectionOutputParameter.confidence_threshold', index=8,
      number=9, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='visualize', full_name='caffe.DetectionOutputParameter.visualize', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='visualize_threshold', full_name='caffe.DetectionOutputParameter.visualize_threshold', index=10,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18228,
  serialized_end=18664,
)


_DROPOUTPARAMETER = _descriptor.Descriptor(
  name='DropoutParameter',
  full_name='caffe.DropoutParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dropout_ratio', full_name='caffe.DropoutParameter.dropout_ratio', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18666,
  serialized_end=18712,
)


_DUMMYDATAPARAMETER = _descriptor.Descriptor(
  name='DummyDataParameter',
  full_name='caffe.DummyDataParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='data_filler', full_name='caffe.DummyDataParameter.data_filler', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shape', full_name='caffe.DummyDataParameter.shape', index=1,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num', full_name='caffe.DummyDataParameter.num', index=2,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='channels', full_name='caffe.DummyDataParameter.channels', index=3,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='caffe.DummyDataParameter.height', index=4,
      number=4, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='caffe.DummyDataParameter.width', index=5,
      number=5, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18715,
  serialized_end=18875,
)


_ELTWISEPARAMETER = _descriptor.Descriptor(
  name='EltwiseParameter',
  full_name='caffe.EltwiseParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='operation', full_name='caffe.EltwiseParameter.operation', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='coeff', full_name='caffe.EltwiseParameter.coeff', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stable_prod_grad', full_name='caffe.EltwiseParameter.stable_prod_grad', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ELTWISEPARAMETER_ELTWISEOP,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18878,
  serialized_end=19043,
)


_EXPPARAMETER = _descriptor.Descriptor(
  name='ExpParameter',
  full_name='caffe.ExpParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='base', full_name='caffe.ExpParameter.base', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(-1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale', full_name='caffe.ExpParameter.scale', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shift', full_name='caffe.ExpParameter.shift', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19045,
  serialized_end=19113,
)


_LOGPARAMETER = _descriptor.Descriptor(
  name='LogParameter',
  full_name='caffe.LogParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='base', full_name='caffe.LogParameter.base', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(-1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale', full_name='caffe.LogParameter.scale', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shift', full_name='caffe.LogParameter.shift', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19115,
  serialized_end=19183,
)


_HDF5DATAPARAMETER = _descriptor.Descriptor(
  name='HDF5DataParameter',
  full_name='caffe.HDF5DataParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source', full_name='caffe.HDF5DataParameter.source', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='batch_size', full_name='caffe.HDF5DataParameter.batch_size', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19185,
  serialized_end=19240,
)


_HDF5OUTPUTPARAMETER = _descriptor.Descriptor(
  name='HDF5OutputParameter',
  full_name='caffe.HDF5OutputParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='file_name', full_name='caffe.HDF5OutputParameter.file_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19242,
  serialized_end=19282,
)


_HINGELOSSPARAMETER = _descriptor.Descriptor(
  name='HingeLossParameter',
  full_name='caffe.HingeLossParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='norm', full_name='caffe.HingeLossParameter.norm', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _HINGELOSSPARAMETER_NORM,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19284,
  serialized_end=19378,
)


_IMAGEDATAPARAMETER = _descriptor.Descriptor(
  name='ImageDataParameter',
  full_name='caffe.ImageDataParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source', full_name='caffe.ImageDataParameter.source', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='batch_size', full_name='caffe.ImageDataParameter.batch_size', index=1,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rand_skip', full_name='caffe.ImageDataParameter.rand_skip', index=2,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shuffle', full_name='caffe.ImageDataParameter.shuffle', index=3,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_height', full_name='caffe.ImageDataParameter.new_height', index=4,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_width', full_name='caffe.ImageDataParameter.new_width', index=5,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='is_color', full_name='caffe.ImageDataParameter.is_color', index=6,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale', full_name='caffe.ImageDataParameter.scale', index=7,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mean_file', full_name='caffe.ImageDataParameter.mean_file', index=8,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='crop_size', full_name='caffe.ImageDataParameter.crop_size', index=9,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mirror', full_name='caffe.ImageDataParameter.mirror', index=10,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='root_folder', full_name='caffe.ImageDataParameter.root_folder', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19381,
  serialized_end=19657,
)


_INFOGAINLOSSPARAMETER = _descriptor.Descriptor(
  name='InfogainLossParameter',
  full_name='caffe.InfogainLossParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source', full_name='caffe.InfogainLossParameter.source', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19659,
  serialized_end=19698,
)


_INNERPRODUCTPARAMETER = _descriptor.Descriptor(
  name='InnerProductParameter',
  full_name='caffe.InnerProductParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_output', full_name='caffe.InnerProductParameter.num_output', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_term', full_name='caffe.InnerProductParameter.bias_term', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_filler', full_name='caffe.InnerProductParameter.weight_filler', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_filler', full_name='caffe.InnerProductParameter.bias_filler', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.InnerProductParameter.axis', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quantize_param', full_name='caffe.InnerProductParameter.quantize_param', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='perchannel_quantize_param', full_name='caffe.InnerProductParameter.perchannel_quantize_param', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19701,
  serialized_end=19989,
)


_LRNPARAMETER = _descriptor.Descriptor(
  name='LRNParameter',
  full_name='caffe.LRNParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='local_size', full_name='caffe.LRNParameter.local_size', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=5,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='alpha', full_name='caffe.LRNParameter.alpha', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beta', full_name='caffe.LRNParameter.beta', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.75),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='norm_region', full_name='caffe.LRNParameter.norm_region', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='k', full_name='caffe.LRNParameter.k', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _LRNPARAMETER_NORMREGION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19992,
  serialized_end=20206,
)


_MEMORYDATAPARAMETER = _descriptor.Descriptor(
  name='MemoryDataParameter',
  full_name='caffe.MemoryDataParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_size', full_name='caffe.MemoryDataParameter.batch_size', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='channels', full_name='caffe.MemoryDataParameter.channels', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='caffe.MemoryDataParameter.height', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='caffe.MemoryDataParameter.width', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20208,
  serialized_end=20298,
)


_MVNPARAMETER = _descriptor.Descriptor(
  name='MVNParameter',
  full_name='caffe.MVNParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='normalize_variance', full_name='caffe.MVNParameter.normalize_variance', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='across_channels', full_name='caffe.MVNParameter.across_channels', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20300,
  serialized_end=20380,
)


_PERMUTEPARAMETER = _descriptor.Descriptor(
  name='PermuteParameter',
  full_name='caffe.PermuteParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='order', full_name='caffe.PermuteParameter.order', index=0,
      number=1, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20382,
  serialized_end=20415,
)


_POOLINGPARAMETER = _descriptor.Descriptor(
  name='PoolingParameter',
  full_name='caffe.PoolingParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pool', full_name='caffe.PoolingParameter.pool', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad', full_name='caffe.PoolingParameter.pad', index=1,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_h', full_name='caffe.PoolingParameter.pad_h', index=2,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_w', full_name='caffe.PoolingParameter.pad_w', index=3,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_size', full_name='caffe.PoolingParameter.kernel_size', index=4,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_h', full_name='caffe.PoolingParameter.kernel_h', index=5,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernel_w', full_name='caffe.PoolingParameter.kernel_w', index=6,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride', full_name='caffe.PoolingParameter.stride', index=7,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_h', full_name='caffe.PoolingParameter.stride_h', index=8,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride_w', full_name='caffe.PoolingParameter.stride_w', index=9,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='engine', full_name='caffe.PoolingParameter.engine', index=10,
      number=11, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='global_pooling', full_name='caffe.PoolingParameter.global_pooling', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ceil_mode', full_name='caffe.PoolingParameter.ceil_mode', index=12,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _POOLINGPARAMETER_POOLMETHOD,
    _POOLINGPARAMETER_ENGINE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20418,
  serialized_end=20923,
)


_POOLINGSPECIFICPARAMETER = _descriptor.Descriptor(
  name='PoolingSpecificParameter',
  full_name='caffe.PoolingSpecificParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pool', full_name='caffe.PoolingSpecificParameter.pool', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spe_w', full_name='caffe.PoolingSpecificParameter.spe_w', index=1,
      number=2, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spe_h', full_name='caffe.PoolingSpecificParameter.spe_h', index=2,
      number=3, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _POOLINGSPECIFICPARAMETER_POOLSPECIFICMETHOD,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20926,
  serialized_end=21088,
)


_POWERPARAMETER = _descriptor.Descriptor(
  name='PowerParameter',
  full_name='caffe.PowerParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='power', full_name='caffe.PowerParameter.power', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale', full_name='caffe.PowerParameter.scale', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shift', full_name='caffe.PowerParameter.shift', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21090,
  serialized_end=21160,
)


_PRIORBOXPARAMETER = _descriptor.Descriptor(
  name='PriorBoxParameter',
  full_name='caffe.PriorBoxParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='min_size', full_name='caffe.PriorBoxParameter.min_size', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_size', full_name='caffe.PriorBoxParameter.max_size', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aspect_ratio', full_name='caffe.PriorBoxParameter.aspect_ratio', index=2,
      number=3, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='flip', full_name='caffe.PriorBoxParameter.flip', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='clip', full_name='caffe.PriorBoxParameter.clip', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='variance', full_name='caffe.PriorBoxParameter.variance', index=5,
      number=6, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PRIORBOXPARAMETER_CODETYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21163,
  serialized_end=21339,
)


_PYTHONPARAMETER = _descriptor.Descriptor(
  name='PythonParameter',
  full_name='caffe.PythonParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='module', full_name='caffe.PythonParameter.module', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='layer', full_name='caffe.PythonParameter.layer', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21341,
  serialized_end=21389,
)


_RELUPARAMETER = _descriptor.Descriptor(
  name='ReLUParameter',
  full_name='caffe.ReLUParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='negative_slope', full_name='caffe.ReLUParameter.negative_slope', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='engine', full_name='caffe.ReLUParameter.engine', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RELUPARAMETER_ENGINE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21392,
  serialized_end=21533,
)


_ROIPOOLINGPARAMETER = _descriptor.Descriptor(
  name='ROIPoolingParameter',
  full_name='caffe.ROIPoolingParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pooled_h', full_name='caffe.ROIPoolingParameter.pooled_h', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pooled_w', full_name='caffe.ROIPoolingParameter.pooled_w', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spatial_scale', full_name='caffe.ROIPoolingParameter.spatial_scale', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21535,
  serialized_end=21624,
)


_RESHAPEPARAMETER = _descriptor.Descriptor(
  name='ReshapeParameter',
  full_name='caffe.ReshapeParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='shape', full_name='caffe.ReshapeParameter.shape', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.ReshapeParameter.axis', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num_axes', full_name='caffe.ReshapeParameter.num_axes', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21626,
  serialized_end=21716,
)


_SIGMOIDPARAMETER = _descriptor.Descriptor(
  name='SigmoidParameter',
  full_name='caffe.SigmoidParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='engine', full_name='caffe.SigmoidParameter.engine', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SIGMOIDPARAMETER_ENGINE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21718,
  serialized_end=21838,
)


_SLICEPARAMETER = _descriptor.Descriptor(
  name='SliceParameter',
  full_name='caffe.SliceParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.SliceParameter.axis', index=0,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slice_point', full_name='caffe.SliceParameter.slice_point', index=1,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slice_dim', full_name='caffe.SliceParameter.slice_dim', index=2,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21840,
  serialized_end=21916,
)


_SOFTMAXPARAMETER = _descriptor.Descriptor(
  name='SoftmaxParameter',
  full_name='caffe.SoftmaxParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='engine', full_name='caffe.SoftmaxParameter.engine', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.SoftmaxParameter.axis', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SOFTMAXPARAMETER_ENGINE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21919,
  serialized_end=22056,
)


_TANHPARAMETER = _descriptor.Descriptor(
  name='TanHParameter',
  full_name='caffe.TanHParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='engine', full_name='caffe.TanHParameter.engine', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _TANHPARAMETER_ENGINE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22058,
  serialized_end=22172,
)


_THRESHOLDPARAMETER = _descriptor.Descriptor(
  name='ThresholdParameter',
  full_name='caffe.ThresholdParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='threshold', full_name='caffe.ThresholdParameter.threshold', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22174,
  serialized_end=22216,
)


_WINDOWDATAPARAMETER = _descriptor.Descriptor(
  name='WindowDataParameter',
  full_name='caffe.WindowDataParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source', full_name='caffe.WindowDataParameter.source', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale', full_name='caffe.WindowDataParameter.scale', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mean_file', full_name='caffe.WindowDataParameter.mean_file', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='batch_size', full_name='caffe.WindowDataParameter.batch_size', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='crop_size', full_name='caffe.WindowDataParameter.crop_size', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mirror', full_name='caffe.WindowDataParameter.mirror', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fg_threshold', full_name='caffe.WindowDataParameter.fg_threshold', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bg_threshold', full_name='caffe.WindowDataParameter.bg_threshold', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fg_fraction', full_name='caffe.WindowDataParameter.fg_fraction', index=8,
      number=9, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.25),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='context_pad', full_name='caffe.WindowDataParameter.context_pad', index=9,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='crop_mode', full_name='caffe.WindowDataParameter.crop_mode', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b'warp'.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cache_images', full_name='caffe.WindowDataParameter.cache_images', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='root_folder', full_name='caffe.WindowDataParameter.root_folder', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22219,
  serialized_end=22540,
)


_BNPARAMETER = _descriptor.Descriptor(
  name='BNParameter',
  full_name='caffe.BNParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='scale_filler', full_name='caffe.BNParameter.scale_filler', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shift_filler', full_name='caffe.BNParameter.shift_filler', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='var_eps', full_name='caffe.BNParameter.var_eps', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1e-10),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='moving_average', full_name='caffe.BNParameter.moving_average', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='decay', full_name='caffe.BNParameter.decay', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.05),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22543,
  serialized_end=22724,
)


_CTCPARAMETER = _descriptor.Descriptor(
  name='CTCParameter',
  full_name='caffe.CTCParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='threshold', full_name='caffe.CTCParameter.threshold', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.7),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='decode_type', full_name='caffe.CTCParameter.decode_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CTCPARAMETER_DECODER,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22727,
  serialized_end=22892,
)


_PRELUPARAMETER = _descriptor.Descriptor(
  name='PReLUParameter',
  full_name='caffe.PReLUParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='filler', full_name='caffe.PReLUParameter.filler', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='channel_shared', full_name='caffe.PReLUParameter.channel_shared', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quantize_param', full_name='caffe.PReLUParameter.quantize_param', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22895,
  serialized_end=23032,
)


_CROPPARAMETER = _descriptor.Descriptor(
  name='CropParameter',
  full_name='caffe.CropParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='caffe.CropParameter.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='crop_w', full_name='caffe.CropParameter.crop_w', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='crop_h', full_name='caffe.CropParameter.crop_h', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='print_info', full_name='caffe.CropParameter.print_info', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CROPPARAMETER_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23035,
  serialized_end=23196,
)


_AFFINETRANSPARAMETER = _descriptor.Descriptor(
  name='AffineTransParameter',
  full_name='caffe.AffineTransParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='output_h', full_name='caffe.AffineTransParameter.output_h', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='output_w', full_name='caffe.AffineTransParameter.output_w', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='border_value', full_name='caffe.AffineTransParameter.border_value', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='affine_mat', full_name='caffe.AffineTransParameter.affine_mat', index=3,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale', full_name='caffe.AffineTransParameter.scale', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='translate_x', full_name='caffe.AffineTransParameter.translate_x', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='translate_y', full_name='caffe.AffineTransParameter.translate_y', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23199,
  serialized_end=23374,
)


_AFFINETRANSPOINTPARAMETER = _descriptor.Descriptor(
  name='AffineTransPointParameter',
  full_name='caffe.AffineTransPointParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='inverse', full_name='caffe.AffineTransPointParameter.inverse', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='translate_x', full_name='caffe.AffineTransPointParameter.translate_x', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='translate_y', full_name='caffe.AffineTransPointParameter.translate_y', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='affine_mat', full_name='caffe.AffineTransPointParameter.affine_mat', index=3,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale', full_name='caffe.AffineTransPointParameter.scale', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23377,
  serialized_end=23514,
)


_CALCAFFINEMATPARAMETER = _descriptor.Descriptor(
  name='CalcAffineMatParameter',
  full_name='caffe.CalcAffineMatParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='landmark_x', full_name='caffe.CalcAffineMatParameter.landmark_x', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='landmark_y', full_name='caffe.CalcAffineMatParameter.landmark_y', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23516,
  serialized_end=23580,
)


_ROIPARAMETER = _descriptor.Descriptor(
  name='ROIParameter',
  full_name='caffe.ROIParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='crop_h', full_name='caffe.ROIParameter.crop_h', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='crop_w', full_name='caffe.ROIParameter.crop_w', index=1,
      number=2, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='center_x', full_name='caffe.ROIParameter.center_x', index=2,
      number=3, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='center_y', full_name='caffe.ROIParameter.center_y', index=3,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='caffe.ROIParameter.type', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23583,
  serialized_end=23715,
)


_INTERPPARAMETER = _descriptor.Descriptor(
  name='InterpParameter',
  full_name='caffe.InterpParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='height', full_name='caffe.InterpParameter.height', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='caffe.InterpParameter.width', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='zoom_factor', full_name='caffe.InterpParameter.zoom_factor', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shrink_factor', full_name='caffe.InterpParameter.shrink_factor', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_beg', full_name='caffe.InterpParameter.pad_beg', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_end', full_name='caffe.InterpParameter.pad_end', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='align_corners', full_name='caffe.InterpParameter.align_corners', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23718,
  serialized_end=23891,
)


_RECURRENTPARAMETER = _descriptor.Descriptor(
  name='RecurrentParameter',
  full_name='caffe.RecurrentParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_output', full_name='caffe.RecurrentParameter.num_output', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_filler', full_name='caffe.RecurrentParameter.weight_filler', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_filler', full_name='caffe.RecurrentParameter.bias_filler', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='debug_info', full_name='caffe.RecurrentParameter.debug_info', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23894,
  serialized_end=24056,
)


_FLATTENPARAMETER = _descriptor.Descriptor(
  name='FlattenParameter',
  full_name='caffe.FlattenParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.FlattenParameter.axis', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='end_axis', full_name='caffe.FlattenParameter.end_axis', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24058,
  serialized_end=24115,
)


_BATCHNORMPARAMETER = _descriptor.Descriptor(
  name='BatchNormParameter',
  full_name='caffe.BatchNormParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='use_global_stats', full_name='caffe.BatchNormParameter.use_global_stats', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='moving_average_fraction', full_name='caffe.BatchNormParameter.moving_average_fraction', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.999),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eps', full_name='caffe.BatchNormParameter.eps', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1e-05),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24117,
  serialized_end=24223,
)


_V1LAYERPARAMETER = _descriptor.Descriptor(
  name='V1LayerParameter',
  full_name='caffe.V1LayerParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bottom', full_name='caffe.V1LayerParameter.bottom', index=0,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='top', full_name='caffe.V1LayerParameter.top', index=1,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='caffe.V1LayerParameter.name', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='include', full_name='caffe.V1LayerParameter.include', index=3,
      number=32, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exclude', full_name='caffe.V1LayerParameter.exclude', index=4,
      number=33, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='caffe.V1LayerParameter.type', index=5,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blobs', full_name='caffe.V1LayerParameter.blobs', index=6,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='param', full_name='caffe.V1LayerParameter.param', index=7,
      number=1001, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blob_share_mode', full_name='caffe.V1LayerParameter.blob_share_mode', index=8,
      number=1002, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blobs_lr', full_name='caffe.V1LayerParameter.blobs_lr', index=9,
      number=7, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_decay', full_name='caffe.V1LayerParameter.weight_decay', index=10,
      number=8, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='loss_weight', full_name='caffe.V1LayerParameter.loss_weight', index=11,
      number=35, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='accuracy_param', full_name='caffe.V1LayerParameter.accuracy_param', index=12,
      number=27, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='argmax_param', full_name='caffe.V1LayerParameter.argmax_param', index=13,
      number=23, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='concat_param', full_name='caffe.V1LayerParameter.concat_param', index=14,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contrastive_loss_param', full_name='caffe.V1LayerParameter.contrastive_loss_param', index=15,
      number=40, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='convolution_param', full_name='caffe.V1LayerParameter.convolution_param', index=16,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_param', full_name='caffe.V1LayerParameter.data_param', index=17,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dropout_param', full_name='caffe.V1LayerParameter.dropout_param', index=18,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dummy_data_param', full_name='caffe.V1LayerParameter.dummy_data_param', index=19,
      number=26, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eltwise_param', full_name='caffe.V1LayerParameter.eltwise_param', index=20,
      number=24, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exp_param', full_name='caffe.V1LayerParameter.exp_param', index=21,
      number=41, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hdf5_data_param', full_name='caffe.V1LayerParameter.hdf5_data_param', index=22,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hdf5_output_param', full_name='caffe.V1LayerParameter.hdf5_output_param', index=23,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hinge_loss_param', full_name='caffe.V1LayerParameter.hinge_loss_param', index=24,
      number=29, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='image_data_param', full_name='caffe.V1LayerParameter.image_data_param', index=25,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='infogain_loss_param', full_name='caffe.V1LayerParameter.infogain_loss_param', index=26,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inner_product_param', full_name='caffe.V1LayerParameter.inner_product_param', index=27,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lrn_param', full_name='caffe.V1LayerParameter.lrn_param', index=28,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='memory_data_param', full_name='caffe.V1LayerParameter.memory_data_param', index=29,
      number=22, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mvn_param', full_name='caffe.V1LayerParameter.mvn_param', index=30,
      number=34, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pooling_param', full_name='caffe.V1LayerParameter.pooling_param', index=31,
      number=19, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='power_param', full_name='caffe.V1LayerParameter.power_param', index=32,
      number=21, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='relu_param', full_name='caffe.V1LayerParameter.relu_param', index=33,
      number=30, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sigmoid_param', full_name='caffe.V1LayerParameter.sigmoid_param', index=34,
      number=38, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='softmax_param', full_name='caffe.V1LayerParameter.softmax_param', index=35,
      number=39, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slice_param', full_name='caffe.V1LayerParameter.slice_param', index=36,
      number=31, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tanh_param', full_name='caffe.V1LayerParameter.tanh_param', index=37,
      number=37, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='threshold_param', full_name='caffe.V1LayerParameter.threshold_param', index=38,
      number=25, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='window_data_param', full_name='caffe.V1LayerParameter.window_data_param', index=39,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transform_param', full_name='caffe.V1LayerParameter.transform_param', index=40,
      number=36, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='loss_param', full_name='caffe.V1LayerParameter.loss_param', index=41,
      number=42, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='layer', full_name='caffe.V1LayerParameter.layer', index=42,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _V1LAYERPARAMETER_LAYERTYPE,
    _V1LAYERPARAMETER_DIMCHECKMODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24226,
  serialized_end=26754,
)


_V0LAYERPARAMETER = _descriptor.Descriptor(
  name='V0LayerParameter',
  full_name='caffe.V0LayerParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='caffe.V0LayerParameter.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='caffe.V0LayerParameter.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='num_output', full_name='caffe.V0LayerParameter.num_output', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='biasterm', full_name='caffe.V0LayerParameter.biasterm', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_filler', full_name='caffe.V0LayerParameter.weight_filler', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_filler', full_name='caffe.V0LayerParameter.bias_filler', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad', full_name='caffe.V0LayerParameter.pad', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kernelsize', full_name='caffe.V0LayerParameter.kernelsize', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='group', full_name='caffe.V0LayerParameter.group', index=8,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stride', full_name='caffe.V0LayerParameter.stride', index=9,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pool', full_name='caffe.V0LayerParameter.pool', index=10,
      number=11, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dropout_ratio', full_name='caffe.V0LayerParameter.dropout_ratio', index=11,
      number=12, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='local_size', full_name='caffe.V0LayerParameter.local_size', index=12,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=5,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='alpha', full_name='caffe.V0LayerParameter.alpha', index=13,
      number=14, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='beta', full_name='caffe.V0LayerParameter.beta', index=14,
      number=15, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.75),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='k', full_name='caffe.V0LayerParameter.k', index=15,
      number=22, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='source', full_name='caffe.V0LayerParameter.source', index=16,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale', full_name='caffe.V0LayerParameter.scale', index=17,
      number=17, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='meanfile', full_name='caffe.V0LayerParameter.meanfile', index=18,
      number=18, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b''.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='batchsize', full_name='caffe.V0LayerParameter.batchsize', index=19,
      number=19, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cropsize', full_name='caffe.V0LayerParameter.cropsize', index=20,
      number=20, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mirror', full_name='caffe.V0LayerParameter.mirror', index=21,
      number=21, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blobs', full_name='caffe.V0LayerParameter.blobs', index=22,
      number=50, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blobs_lr', full_name='caffe.V0LayerParameter.blobs_lr', index=23,
      number=51, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='weight_decay', full_name='caffe.V0LayerParameter.weight_decay', index=24,
      number=52, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rand_skip', full_name='caffe.V0LayerParameter.rand_skip', index=25,
      number=53, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='det_fg_threshold', full_name='caffe.V0LayerParameter.det_fg_threshold', index=26,
      number=54, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='det_bg_threshold', full_name='caffe.V0LayerParameter.det_bg_threshold', index=27,
      number=55, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='det_fg_fraction', full_name='caffe.V0LayerParameter.det_fg_fraction', index=28,
      number=56, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.25),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='det_context_pad', full_name='caffe.V0LayerParameter.det_context_pad', index=29,
      number=58, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='det_crop_mode', full_name='caffe.V0LayerParameter.det_crop_mode', index=30,
      number=59, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b'warp'.decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_num', full_name='caffe.V0LayerParameter.new_num', index=31,
      number=60, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_channels', full_name='caffe.V0LayerParameter.new_channels', index=32,
      number=61, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_height', full_name='caffe.V0LayerParameter.new_height', index=33,
      number=62, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='new_width', full_name='caffe.V0LayerParameter.new_width', index=34,
      number=63, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shuffle_images', full_name='caffe.V0LayerParameter.shuffle_images', index=35,
      number=64, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='concat_dim', full_name='caffe.V0LayerParameter.concat_dim', index=36,
      number=65, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ntile_width', full_name='caffe.V0LayerParameter.ntile_width', index=37,
      number=72, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ntile_height', full_name='caffe.V0LayerParameter.ntile_height', index=38,
      number=73, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hdf5_output_param', full_name='caffe.V0LayerParameter.hdf5_output_param', index=39,
      number=1001, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _V0LAYERPARAMETER_POOLMETHOD,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26757,
  serialized_end=27827,
)


_CHANNELSHUFFLEPARAMETER = _descriptor.Descriptor(
  name='ChannelShuffleParameter',
  full_name='caffe.ChannelShuffleParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='group', full_name='caffe.ChannelShuffleParameter.group', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=27829,
  serialized_end=27872,
)


_MOVINGAVGPARAMETER = _descriptor.Descriptor(
  name='MovingAvgParameter',
  full_name='caffe.MovingAvgParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='decay', full_name='caffe.MovingAvgParameter.decay', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=27874,
  serialized_end=27912,
)


_SLICINGPARAMETER = _descriptor.Descriptor(
  name='SlicingParameter',
  full_name='caffe.SlicingParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='coefficient_len', full_name='caffe.SlicingParameter.coefficient_len', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='depth_d', full_name='caffe.SlicingParameter.depth_d', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale_x', full_name='caffe.SlicingParameter.scale_x', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(-1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scale_y', full_name='caffe.SlicingParameter.scale_y', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(-1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offset_x', full_name='caffe.SlicingParameter.offset_x', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offset_y', full_name='caffe.SlicingParameter.offset_y', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offset_z', full_name='caffe.SlicingParameter.offset_z', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=27915,
  serialized_end=28087,
)


_BILATERALSLICINGPARAMETER = _descriptor.Descriptor(
  name='BilateralSlicingParameter',
  full_name='caffe.BilateralSlicingParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='coefficient_len', full_name='caffe.BilateralSlicingParameter.coefficient_len', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28089,
  serialized_end=28144,
)


_BILATERALSLICEAPPLYPARAMETER = _descriptor.Descriptor(
  name='BilateralSliceApplyParameter',
  full_name='caffe.BilateralSliceApplyParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='coefficient_len', full_name='caffe.BilateralSliceApplyParameter.coefficient_len', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='has_offset', full_name='caffe.BilateralSliceApplyParameter.has_offset', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28146,
  serialized_end=28230,
)


_ADAPTIVEBILATERALSLICEAPPLYPARAMETER = _descriptor.Descriptor(
  name='AdaptiveBilateralSliceApplyParameter',
  full_name='caffe.AdaptiveBilateralSliceApplyParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='coefficient_len', full_name='caffe.AdaptiveBilateralSliceApplyParameter.coefficient_len', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='has_offset', full_name='caffe.AdaptiveBilateralSliceApplyParameter.has_offset', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='depth_d', full_name='caffe.AdaptiveBilateralSliceApplyParameter.depth_d', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28232,
  serialized_end=28344,
)


_POINTSCURVEPARAMETER = _descriptor.Descriptor(
  name='PointsCurveParameter',
  full_name='caffe.PointsCurveParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='numpoints', full_name='caffe.PointsCurveParameter.numpoints', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=16,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slope_filler', full_name='caffe.PointsCurveParameter.slope_filler', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bias_filler', full_name='caffe.PointsCurveParameter.bias_filler', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28347,
  serialized_end=28483,
)


_SUBPIXELDOWNPARAMETER = _descriptor.Descriptor(
  name='SubpixelDownParameter',
  full_name='caffe.SubpixelDownParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='downsample', full_name='caffe.SubpixelDownParameter.downsample', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28485,
  serialized_end=28531,
)


_SUBPIXELUPPARAMETER = _descriptor.Descriptor(
  name='SubpixelUpParameter',
  full_name='caffe.SubpixelUpParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='upsample', full_name='caffe.SubpixelUpParameter.upsample', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='output_datatype', full_name='caffe.SubpixelUpParameter.output_datatype', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mode', full_name='caffe.SubpixelUpParameter.mode', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SUBPIXELUPPARAMETER_DATATYPE,
    _SUBPIXELUPPARAMETER_MODETYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28534,
  serialized_end=28779,
)


_CLIPPARAMETER = _descriptor.Descriptor(
  name='ClipParameter',
  full_name='caffe.ClipParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='min', full_name='caffe.ClipParameter.min', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max', full_name='caffe.ClipParameter.max', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28781,
  serialized_end=28822,
)


_ELTWISEAFFINETRANSFORMPARAMETER = _descriptor.Descriptor(
  name='EltwiseAffineTransformParameter',
  full_name='caffe.EltwiseAffineTransformParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28824,
  serialized_end=28857,
)


_TILEPARAMETER = _descriptor.Descriptor(
  name='TileParameter',
  full_name='caffe.TileParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.TileParameter.axis', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tiles', full_name='caffe.TileParameter.tiles', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28859,
  serialized_end=28909,
)


_REDUCEPARAMETER = _descriptor.Descriptor(
  name='ReduceParameter',
  full_name='caffe.ReduceParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.ReduceParameter.axis', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mode', full_name='caffe.ReduceParameter.mode', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _REDUCEPARAMETER_REDUCEOP,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28911,
  serialized_end=28984,
)


_PADPARAMETER = _descriptor.Descriptor(
  name='PadParameter',
  full_name='caffe.PadParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pad', full_name='caffe.PadParameter.pad', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mode', full_name='caffe.PadParameter.mode', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_w', full_name='caffe.PadParameter.pad_w', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_h', full_name='caffe.PadParameter.pad_h', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PADPARAMETER_PADMODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28987,
  serialized_end=29135,
)


_RELU6PARAMETER = _descriptor.Descriptor(
  name='ReLU6Parameter',
  full_name='caffe.ReLU6Parameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='negative_slope', full_name='caffe.ReLU6Parameter.negative_slope', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29137,
  serialized_end=29180,
)


_CORRELATION2DPARAMETER = _descriptor.Descriptor(
  name='Correlation2DParameter',
  full_name='caffe.Correlation2DParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='groups', full_name='caffe.Correlation2DParameter.groups', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29182,
  serialized_end=29225,
)


_MATMULPARAMETER = _descriptor.Descriptor(
  name='MatMulParameter',
  full_name='caffe.MatMulParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mode', full_name='caffe.MatMulParameter.mode', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MATMULPARAMETER_MATMULMODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29227,
  serialized_end=29343,
)


_PARAMETERPARAMETER = _descriptor.Descriptor(
  name='ParameterParameter',
  full_name='caffe.ParameterParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch', full_name='caffe.ParameterParameter.batch', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='m', full_name='caffe.ParameterParameter.m', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='n', full_name='caffe.ParameterParameter.n', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='channel', full_name='caffe.ParameterParameter.channel', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='caffe.ParameterParameter.height', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='caffe.ParameterParameter.width', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=-1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29346,
  serialized_end=29474,
)


_PIXELSHUFFLEPARAMETER = _descriptor.Descriptor(
  name='PixelShuffleParameter',
  full_name='caffe.PixelShuffleParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='upscale_factor', full_name='caffe.PixelShuffleParameter.upscale_factor', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29476,
  serialized_end=29526,
)


_ROITRANSFORMPARAMETER = _descriptor.Descriptor(
  name='ROITransformParameter',
  full_name='caffe.ROITransformParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='roi_pooled_h', full_name='caffe.ROITransformParameter.roi_pooled_h', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_pooled_w', full_name='caffe.ROITransformParameter.roi_pooled_w', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29528,
  serialized_end=29595,
)


_MAINTRANSFORMPARAMETER = _descriptor.Descriptor(
  name='MainTransformParameter',
  full_name='caffe.MainTransformParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='roi_pooled_h', full_name='caffe.MainTransformParameter.roi_pooled_h', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='roi_pooled_w', full_name='caffe.MainTransformParameter.roi_pooled_w', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29597,
  serialized_end=29665,
)


_INSTANCENORMPARAMETER = _descriptor.Descriptor(
  name='InstanceNormParameter',
  full_name='caffe.InstanceNormParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_features', full_name='caffe.InstanceNormParameter.num_features', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='eps', full_name='caffe.InstanceNormParameter.eps', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1e-05),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='affine', full_name='caffe.InstanceNormParameter.affine', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29667,
  serialized_end=29755,
)


_GRIDSAMPLEPARAMETER = _descriptor.Descriptor(
  name='GridSampleParameter',
  full_name='caffe.GridSampleParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='padding_mode', full_name='caffe.GridSampleParameter.padding_mode', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _GRIDSAMPLEPARAMETER_PADDINGMODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29758,
  serialized_end=29886,
)


_GRIDSAMPLE3DPARAMETER = _descriptor.Descriptor(
  name='GridSample3DParameter',
  full_name='caffe.GridSample3DParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='padding_mode', full_name='caffe.GridSample3DParameter.padding_mode', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='align_corners', full_name='caffe.GridSample3DParameter.align_corners', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='output_channel', full_name='caffe.GridSample3DParameter.output_channel', index=2,
      number=3, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _GRIDSAMPLE3DPARAMETER_PADDINGMODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29889,
  serialized_end=30075,
)


_REDUCEL2PARAMETER = _descriptor.Descriptor(
  name='ReduceL2Parameter',
  full_name='caffe.ReduceL2Parameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='axes', full_name='caffe.ReduceL2Parameter.axes', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=4,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keepdims', full_name='caffe.ReduceL2Parameter.keepdims', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30077,
  serialized_end=30134,
)


_VARIANCEPARAMETER = _descriptor.Descriptor(
  name='VarianceParameter',
  full_name='caffe.VarianceParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dim', full_name='caffe.VarianceParameter.dim', index=0,
      number=1, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keepdim', full_name='caffe.VarianceParameter.keepdim', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30136,
  serialized_end=30191,
)


_MEANPARAMETER = _descriptor.Descriptor(
  name='MeanParameter',
  full_name='caffe.MeanParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dim', full_name='caffe.MeanParameter.dim', index=0,
      number=1, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keepdim', full_name='caffe.MeanParameter.keepdim', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30193,
  serialized_end=30244,
)


_INTERP3DPARAMETER = _descriptor.Descriptor(
  name='Interp3dParameter',
  full_name='caffe.Interp3dParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='depth', full_name='caffe.Interp3dParameter.depth', index=0,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='height', full_name='caffe.Interp3dParameter.height', index=1,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='width', full_name='caffe.Interp3dParameter.width', index=2,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='zoom_factor', full_name='caffe.Interp3dParameter.zoom_factor', index=3,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shrink_factor', full_name='caffe.Interp3dParameter.shrink_factor', index=4,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_beg', full_name='caffe.Interp3dParameter.pad_beg', index=5,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pad_end', full_name='caffe.Interp3dParameter.pad_end', index=6,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='align_corners', full_name='caffe.Interp3dParameter.align_corners', index=7,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30247,
  serialized_end=30440,
)


_GRUPARAMETER = _descriptor.Descriptor(
  name='GRUParameter',
  full_name='caffe.GRUParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hidden_size', full_name='caffe.GRUParameter.hidden_size', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bidirectional', full_name='caffe.GRUParameter.bidirectional', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30442,
  serialized_end=30507,
)


_CORRELATIONMIGPARAMETER = _descriptor.Descriptor(
  name='CorrelationMigParameter',
  full_name='caffe.CorrelationMigParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='shift_step', full_name='caffe.CorrelationMigParameter.shift_step', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=9,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30509,
  serialized_end=30557,
)


_ARGSORTPARAMETER = _descriptor.Descriptor(
  name='ArgSortParameter',
  full_name='caffe.ArgSortParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='descending', full_name='caffe.ArgSortParameter.descending', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.ArgSortParameter.axis', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30559,
  serialized_end=30618,
)


_CUMPRODPARAMETER = _descriptor.Descriptor(
  name='CumProdParameter',
  full_name='caffe.CumProdParameter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='axis', full_name='caffe.CumProdParameter.axis', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30620,
  serialized_end=30652,
)

_BLOBPROTO.fields_by_name['shape'].message_type = _BLOBSHAPE
_BLOBPROTOVECTOR.fields_by_name['blobs'].message_type = _BLOBPROTO
_NETPARAMETER.fields_by_name['input_shape'].message_type = _BLOBSHAPE
_NETPARAMETER.fields_by_name['state'].message_type = _NETSTATE
_NETPARAMETER.fields_by_name['layer'].message_type = _LAYERPARAMETER
_NETPARAMETER.fields_by_name['layers'].message_type = _V1LAYERPARAMETER
_SOLVERPARAMETER.fields_by_name['net_param'].message_type = _NETPARAMETER
_SOLVERPARAMETER.fields_by_name['train_net_param'].message_type = _NETPARAMETER
_SOLVERPARAMETER.fields_by_name['test_net_param'].message_type = _NETPARAMETER
_SOLVERPARAMETER.fields_by_name['train_state'].message_type = _NETSTATE
_SOLVERPARAMETER.fields_by_name['test_state'].message_type = _NETSTATE
_SOLVERPARAMETER.fields_by_name['solver_mode'].enum_type = _SOLVERPARAMETER_SOLVERMODE
_SOLVERPARAMETER.fields_by_name['solver_type'].enum_type = _SOLVERPARAMETER_SOLVERTYPE
_SOLVERPARAMETER_SOLVERMODE.containing_type = _SOLVERPARAMETER
_SOLVERPARAMETER_SOLVERTYPE.containing_type = _SOLVERPARAMETER
_SOLVERSTATE.fields_by_name['history'].message_type = _BLOBPROTO
_NETSTATE.fields_by_name['phase'].enum_type = _PHASE
_NETSTATERULE.fields_by_name['phase'].enum_type = _PHASE
_PARAMSPEC.fields_by_name['share_mode'].enum_type = _PARAMSPEC_DIMCHECKMODE
_PARAMSPEC_DIMCHECKMODE.containing_type = _PARAMSPEC
_LAYERPARAMETER.fields_by_name['phase'].enum_type = _PHASE
_LAYERPARAMETER.fields_by_name['param'].message_type = _PARAMSPEC
_LAYERPARAMETER.fields_by_name['blobs'].message_type = _BLOBPROTO
_LAYERPARAMETER.fields_by_name['include'].message_type = _NETSTATERULE
_LAYERPARAMETER.fields_by_name['exclude'].message_type = _NETSTATERULE
_LAYERPARAMETER.fields_by_name['forward_precision'].enum_type = _LAYERPARAMETER_PRECISIONTYPE
_LAYERPARAMETER.fields_by_name['transform_param'].message_type = _TRANSFORMATIONPARAMETER
_LAYERPARAMETER.fields_by_name['loss_param'].message_type = _LOSSPARAMETER
_LAYERPARAMETER.fields_by_name['accuracy_param'].message_type = _ACCURACYPARAMETER
_LAYERPARAMETER.fields_by_name['argmax_param'].message_type = _ARGMAXPARAMETER
_LAYERPARAMETER.fields_by_name['concat_param'].message_type = _CONCATPARAMETER
_LAYERPARAMETER.fields_by_name['contrastive_loss_param'].message_type = _CONTRASTIVELOSSPARAMETER
_LAYERPARAMETER.fields_by_name['convolution_param'].message_type = _CONVOLUTIONPARAMETER
_LAYERPARAMETER.fields_by_name['data_param'].message_type = _DATAPARAMETER
_LAYERPARAMETER.fields_by_name['dropout_param'].message_type = _DROPOUTPARAMETER
_LAYERPARAMETER.fields_by_name['dummy_data_param'].message_type = _DUMMYDATAPARAMETER
_LAYERPARAMETER.fields_by_name['eltwise_param'].message_type = _ELTWISEPARAMETER
_LAYERPARAMETER.fields_by_name['exp_param'].message_type = _EXPPARAMETER
_LAYERPARAMETER.fields_by_name['hdf5_data_param'].message_type = _HDF5DATAPARAMETER
_LAYERPARAMETER.fields_by_name['hdf5_output_param'].message_type = _HDF5OUTPUTPARAMETER
_LAYERPARAMETER.fields_by_name['hinge_loss_param'].message_type = _HINGELOSSPARAMETER
_LAYERPARAMETER.fields_by_name['image_data_param'].message_type = _IMAGEDATAPARAMETER
_LAYERPARAMETER.fields_by_name['infogain_loss_param'].message_type = _INFOGAINLOSSPARAMETER
_LAYERPARAMETER.fields_by_name['inner_product_param'].message_type = _INNERPRODUCTPARAMETER
_LAYERPARAMETER.fields_by_name['lrn_param'].message_type = _LRNPARAMETER
_LAYERPARAMETER.fields_by_name['memory_data_param'].message_type = _MEMORYDATAPARAMETER
_LAYERPARAMETER.fields_by_name['mvn_param'].message_type = _MVNPARAMETER
_LAYERPARAMETER.fields_by_name['pooling_param'].message_type = _POOLINGPARAMETER
_LAYERPARAMETER.fields_by_name['pooling_specific_param'].message_type = _POOLINGSPECIFICPARAMETER
_LAYERPARAMETER.fields_by_name['power_param'].message_type = _POWERPARAMETER
_LAYERPARAMETER.fields_by_name['relu_param'].message_type = _RELUPARAMETER
_LAYERPARAMETER.fields_by_name['sigmoid_param'].message_type = _SIGMOIDPARAMETER
_LAYERPARAMETER.fields_by_name['softmax_param'].message_type = _SOFTMAXPARAMETER
_LAYERPARAMETER.fields_by_name['slice_param'].message_type = _SLICEPARAMETER
_LAYERPARAMETER.fields_by_name['tanh_param'].message_type = _TANHPARAMETER
_LAYERPARAMETER.fields_by_name['threshold_param'].message_type = _THRESHOLDPARAMETER
_LAYERPARAMETER.fields_by_name['window_data_param'].message_type = _WINDOWDATAPARAMETER
_LAYERPARAMETER.fields_by_name['python_param'].message_type = _PYTHONPARAMETER
_LAYERPARAMETER.fields_by_name['ctc_param'].message_type = _CTCPARAMETER
_LAYERPARAMETER.fields_by_name['prelu_param'].message_type = _PRELUPARAMETER
_LAYERPARAMETER.fields_by_name['reshape_param'].message_type = _RESHAPEPARAMETER
_LAYERPARAMETER.fields_by_name['roi_pooling_param'].message_type = _ROIPOOLINGPARAMETER
_LAYERPARAMETER.fields_by_name['affine_trans_param'].message_type = _AFFINETRANSPARAMETER
_LAYERPARAMETER.fields_by_name['roi_param'].message_type = _ROIPARAMETER
_LAYERPARAMETER.fields_by_name['affine_trans_point_param'].message_type = _AFFINETRANSPOINTPARAMETER
_LAYERPARAMETER.fields_by_name['permute_param'].message_type = _PERMUTEPARAMETER
_LAYERPARAMETER.fields_by_name['prior_box_param'].message_type = _PRIORBOXPARAMETER
_LAYERPARAMETER.fields_by_name['calc_affine_mat_param'].message_type = _CALCAFFINEMATPARAMETER
_LAYERPARAMETER.fields_by_name['interp_param'].message_type = _INTERPPARAMETER
_LAYERPARAMETER.fields_by_name['detection_output_param'].message_type = _DETECTIONOUTPUTPARAMETER
_LAYERPARAMETER.fields_by_name['relu6_param'].message_type = _RELU6PARAMETER
_LAYERPARAMETER.fields_by_name['bias_param'].message_type = _BIASPARAMETER
_LAYERPARAMETER.fields_by_name['zxybn_param'].message_type = _ZXYBNPARAMETER
_LAYERPARAMETER.fields_by_name['correlation_param'].message_type = _CORRELATIONPARAMETER
_LAYERPARAMETER.fields_by_name['psroi_pooling_param'].message_type = _PSROIPOOLINGPARAMETER
_LAYERPARAMETER.fields_by_name['nn_upsample_param'].message_type = _NNUPSAMPLEPARAMETER
_LAYERPARAMETER.fields_by_name['roi_align_param'].message_type = _ROIALIGNPARAMETER
_LAYERPARAMETER.fields_by_name['rpn_proposal_param'].message_type = _RPNPROPOSALLAYERPARAMETER
_LAYERPARAMETER.fields_by_name['roi_mask_pooling_param'].message_type = _ROIMASKPOOLINGPARAMETER
_LAYERPARAMETER.fields_by_name['channel_shuffle_param'].message_type = _CHANNELSHUFFLEPARAMETER
_LAYERPARAMETER.fields_by_name['bn_param'].message_type = _BNPARAMETER
_LAYERPARAMETER.fields_by_name['roi_align_pooling_pod_param'].message_type = _ROIALIGNPOOLINGPODPARAMETER
_LAYERPARAMETER.fields_by_name['roi_align_pooling_param'].message_type = _ROIALIGNPOOLINGPARAMETER
_LAYERPARAMETER.fields_by_name['psroi_align_pooling_param'].message_type = _PSROIALIGNPOOLINGPARAMETER
_LAYERPARAMETER.fields_by_name['alignment_to_roi_param'].message_type = _ALIGNMENTTOROIPARAMETER
_LAYERPARAMETER.fields_by_name['unpooling_param'].message_type = _UNPOOLINGPARAMETER
_LAYERPARAMETER.fields_by_name['log_param'].message_type = _LOGPARAMETER
_LAYERPARAMETER.fields_by_name['reverse_param'].message_type = _REVERSEPARAMETER
_LAYERPARAMETER.fields_by_name['lstm_param'].message_type = _LSTMPARAMETER
_LAYERPARAMETER.fields_by_name['prior_vbox_param'].message_type = _PRIORVBOXPARAMETER
_LAYERPARAMETER.fields_by_name['btcostvolume_param'].message_type = _BTCOSTVOLUMEPARAMETER
_LAYERPARAMETER.fields_by_name['moving_avg_param'].message_type = _MOVINGAVGPARAMETER
_LAYERPARAMETER.fields_by_name['slicing_param'].message_type = _SLICINGPARAMETER
_LAYERPARAMETER.fields_by_name['bilateral_slicing_param'].message_type = _BILATERALSLICINGPARAMETER
_LAYERPARAMETER.fields_by_name['bilateral_slice_apply_param'].message_type = _BILATERALSLICEAPPLYPARAMETER
_LAYERPARAMETER.fields_by_name['adaptive_bilateral_slice_apply_param'].message_type = _ADAPTIVEBILATERALSLICEAPPLYPARAMETER
_LAYERPARAMETER.fields_by_name['pointscurve_param'].message_type = _POINTSCURVEPARAMETER
_LAYERPARAMETER.fields_by_name['subpixel_down_param'].message_type = _SUBPIXELDOWNPARAMETER
_LAYERPARAMETER.fields_by_name['subpixel_up_param'].message_type = _SUBPIXELUPPARAMETER
_LAYERPARAMETER.fields_by_name['clip_param'].message_type = _CLIPPARAMETER
_LAYERPARAMETER.fields_by_name['eltwise_affine_transform_param'].message_type = _ELTWISEAFFINETRANSFORMPARAMETER
_LAYERPARAMETER.fields_by_name['tile_param'].message_type = _TILEPARAMETER
_LAYERPARAMETER.fields_by_name['pad_param'].message_type = _PADPARAMETER
_LAYERPARAMETER.fields_by_name['reduce_param'].message_type = _REDUCEPARAMETER
_LAYERPARAMETER.fields_by_name['convolution3d_param'].message_type = _CONVOLUTION3DPARAMETER
_LAYERPARAMETER.fields_by_name['deconvolution3d_param'].message_type = _DECONVOLUTION3DPARAMETER
_LAYERPARAMETER.fields_by_name['relu3d_param'].message_type = _RELU3DPARAMETER
_LAYERPARAMETER.fields_by_name['prelu3d_param'].message_type = _PRELUPARAMETER
_LAYERPARAMETER.fields_by_name['leaky_relu3d_param'].message_type = _RELU3DPARAMETER
_LAYERPARAMETER.fields_by_name['elu3d_param'].message_type = _RELU3DPARAMETER
_LAYERPARAMETER.fields_by_name['groupnorm3d_param'].message_type = _GROUPNORM3DPARAMETER
_LAYERPARAMETER.fields_by_name['batchnorm3d_param'].message_type = _BATCHNORM3DPARAMETER
_LAYERPARAMETER.fields_by_name['maxpooling3d_param'].message_type = _POOLING3DPARAMETER
_LAYERPARAMETER.fields_by_name['avepooling3d_param'].message_type = _POOLING3DPARAMETER
_LAYERPARAMETER.fields_by_name['dropout3d_param'].message_type = _DROPOUTPARAMETER
_LAYERPARAMETER.fields_by_name['interp3d_param'].message_type = _INTERP3DPARAMETER
_LAYERPARAMETER.fields_by_name['pooling3d_param'].message_type = _POOLING3DPARAMETER
_LAYERPARAMETER.fields_by_name['reshape3d_param'].message_type = _RESHAPEPARAMETER
_LAYERPARAMETER.fields_by_name['softmax3d_param'].message_type = _SOFTMAXPARAMETER
_LAYERPARAMETER.fields_by_name['argmax3d_param'].message_type = _ARGMAXPARAMETER
_LAYERPARAMETER.fields_by_name['inner_product3d_param'].message_type = _INNERPRODUCTPARAMETER
_LAYERPARAMETER.fields_by_name['slice3d_param'].message_type = _SLICEPARAMETER
_LAYERPARAMETER.fields_by_name['concat3d_param'].message_type = _CONCATPARAMETER
_LAYERPARAMETER.fields_by_name['bn3d_param'].message_type = _BNPARAMETER
_LAYERPARAMETER.fields_by_name['matmul3d_param'].message_type = _MATMULPARAMETER
_LAYERPARAMETER.fields_by_name['transpose3d_param'].message_type = _TRANSPOSEPARAMETER
_LAYERPARAMETER.fields_by_name['scale_param'].message_type = _SCALEPARAMETER
_LAYERPARAMETER.fields_by_name['crop_param'].message_type = _CROPPARAMETER
_LAYERPARAMETER.fields_by_name['recurrent_param'].message_type = _RECURRENTPARAMETER
_LAYERPARAMETER.fields_by_name['batch_norm_param'].message_type = _BATCHNORMPARAMETER
_LAYERPARAMETER.fields_by_name['flatten_param'].message_type = _FLATTENPARAMETER
_LAYERPARAMETER.fields_by_name['scales_param'].message_type = _SCALESPARAMETER
_LAYERPARAMETER.fields_by_name['transpose_param'].message_type = _TRANSPOSEPARAMETER
_LAYERPARAMETER.fields_by_name['heatmap_param'].message_type = _HEATMAP2COORDPARAMETER
_LAYERPARAMETER.fields_by_name['normalize_param'].message_type = _NORMALIZEPARAMETER
_LAYERPARAMETER.fields_by_name['correlation2d_param'].message_type = _CORRELATION2DPARAMETER
_LAYERPARAMETER.fields_by_name['matmul_param'].message_type = _MATMULPARAMETER
_LAYERPARAMETER.fields_by_name['parameter_param'].message_type = _PARAMETERPARAMETER
_LAYERPARAMETER.fields_by_name['pixelshuffle_param'].message_type = _PIXELSHUFFLEPARAMETER
_LAYERPARAMETER.fields_by_name['roi_transform_param'].message_type = _ROITRANSFORMPARAMETER
_LAYERPARAMETER.fields_by_name['instance_norm_param'].message_type = _INSTANCENORMPARAMETER
_LAYERPARAMETER.fields_by_name['grid_sample_param'].message_type = _GRIDSAMPLEPARAMETER
_LAYERPARAMETER.fields_by_name['reducel2_param'].message_type = _REDUCEL2PARAMETER
_LAYERPARAMETER.fields_by_name['mean_param'].message_type = _MEANPARAMETER
_LAYERPARAMETER.fields_by_name['variance_param'].message_type = _VARIANCEPARAMETER
_LAYERPARAMETER.fields_by_name['grid_sample_3d_param'].message_type = _GRIDSAMPLE3DPARAMETER
_LAYERPARAMETER.fields_by_name['gru_param'].message_type = _GRUPARAMETER
_LAYERPARAMETER.fields_by_name['correlationmig_param'].message_type = _CORRELATIONMIGPARAMETER
_LAYERPARAMETER.fields_by_name['argsort_param'].message_type = _ARGSORTPARAMETER
_LAYERPARAMETER.fields_by_name['cumprod_param'].message_type = _CUMPRODPARAMETER
_LAYERPARAMETER.fields_by_name['main_transform_param'].message_type = _MAINTRANSFORMPARAMETER
_LAYERPARAMETER.fields_by_name['quantize_param'].message_type = _QUANTIZEPARAMETER
_LAYERPARAMETER_PRECISIONTYPE.containing_type = _LAYERPARAMETER
_POOLING3DPARAMETER.fields_by_name['pool'].enum_type = _POOLING3DPARAMETER_POOLMETHOD
_POOLING3DPARAMETER_POOLMETHOD.containing_type = _POOLING3DPARAMETER
_NORMALIZEPARAMETER.fields_by_name['mode'].enum_type = _NORMALIZEPARAMETER_NORMALIZEMETHOD
_NORMALIZEPARAMETER_NORMALIZEMETHOD.containing_type = _NORMALIZEPARAMETER
_BIASPARAMETER.fields_by_name['filler'].message_type = _FILLERPARAMETER
_UNPOOLINGPARAMETER.fields_by_name['unpool'].enum_type = _UNPOOLINGPARAMETER_UNPOOLMETHOD
_UNPOOLINGPARAMETER.fields_by_name['engine'].enum_type = _UNPOOLINGPARAMETER_ENGINE
_UNPOOLINGPARAMETER_UNPOOLMETHOD.containing_type = _UNPOOLINGPARAMETER
_UNPOOLINGPARAMETER_ENGINE.containing_type = _UNPOOLINGPARAMETER
_LSTMPARAMETER.fields_by_name['weight_filler'].message_type = _FILLERPARAMETER
_LSTMPARAMETER.fields_by_name['bias_filler'].message_type = _FILLERPARAMETER
_BTCOSTVOLUMEPARAMETER.fields_by_name['cost_domain_type'].enum_type = _BTCOSTVOLUMEPARAMETER_COSTDOMAINTYPE
_BTCOSTVOLUMEPARAMETER.fields_by_name['implement_mode'].enum_type = _BTCOSTVOLUMEPARAMETER_IMPLEMENTMODE
_BTCOSTVOLUMEPARAMETER_COSTDOMAINTYPE.containing_type = _BTCOSTVOLUMEPARAMETER
_BTCOSTVOLUMEPARAMETER_IMPLEMENTMODE.containing_type = _BTCOSTVOLUMEPARAMETER
_CORRELATIONPARAMETER.fields_by_name['correlation_type'].enum_type = _CORRELATIONPARAMETER_CORRELATIONTYPE
_CORRELATIONPARAMETER_CORRELATIONTYPE.containing_type = _CORRELATIONPARAMETER
_SCALEPARAMETER.fields_by_name['filler'].message_type = _FILLERPARAMETER
_SCALEPARAMETER.fields_by_name['bias_filler'].message_type = _FILLERPARAMETER
_RPNPROPOSALLAYERPARAMETER_TESTDESCPARAM.containing_type = _RPNPROPOSALLAYERPARAMETER
_RPNPROPOSALLAYERPARAMETER_TRAINDESCPARAM.containing_type = _RPNPROPOSALLAYERPARAMETER
_RPNPROPOSALLAYERPARAMETER.fields_by_name['test_desc_param'].message_type = _RPNPROPOSALLAYERPARAMETER_TESTDESCPARAM
_RPNPROPOSALLAYERPARAMETER.fields_by_name['train_desc_param'].message_type = _RPNPROPOSALLAYERPARAMETER_TRAINDESCPARAM
_ZXYBNPARAMETER.fields_by_name['slope_filler'].message_type = _FILLERPARAMETER
_ZXYBNPARAMETER.fields_by_name['bias_filler'].message_type = _FILLERPARAMETER
_ZXYBNPARAMETER.fields_by_name['engine'].enum_type = _ZXYBNPARAMETER_ENGINE
_ZXYBNPARAMETER_ENGINE.containing_type = _ZXYBNPARAMETER
_CONVOLUTIONPARAMETER.fields_by_name['weight_filler'].message_type = _FILLERPARAMETER
_CONVOLUTIONPARAMETER.fields_by_name['bias_filler'].message_type = _FILLERPARAMETER
_CONVOLUTIONPARAMETER.fields_by_name['engine'].enum_type = _CONVOLUTIONPARAMETER_ENGINE
_CONVOLUTIONPARAMETER.fields_by_name['ForwardAlgo'].enum_type = _CONVOLUTIONPARAMETER_PPLCONVOLUTIONFORWARDALGO_T
_CONVOLUTIONPARAMETER.fields_by_name['quantize_param'].message_type = _QUANTIZEPARAMETER
_CONVOLUTIONPARAMETER.fields_by_name['perchannel_quantize_param'].message_type = _QUANTIZEPARAMETER
_CONVOLUTIONPARAMETER_ENGINE.containing_type = _CONVOLUTIONPARAMETER
_CONVOLUTIONPARAMETER_PPLCONVOLUTIONFORWARDALGO_T.containing_type = _CONVOLUTIONPARAMETER
_DATAPARAMETER.fields_by_name['backend'].enum_type = _DATAPARAMETER_DB
_DATAPARAMETER_DB.containing_type = _DATAPARAMETER
_DETECTIONOUTPUTPARAMETER.fields_by_name['nms_param'].message_type = _NONMAXIMUMSUPPRESSIONPARAMETER
_DETECTIONOUTPUTPARAMETER.fields_by_name['save_output_param'].message_type = _SAVEOUTPUTPARAMETER
_DETECTIONOUTPUTPARAMETER.fields_by_name['code_type'].enum_type = _PRIORBOXPARAMETER_CODETYPE
_DUMMYDATAPARAMETER.fields_by_name['data_filler'].message_type = _FILLERPARAMETER
_DUMMYDATAPARAMETER.fields_by_name['shape'].message_type = _BLOBSHAPE
_ELTWISEPARAMETER.fields_by_name['operation'].enum_type = _ELTWISEPARAMETER_ELTWISEOP
_ELTWISEPARAMETER_ELTWISEOP.containing_type = _ELTWISEPARAMETER
_HINGELOSSPARAMETER.fields_by_name['norm'].enum_type = _HINGELOSSPARAMETER_NORM
_HINGELOSSPARAMETER_NORM.containing_type = _HINGELOSSPARAMETER
_INNERPRODUCTPARAMETER.fields_by_name['weight_filler'].message_type = _FILLERPARAMETER
_INNERPRODUCTPARAMETER.fields_by_name['bias_filler'].message_type = _FILLERPARAMETER
_INNERPRODUCTPARAMETER.fields_by_name['quantize_param'].message_type = _QUANTIZEPARAMETER
_INNERPRODUCTPARAMETER.fields_by_name['perchannel_quantize_param'].message_type = _QUANTIZEPARAMETER
_LRNPARAMETER.fields_by_name['norm_region'].enum_type = _LRNPARAMETER_NORMREGION
_LRNPARAMETER_NORMREGION.containing_type = _LRNPARAMETER
_POOLINGPARAMETER.fields_by_name['pool'].enum_type = _POOLINGPARAMETER_POOLMETHOD
_POOLINGPARAMETER.fields_by_name['engine'].enum_type = _POOLINGPARAMETER_ENGINE
_POOLINGPARAMETER_POOLMETHOD.containing_type = _POOLINGPARAMETER
_POOLINGPARAMETER_ENGINE.containing_type = _POOLINGPARAMETER
_POOLINGSPECIFICPARAMETER.fields_by_name['pool'].enum_type = _POOLINGSPECIFICPARAMETER_POOLSPECIFICMETHOD
_POOLINGSPECIFICPARAMETER_POOLSPECIFICMETHOD.containing_type = _POOLINGSPECIFICPARAMETER
_PRIORBOXPARAMETER_CODETYPE.containing_type = _PRIORBOXPARAMETER
_RELUPARAMETER.fields_by_name['engine'].enum_type = _RELUPARAMETER_ENGINE
_RELUPARAMETER_ENGINE.containing_type = _RELUPARAMETER
_RESHAPEPARAMETER.fields_by_name['shape'].message_type = _BLOBSHAPE
_SIGMOIDPARAMETER.fields_by_name['engine'].enum_type = _SIGMOIDPARAMETER_ENGINE
_SIGMOIDPARAMETER_ENGINE.containing_type = _SIGMOIDPARAMETER
_SOFTMAXPARAMETER.fields_by_name['engine'].enum_type = _SOFTMAXPARAMETER_ENGINE
_SOFTMAXPARAMETER_ENGINE.containing_type = _SOFTMAXPARAMETER
_TANHPARAMETER.fields_by_name['engine'].enum_type = _TANHPARAMETER_ENGINE
_TANHPARAMETER_ENGINE.containing_type = _TANHPARAMETER
_BNPARAMETER.fields_by_name['scale_filler'].message_type = _FILLERPARAMETER
_BNPARAMETER.fields_by_name['shift_filler'].message_type = _FILLERPARAMETER
_CTCPARAMETER.fields_by_name['decode_type'].enum_type = _CTCPARAMETER_DECODER
_CTCPARAMETER_DECODER.containing_type = _CTCPARAMETER
_PRELUPARAMETER.fields_by_name['filler'].message_type = _FILLERPARAMETER
_PRELUPARAMETER.fields_by_name['quantize_param'].message_type = _QUANTIZEPARAMETER
_CROPPARAMETER.fields_by_name['type'].enum_type = _CROPPARAMETER_TYPE
_CROPPARAMETER_TYPE.containing_type = _CROPPARAMETER
_ROIPARAMETER.fields_by_name['type'].enum_type = _INTERPTYPE
_RECURRENTPARAMETER.fields_by_name['weight_filler'].message_type = _FILLERPARAMETER
_RECURRENTPARAMETER.fields_by_name['bias_filler'].message_type = _FILLERPARAMETER
_V1LAYERPARAMETER.fields_by_name['include'].message_type = _NETSTATERULE
_V1LAYERPARAMETER.fields_by_name['exclude'].message_type = _NETSTATERULE
_V1LAYERPARAMETER.fields_by_name['type'].enum_type = _V1LAYERPARAMETER_LAYERTYPE
_V1LAYERPARAMETER.fields_by_name['blobs'].message_type = _BLOBPROTO
_V1LAYERPARAMETER.fields_by_name['blob_share_mode'].enum_type = _V1LAYERPARAMETER_DIMCHECKMODE
_V1LAYERPARAMETER.fields_by_name['accuracy_param'].message_type = _ACCURACYPARAMETER
_V1LAYERPARAMETER.fields_by_name['argmax_param'].message_type = _ARGMAXPARAMETER
_V1LAYERPARAMETER.fields_by_name['concat_param'].message_type = _CONCATPARAMETER
_V1LAYERPARAMETER.fields_by_name['contrastive_loss_param'].message_type = _CONTRASTIVELOSSPARAMETER
_V1LAYERPARAMETER.fields_by_name['convolution_param'].message_type = _CONVOLUTIONPARAMETER
_V1LAYERPARAMETER.fields_by_name['data_param'].message_type = _DATAPARAMETER
_V1LAYERPARAMETER.fields_by_name['dropout_param'].message_type = _DROPOUTPARAMETER
_V1LAYERPARAMETER.fields_by_name['dummy_data_param'].message_type = _DUMMYDATAPARAMETER
_V1LAYERPARAMETER.fields_by_name['eltwise_param'].message_type = _ELTWISEPARAMETER
_V1LAYERPARAMETER.fields_by_name['exp_param'].message_type = _EXPPARAMETER
_V1LAYERPARAMETER.fields_by_name['hdf5_data_param'].message_type = _HDF5DATAPARAMETER
_V1LAYERPARAMETER.fields_by_name['hdf5_output_param'].message_type = _HDF5OUTPUTPARAMETER
_V1LAYERPARAMETER.fields_by_name['hinge_loss_param'].message_type = _HINGELOSSPARAMETER
_V1LAYERPARAMETER.fields_by_name['image_data_param'].message_type = _IMAGEDATAPARAMETER
_V1LAYERPARAMETER.fields_by_name['infogain_loss_param'].message_type = _INFOGAINLOSSPARAMETER
_V1LAYERPARAMETER.fields_by_name['inner_product_param'].message_type = _INNERPRODUCTPARAMETER
_V1LAYERPARAMETER.fields_by_name['lrn_param'].message_type = _LRNPARAMETER
_V1LAYERPARAMETER.fields_by_name['memory_data_param'].message_type = _MEMORYDATAPARAMETER
_V1LAYERPARAMETER.fields_by_name['mvn_param'].message_type = _MVNPARAMETER
_V1LAYERPARAMETER.fields_by_name['pooling_param'].message_type = _POOLINGPARAMETER
_V1LAYERPARAMETER.fields_by_name['power_param'].message_type = _POWERPARAMETER
_V1LAYERPARAMETER.fields_by_name['relu_param'].message_type = _RELUPARAMETER
_V1LAYERPARAMETER.fields_by_name['sigmoid_param'].message_type = _SIGMOIDPARAMETER
_V1LAYERPARAMETER.fields_by_name['softmax_param'].message_type = _SOFTMAXPARAMETER
_V1LAYERPARAMETER.fields_by_name['slice_param'].message_type = _SLICEPARAMETER
_V1LAYERPARAMETER.fields_by_name['tanh_param'].message_type = _TANHPARAMETER
_V1LAYERPARAMETER.fields_by_name['threshold_param'].message_type = _THRESHOLDPARAMETER
_V1LAYERPARAMETER.fields_by_name['window_data_param'].message_type = _WINDOWDATAPARAMETER
_V1LAYERPARAMETER.fields_by_name['transform_param'].message_type = _TRANSFORMATIONPARAMETER
_V1LAYERPARAMETER.fields_by_name['loss_param'].message_type = _LOSSPARAMETER
_V1LAYERPARAMETER.fields_by_name['layer'].message_type = _V0LAYERPARAMETER
_V1LAYERPARAMETER_LAYERTYPE.containing_type = _V1LAYERPARAMETER
_V1LAYERPARAMETER_DIMCHECKMODE.containing_type = _V1LAYERPARAMETER
_V0LAYERPARAMETER.fields_by_name['weight_filler'].message_type = _FILLERPARAMETER
_V0LAYERPARAMETER.fields_by_name['bias_filler'].message_type = _FILLERPARAMETER
_V0LAYERPARAMETER.fields_by_name['pool'].enum_type = _V0LAYERPARAMETER_POOLMETHOD
_V0LAYERPARAMETER.fields_by_name['blobs'].message_type = _BLOBPROTO
_V0LAYERPARAMETER.fields_by_name['hdf5_output_param'].message_type = _HDF5OUTPUTPARAMETER
_V0LAYERPARAMETER_POOLMETHOD.containing_type = _V0LAYERPARAMETER
_POINTSCURVEPARAMETER.fields_by_name['slope_filler'].message_type = _FILLERPARAMETER
_POINTSCURVEPARAMETER.fields_by_name['bias_filler'].message_type = _FILLERPARAMETER
_SUBPIXELUPPARAMETER.fields_by_name['output_datatype'].enum_type = _SUBPIXELUPPARAMETER_DATATYPE
_SUBPIXELUPPARAMETER.fields_by_name['mode'].enum_type = _SUBPIXELUPPARAMETER_MODETYPE
_SUBPIXELUPPARAMETER_DATATYPE.containing_type = _SUBPIXELUPPARAMETER
_SUBPIXELUPPARAMETER_MODETYPE.containing_type = _SUBPIXELUPPARAMETER
_REDUCEPARAMETER_REDUCEOP.containing_type = _REDUCEPARAMETER
_PADPARAMETER.fields_by_name['mode'].enum_type = _PADPARAMETER_PADMODE
_PADPARAMETER_PADMODE.containing_type = _PADPARAMETER
_MATMULPARAMETER.fields_by_name['mode'].enum_type = _MATMULPARAMETER_MATMULMODE
_MATMULPARAMETER_MATMULMODE.containing_type = _MATMULPARAMETER
_GRIDSAMPLEPARAMETER.fields_by_name['padding_mode'].enum_type = _GRIDSAMPLEPARAMETER_PADDINGMODE
_GRIDSAMPLEPARAMETER_PADDINGMODE.containing_type = _GRIDSAMPLEPARAMETER
_GRIDSAMPLE3DPARAMETER.fields_by_name['padding_mode'].enum_type = _GRIDSAMPLE3DPARAMETER_PADDINGMODE
_GRIDSAMPLE3DPARAMETER_PADDINGMODE.containing_type = _GRIDSAMPLE3DPARAMETER
DESCRIPTOR.message_types_by_name['BlobShape'] = _BLOBSHAPE
DESCRIPTOR.message_types_by_name['BlobProto'] = _BLOBPROTO
DESCRIPTOR.message_types_by_name['BlobProtoVector'] = _BLOBPROTOVECTOR
DESCRIPTOR.message_types_by_name['QuantizeParameter'] = _QUANTIZEPARAMETER
DESCRIPTOR.message_types_by_name['Datum'] = _DATUM
DESCRIPTOR.message_types_by_name['FillerParameter'] = _FILLERPARAMETER
DESCRIPTOR.message_types_by_name['NetParameter'] = _NETPARAMETER
DESCRIPTOR.message_types_by_name['SolverParameter'] = _SOLVERPARAMETER
DESCRIPTOR.message_types_by_name['SolverState'] = _SOLVERSTATE
DESCRIPTOR.message_types_by_name['NetState'] = _NETSTATE
DESCRIPTOR.message_types_by_name['NetStateRule'] = _NETSTATERULE
DESCRIPTOR.message_types_by_name['ParamSpec'] = _PARAMSPEC
DESCRIPTOR.message_types_by_name['LayerParameter'] = _LAYERPARAMETER
DESCRIPTOR.message_types_by_name['Convolution3dParameter'] = _CONVOLUTION3DPARAMETER
DESCRIPTOR.message_types_by_name['ConvolutionTranspose3dParameter'] = _CONVOLUTIONTRANSPOSE3DPARAMETER
DESCRIPTOR.message_types_by_name['Deconvolution3dParameter'] = _DECONVOLUTION3DPARAMETER
DESCRIPTOR.message_types_by_name['ReLU3dParameter'] = _RELU3DPARAMETER
DESCRIPTOR.message_types_by_name['GroupNorm3dParameter'] = _GROUPNORM3DPARAMETER
DESCRIPTOR.message_types_by_name['BatchNorm3dParameter'] = _BATCHNORM3DPARAMETER
DESCRIPTOR.message_types_by_name['Pooling3dParameter'] = _POOLING3DPARAMETER
DESCRIPTOR.message_types_by_name['NormalizeParameter'] = _NORMALIZEPARAMETER
DESCRIPTOR.message_types_by_name['HeatMap2CoordParameter'] = _HEATMAP2COORDPARAMETER
DESCRIPTOR.message_types_by_name['BiasParameter'] = _BIASPARAMETER
DESCRIPTOR.message_types_by_name['AlignmentToROIParameter'] = _ALIGNMENTTOROIPARAMETER
DESCRIPTOR.message_types_by_name['ROIAlignPoolingParameter'] = _ROIALIGNPOOLINGPARAMETER
DESCRIPTOR.message_types_by_name['ROIAlignPoolingPodParameter'] = _ROIALIGNPOOLINGPODPARAMETER
DESCRIPTOR.message_types_by_name['PSROIAlignPoolingParameter'] = _PSROIALIGNPOOLINGPARAMETER
DESCRIPTOR.message_types_by_name['UnpoolingParameter'] = _UNPOOLINGPARAMETER
DESCRIPTOR.message_types_by_name['ReverseParameter'] = _REVERSEPARAMETER
DESCRIPTOR.message_types_by_name['LSTMParameter'] = _LSTMPARAMETER
DESCRIPTOR.message_types_by_name['PriorVBoxParameter'] = _PRIORVBOXPARAMETER
DESCRIPTOR.message_types_by_name['NNUpsampleParameter'] = _NNUPSAMPLEPARAMETER
DESCRIPTOR.message_types_by_name['BTCostVolumeParameter'] = _BTCOSTVOLUMEPARAMETER
DESCRIPTOR.message_types_by_name['CorrelationParameter'] = _CORRELATIONPARAMETER
DESCRIPTOR.message_types_by_name['PSROIPoolingParameter'] = _PSROIPOOLINGPARAMETER
DESCRIPTOR.message_types_by_name['ScaleParameter'] = _SCALEPARAMETER
DESCRIPTOR.message_types_by_name['ScalesParameter'] = _SCALESPARAMETER
DESCRIPTOR.message_types_by_name['TransposeParameter'] = _TRANSPOSEPARAMETER
DESCRIPTOR.message_types_by_name['ROIAlignParameter'] = _ROIALIGNPARAMETER
DESCRIPTOR.message_types_by_name['RpnProposalLayerParameter'] = _RPNPROPOSALLAYERPARAMETER
DESCRIPTOR.message_types_by_name['ROIMaskPoolingParameter'] = _ROIMASKPOOLINGPARAMETER
DESCRIPTOR.message_types_by_name['ZXYBNParameter'] = _ZXYBNPARAMETER
DESCRIPTOR.message_types_by_name['TransformationParameter'] = _TRANSFORMATIONPARAMETER
DESCRIPTOR.message_types_by_name['LossParameter'] = _LOSSPARAMETER
DESCRIPTOR.message_types_by_name['AccuracyParameter'] = _ACCURACYPARAMETER
DESCRIPTOR.message_types_by_name['ArgMaxParameter'] = _ARGMAXPARAMETER
DESCRIPTOR.message_types_by_name['ConcatParameter'] = _CONCATPARAMETER
DESCRIPTOR.message_types_by_name['ContrastiveLossParameter'] = _CONTRASTIVELOSSPARAMETER
DESCRIPTOR.message_types_by_name['ConvolutionParameter'] = _CONVOLUTIONPARAMETER
DESCRIPTOR.message_types_by_name['DataParameter'] = _DATAPARAMETER
DESCRIPTOR.message_types_by_name['NonMaximumSuppressionParameter'] = _NONMAXIMUMSUPPRESSIONPARAMETER
DESCRIPTOR.message_types_by_name['SaveOutputParameter'] = _SAVEOUTPUTPARAMETER
DESCRIPTOR.message_types_by_name['DetectionOutputParameter'] = _DETECTIONOUTPUTPARAMETER
DESCRIPTOR.message_types_by_name['DropoutParameter'] = _DROPOUTPARAMETER
DESCRIPTOR.message_types_by_name['DummyDataParameter'] = _DUMMYDATAPARAMETER
DESCRIPTOR.message_types_by_name['EltwiseParameter'] = _ELTWISEPARAMETER
DESCRIPTOR.message_types_by_name['ExpParameter'] = _EXPPARAMETER
DESCRIPTOR.message_types_by_name['LogParameter'] = _LOGPARAMETER
DESCRIPTOR.message_types_by_name['HDF5DataParameter'] = _HDF5DATAPARAMETER
DESCRIPTOR.message_types_by_name['HDF5OutputParameter'] = _HDF5OUTPUTPARAMETER
DESCRIPTOR.message_types_by_name['HingeLossParameter'] = _HINGELOSSPARAMETER
DESCRIPTOR.message_types_by_name['ImageDataParameter'] = _IMAGEDATAPARAMETER
DESCRIPTOR.message_types_by_name['InfogainLossParameter'] = _INFOGAINLOSSPARAMETER
DESCRIPTOR.message_types_by_name['InnerProductParameter'] = _INNERPRODUCTPARAMETER
DESCRIPTOR.message_types_by_name['LRNParameter'] = _LRNPARAMETER
DESCRIPTOR.message_types_by_name['MemoryDataParameter'] = _MEMORYDATAPARAMETER
DESCRIPTOR.message_types_by_name['MVNParameter'] = _MVNPARAMETER
DESCRIPTOR.message_types_by_name['PermuteParameter'] = _PERMUTEPARAMETER
DESCRIPTOR.message_types_by_name['PoolingParameter'] = _POOLINGPARAMETER
DESCRIPTOR.message_types_by_name['PoolingSpecificParameter'] = _POOLINGSPECIFICPARAMETER
DESCRIPTOR.message_types_by_name['PowerParameter'] = _POWERPARAMETER
DESCRIPTOR.message_types_by_name['PriorBoxParameter'] = _PRIORBOXPARAMETER
DESCRIPTOR.message_types_by_name['PythonParameter'] = _PYTHONPARAMETER
DESCRIPTOR.message_types_by_name['ReLUParameter'] = _RELUPARAMETER
DESCRIPTOR.message_types_by_name['ROIPoolingParameter'] = _ROIPOOLINGPARAMETER
DESCRIPTOR.message_types_by_name['ReshapeParameter'] = _RESHAPEPARAMETER
DESCRIPTOR.message_types_by_name['SigmoidParameter'] = _SIGMOIDPARAMETER
DESCRIPTOR.message_types_by_name['SliceParameter'] = _SLICEPARAMETER
DESCRIPTOR.message_types_by_name['SoftmaxParameter'] = _SOFTMAXPARAMETER
DESCRIPTOR.message_types_by_name['TanHParameter'] = _TANHPARAMETER
DESCRIPTOR.message_types_by_name['ThresholdParameter'] = _THRESHOLDPARAMETER
DESCRIPTOR.message_types_by_name['WindowDataParameter'] = _WINDOWDATAPARAMETER
DESCRIPTOR.message_types_by_name['BNParameter'] = _BNPARAMETER
DESCRIPTOR.message_types_by_name['CTCParameter'] = _CTCPARAMETER
DESCRIPTOR.message_types_by_name['PReLUParameter'] = _PRELUPARAMETER
DESCRIPTOR.message_types_by_name['CropParameter'] = _CROPPARAMETER
DESCRIPTOR.message_types_by_name['AffineTransParameter'] = _AFFINETRANSPARAMETER
DESCRIPTOR.message_types_by_name['AffineTransPointParameter'] = _AFFINETRANSPOINTPARAMETER
DESCRIPTOR.message_types_by_name['CalcAffineMatParameter'] = _CALCAFFINEMATPARAMETER
DESCRIPTOR.message_types_by_name['ROIParameter'] = _ROIPARAMETER
DESCRIPTOR.message_types_by_name['InterpParameter'] = _INTERPPARAMETER
DESCRIPTOR.message_types_by_name['RecurrentParameter'] = _RECURRENTPARAMETER
DESCRIPTOR.message_types_by_name['FlattenParameter'] = _FLATTENPARAMETER
DESCRIPTOR.message_types_by_name['BatchNormParameter'] = _BATCHNORMPARAMETER
DESCRIPTOR.message_types_by_name['V1LayerParameter'] = _V1LAYERPARAMETER
DESCRIPTOR.message_types_by_name['V0LayerParameter'] = _V0LAYERPARAMETER
DESCRIPTOR.message_types_by_name['ChannelShuffleParameter'] = _CHANNELSHUFFLEPARAMETER
DESCRIPTOR.message_types_by_name['MovingAvgParameter'] = _MOVINGAVGPARAMETER
DESCRIPTOR.message_types_by_name['SlicingParameter'] = _SLICINGPARAMETER
DESCRIPTOR.message_types_by_name['BilateralSlicingParameter'] = _BILATERALSLICINGPARAMETER
DESCRIPTOR.message_types_by_name['BilateralSliceApplyParameter'] = _BILATERALSLICEAPPLYPARAMETER
DESCRIPTOR.message_types_by_name['AdaptiveBilateralSliceApplyParameter'] = _ADAPTIVEBILATERALSLICEAPPLYPARAMETER
DESCRIPTOR.message_types_by_name['PointsCurveParameter'] = _POINTSCURVEPARAMETER
DESCRIPTOR.message_types_by_name['SubpixelDownParameter'] = _SUBPIXELDOWNPARAMETER
DESCRIPTOR.message_types_by_name['SubpixelUpParameter'] = _SUBPIXELUPPARAMETER
DESCRIPTOR.message_types_by_name['ClipParameter'] = _CLIPPARAMETER
DESCRIPTOR.message_types_by_name['EltwiseAffineTransformParameter'] = _ELTWISEAFFINETRANSFORMPARAMETER
DESCRIPTOR.message_types_by_name['TileParameter'] = _TILEPARAMETER
DESCRIPTOR.message_types_by_name['ReduceParameter'] = _REDUCEPARAMETER
DESCRIPTOR.message_types_by_name['PadParameter'] = _PADPARAMETER
DESCRIPTOR.message_types_by_name['ReLU6Parameter'] = _RELU6PARAMETER
DESCRIPTOR.message_types_by_name['Correlation2DParameter'] = _CORRELATION2DPARAMETER
DESCRIPTOR.message_types_by_name['MatMulParameter'] = _MATMULPARAMETER
DESCRIPTOR.message_types_by_name['ParameterParameter'] = _PARAMETERPARAMETER
DESCRIPTOR.message_types_by_name['PixelShuffleParameter'] = _PIXELSHUFFLEPARAMETER
DESCRIPTOR.message_types_by_name['ROITransformParameter'] = _ROITRANSFORMPARAMETER
DESCRIPTOR.message_types_by_name['MainTransformParameter'] = _MAINTRANSFORMPARAMETER
DESCRIPTOR.message_types_by_name['InstanceNormParameter'] = _INSTANCENORMPARAMETER
DESCRIPTOR.message_types_by_name['GridSampleParameter'] = _GRIDSAMPLEPARAMETER
DESCRIPTOR.message_types_by_name['GridSample3DParameter'] = _GRIDSAMPLE3DPARAMETER
DESCRIPTOR.message_types_by_name['ReduceL2Parameter'] = _REDUCEL2PARAMETER
DESCRIPTOR.message_types_by_name['VarianceParameter'] = _VARIANCEPARAMETER
DESCRIPTOR.message_types_by_name['MeanParameter'] = _MEANPARAMETER
DESCRIPTOR.message_types_by_name['Interp3dParameter'] = _INTERP3DPARAMETER
DESCRIPTOR.message_types_by_name['GRUParameter'] = _GRUPARAMETER
DESCRIPTOR.message_types_by_name['CorrelationMigParameter'] = _CORRELATIONMIGPARAMETER
DESCRIPTOR.message_types_by_name['ArgSortParameter'] = _ARGSORTPARAMETER
DESCRIPTOR.message_types_by_name['CumProdParameter'] = _CUMPRODPARAMETER
DESCRIPTOR.enum_types_by_name['Phase'] = _PHASE
DESCRIPTOR.enum_types_by_name['InterpType'] = _INTERPTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

BlobShape = _reflection.GeneratedProtocolMessageType('BlobShape', (_message.Message,), {
  'DESCRIPTOR' : _BLOBSHAPE,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.BlobShape)
  })
_sym_db.RegisterMessage(BlobShape)

BlobProto = _reflection.GeneratedProtocolMessageType('BlobProto', (_message.Message,), {
  'DESCRIPTOR' : _BLOBPROTO,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.BlobProto)
  })
_sym_db.RegisterMessage(BlobProto)

BlobProtoVector = _reflection.GeneratedProtocolMessageType('BlobProtoVector', (_message.Message,), {
  'DESCRIPTOR' : _BLOBPROTOVECTOR,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.BlobProtoVector)
  })
_sym_db.RegisterMessage(BlobProtoVector)

QuantizeParameter = _reflection.GeneratedProtocolMessageType('QuantizeParameter', (_message.Message,), {
  'DESCRIPTOR' : _QUANTIZEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.QuantizeParameter)
  })
_sym_db.RegisterMessage(QuantizeParameter)

Datum = _reflection.GeneratedProtocolMessageType('Datum', (_message.Message,), {
  'DESCRIPTOR' : _DATUM,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.Datum)
  })
_sym_db.RegisterMessage(Datum)

FillerParameter = _reflection.GeneratedProtocolMessageType('FillerParameter', (_message.Message,), {
  'DESCRIPTOR' : _FILLERPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.FillerParameter)
  })
_sym_db.RegisterMessage(FillerParameter)

NetParameter = _reflection.GeneratedProtocolMessageType('NetParameter', (_message.Message,), {
  'DESCRIPTOR' : _NETPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.NetParameter)
  })
_sym_db.RegisterMessage(NetParameter)

SolverParameter = _reflection.GeneratedProtocolMessageType('SolverParameter', (_message.Message,), {
  'DESCRIPTOR' : _SOLVERPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.SolverParameter)
  })
_sym_db.RegisterMessage(SolverParameter)

SolverState = _reflection.GeneratedProtocolMessageType('SolverState', (_message.Message,), {
  'DESCRIPTOR' : _SOLVERSTATE,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.SolverState)
  })
_sym_db.RegisterMessage(SolverState)

NetState = _reflection.GeneratedProtocolMessageType('NetState', (_message.Message,), {
  'DESCRIPTOR' : _NETSTATE,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.NetState)
  })
_sym_db.RegisterMessage(NetState)

NetStateRule = _reflection.GeneratedProtocolMessageType('NetStateRule', (_message.Message,), {
  'DESCRIPTOR' : _NETSTATERULE,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.NetStateRule)
  })
_sym_db.RegisterMessage(NetStateRule)

ParamSpec = _reflection.GeneratedProtocolMessageType('ParamSpec', (_message.Message,), {
  'DESCRIPTOR' : _PARAMSPEC,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ParamSpec)
  })
_sym_db.RegisterMessage(ParamSpec)

LayerParameter = _reflection.GeneratedProtocolMessageType('LayerParameter', (_message.Message,), {
  'DESCRIPTOR' : _LAYERPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.LayerParameter)
  })
_sym_db.RegisterMessage(LayerParameter)

Convolution3dParameter = _reflection.GeneratedProtocolMessageType('Convolution3dParameter', (_message.Message,), {
  'DESCRIPTOR' : _CONVOLUTION3DPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.Convolution3dParameter)
  })
_sym_db.RegisterMessage(Convolution3dParameter)

ConvolutionTranspose3dParameter = _reflection.GeneratedProtocolMessageType('ConvolutionTranspose3dParameter', (_message.Message,), {
  'DESCRIPTOR' : _CONVOLUTIONTRANSPOSE3DPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ConvolutionTranspose3dParameter)
  })
_sym_db.RegisterMessage(ConvolutionTranspose3dParameter)

Deconvolution3dParameter = _reflection.GeneratedProtocolMessageType('Deconvolution3dParameter', (_message.Message,), {
  'DESCRIPTOR' : _DECONVOLUTION3DPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.Deconvolution3dParameter)
  })
_sym_db.RegisterMessage(Deconvolution3dParameter)

ReLU3dParameter = _reflection.GeneratedProtocolMessageType('ReLU3dParameter', (_message.Message,), {
  'DESCRIPTOR' : _RELU3DPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ReLU3dParameter)
  })
_sym_db.RegisterMessage(ReLU3dParameter)

GroupNorm3dParameter = _reflection.GeneratedProtocolMessageType('GroupNorm3dParameter', (_message.Message,), {
  'DESCRIPTOR' : _GROUPNORM3DPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.GroupNorm3dParameter)
  })
_sym_db.RegisterMessage(GroupNorm3dParameter)

BatchNorm3dParameter = _reflection.GeneratedProtocolMessageType('BatchNorm3dParameter', (_message.Message,), {
  'DESCRIPTOR' : _BATCHNORM3DPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.BatchNorm3dParameter)
  })
_sym_db.RegisterMessage(BatchNorm3dParameter)

Pooling3dParameter = _reflection.GeneratedProtocolMessageType('Pooling3dParameter', (_message.Message,), {
  'DESCRIPTOR' : _POOLING3DPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.Pooling3dParameter)
  })
_sym_db.RegisterMessage(Pooling3dParameter)

NormalizeParameter = _reflection.GeneratedProtocolMessageType('NormalizeParameter', (_message.Message,), {
  'DESCRIPTOR' : _NORMALIZEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.NormalizeParameter)
  })
_sym_db.RegisterMessage(NormalizeParameter)

HeatMap2CoordParameter = _reflection.GeneratedProtocolMessageType('HeatMap2CoordParameter', (_message.Message,), {
  'DESCRIPTOR' : _HEATMAP2COORDPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.HeatMap2CoordParameter)
  })
_sym_db.RegisterMessage(HeatMap2CoordParameter)

BiasParameter = _reflection.GeneratedProtocolMessageType('BiasParameter', (_message.Message,), {
  'DESCRIPTOR' : _BIASPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.BiasParameter)
  })
_sym_db.RegisterMessage(BiasParameter)

AlignmentToROIParameter = _reflection.GeneratedProtocolMessageType('AlignmentToROIParameter', (_message.Message,), {
  'DESCRIPTOR' : _ALIGNMENTTOROIPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.AlignmentToROIParameter)
  })
_sym_db.RegisterMessage(AlignmentToROIParameter)

ROIAlignPoolingParameter = _reflection.GeneratedProtocolMessageType('ROIAlignPoolingParameter', (_message.Message,), {
  'DESCRIPTOR' : _ROIALIGNPOOLINGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ROIAlignPoolingParameter)
  })
_sym_db.RegisterMessage(ROIAlignPoolingParameter)

ROIAlignPoolingPodParameter = _reflection.GeneratedProtocolMessageType('ROIAlignPoolingPodParameter', (_message.Message,), {
  'DESCRIPTOR' : _ROIALIGNPOOLINGPODPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ROIAlignPoolingPodParameter)
  })
_sym_db.RegisterMessage(ROIAlignPoolingPodParameter)

PSROIAlignPoolingParameter = _reflection.GeneratedProtocolMessageType('PSROIAlignPoolingParameter', (_message.Message,), {
  'DESCRIPTOR' : _PSROIALIGNPOOLINGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PSROIAlignPoolingParameter)
  })
_sym_db.RegisterMessage(PSROIAlignPoolingParameter)

UnpoolingParameter = _reflection.GeneratedProtocolMessageType('UnpoolingParameter', (_message.Message,), {
  'DESCRIPTOR' : _UNPOOLINGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.UnpoolingParameter)
  })
_sym_db.RegisterMessage(UnpoolingParameter)

ReverseParameter = _reflection.GeneratedProtocolMessageType('ReverseParameter', (_message.Message,), {
  'DESCRIPTOR' : _REVERSEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ReverseParameter)
  })
_sym_db.RegisterMessage(ReverseParameter)

LSTMParameter = _reflection.GeneratedProtocolMessageType('LSTMParameter', (_message.Message,), {
  'DESCRIPTOR' : _LSTMPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.LSTMParameter)
  })
_sym_db.RegisterMessage(LSTMParameter)

PriorVBoxParameter = _reflection.GeneratedProtocolMessageType('PriorVBoxParameter', (_message.Message,), {
  'DESCRIPTOR' : _PRIORVBOXPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PriorVBoxParameter)
  })
_sym_db.RegisterMessage(PriorVBoxParameter)

NNUpsampleParameter = _reflection.GeneratedProtocolMessageType('NNUpsampleParameter', (_message.Message,), {
  'DESCRIPTOR' : _NNUPSAMPLEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.NNUpsampleParameter)
  })
_sym_db.RegisterMessage(NNUpsampleParameter)

BTCostVolumeParameter = _reflection.GeneratedProtocolMessageType('BTCostVolumeParameter', (_message.Message,), {
  'DESCRIPTOR' : _BTCOSTVOLUMEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.BTCostVolumeParameter)
  })
_sym_db.RegisterMessage(BTCostVolumeParameter)

CorrelationParameter = _reflection.GeneratedProtocolMessageType('CorrelationParameter', (_message.Message,), {
  'DESCRIPTOR' : _CORRELATIONPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.CorrelationParameter)
  })
_sym_db.RegisterMessage(CorrelationParameter)

PSROIPoolingParameter = _reflection.GeneratedProtocolMessageType('PSROIPoolingParameter', (_message.Message,), {
  'DESCRIPTOR' : _PSROIPOOLINGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PSROIPoolingParameter)
  })
_sym_db.RegisterMessage(PSROIPoolingParameter)

ScaleParameter = _reflection.GeneratedProtocolMessageType('ScaleParameter', (_message.Message,), {
  'DESCRIPTOR' : _SCALEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ScaleParameter)
  })
_sym_db.RegisterMessage(ScaleParameter)

ScalesParameter = _reflection.GeneratedProtocolMessageType('ScalesParameter', (_message.Message,), {
  'DESCRIPTOR' : _SCALESPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ScalesParameter)
  })
_sym_db.RegisterMessage(ScalesParameter)

TransposeParameter = _reflection.GeneratedProtocolMessageType('TransposeParameter', (_message.Message,), {
  'DESCRIPTOR' : _TRANSPOSEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.TransposeParameter)
  })
_sym_db.RegisterMessage(TransposeParameter)

ROIAlignParameter = _reflection.GeneratedProtocolMessageType('ROIAlignParameter', (_message.Message,), {
  'DESCRIPTOR' : _ROIALIGNPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ROIAlignParameter)
  })
_sym_db.RegisterMessage(ROIAlignParameter)

RpnProposalLayerParameter = _reflection.GeneratedProtocolMessageType('RpnProposalLayerParameter', (_message.Message,), {

  'TestDescParam' : _reflection.GeneratedProtocolMessageType('TestDescParam', (_message.Message,), {
    'DESCRIPTOR' : _RPNPROPOSALLAYERPARAMETER_TESTDESCPARAM,
    '__module__' : 'caffe_pb2'
    # @@protoc_insertion_point(class_scope:caffe.RpnProposalLayerParameter.TestDescParam)
    })
  ,

  'TrainDescParam' : _reflection.GeneratedProtocolMessageType('TrainDescParam', (_message.Message,), {
    'DESCRIPTOR' : _RPNPROPOSALLAYERPARAMETER_TRAINDESCPARAM,
    '__module__' : 'caffe_pb2'
    # @@protoc_insertion_point(class_scope:caffe.RpnProposalLayerParameter.TrainDescParam)
    })
  ,
  'DESCRIPTOR' : _RPNPROPOSALLAYERPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.RpnProposalLayerParameter)
  })
_sym_db.RegisterMessage(RpnProposalLayerParameter)
_sym_db.RegisterMessage(RpnProposalLayerParameter.TestDescParam)
_sym_db.RegisterMessage(RpnProposalLayerParameter.TrainDescParam)

ROIMaskPoolingParameter = _reflection.GeneratedProtocolMessageType('ROIMaskPoolingParameter', (_message.Message,), {
  'DESCRIPTOR' : _ROIMASKPOOLINGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ROIMaskPoolingParameter)
  })
_sym_db.RegisterMessage(ROIMaskPoolingParameter)

ZXYBNParameter = _reflection.GeneratedProtocolMessageType('ZXYBNParameter', (_message.Message,), {
  'DESCRIPTOR' : _ZXYBNPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ZXYBNParameter)
  })
_sym_db.RegisterMessage(ZXYBNParameter)

TransformationParameter = _reflection.GeneratedProtocolMessageType('TransformationParameter', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFORMATIONPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.TransformationParameter)
  })
_sym_db.RegisterMessage(TransformationParameter)

LossParameter = _reflection.GeneratedProtocolMessageType('LossParameter', (_message.Message,), {
  'DESCRIPTOR' : _LOSSPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.LossParameter)
  })
_sym_db.RegisterMessage(LossParameter)

AccuracyParameter = _reflection.GeneratedProtocolMessageType('AccuracyParameter', (_message.Message,), {
  'DESCRIPTOR' : _ACCURACYPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.AccuracyParameter)
  })
_sym_db.RegisterMessage(AccuracyParameter)

ArgMaxParameter = _reflection.GeneratedProtocolMessageType('ArgMaxParameter', (_message.Message,), {
  'DESCRIPTOR' : _ARGMAXPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ArgMaxParameter)
  })
_sym_db.RegisterMessage(ArgMaxParameter)

ConcatParameter = _reflection.GeneratedProtocolMessageType('ConcatParameter', (_message.Message,), {
  'DESCRIPTOR' : _CONCATPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ConcatParameter)
  })
_sym_db.RegisterMessage(ConcatParameter)

ContrastiveLossParameter = _reflection.GeneratedProtocolMessageType('ContrastiveLossParameter', (_message.Message,), {
  'DESCRIPTOR' : _CONTRASTIVELOSSPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ContrastiveLossParameter)
  })
_sym_db.RegisterMessage(ContrastiveLossParameter)

ConvolutionParameter = _reflection.GeneratedProtocolMessageType('ConvolutionParameter', (_message.Message,), {
  'DESCRIPTOR' : _CONVOLUTIONPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ConvolutionParameter)
  })
_sym_db.RegisterMessage(ConvolutionParameter)

DataParameter = _reflection.GeneratedProtocolMessageType('DataParameter', (_message.Message,), {
  'DESCRIPTOR' : _DATAPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.DataParameter)
  })
_sym_db.RegisterMessage(DataParameter)

NonMaximumSuppressionParameter = _reflection.GeneratedProtocolMessageType('NonMaximumSuppressionParameter', (_message.Message,), {
  'DESCRIPTOR' : _NONMAXIMUMSUPPRESSIONPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.NonMaximumSuppressionParameter)
  })
_sym_db.RegisterMessage(NonMaximumSuppressionParameter)

SaveOutputParameter = _reflection.GeneratedProtocolMessageType('SaveOutputParameter', (_message.Message,), {
  'DESCRIPTOR' : _SAVEOUTPUTPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.SaveOutputParameter)
  })
_sym_db.RegisterMessage(SaveOutputParameter)

DetectionOutputParameter = _reflection.GeneratedProtocolMessageType('DetectionOutputParameter', (_message.Message,), {
  'DESCRIPTOR' : _DETECTIONOUTPUTPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.DetectionOutputParameter)
  })
_sym_db.RegisterMessage(DetectionOutputParameter)

DropoutParameter = _reflection.GeneratedProtocolMessageType('DropoutParameter', (_message.Message,), {
  'DESCRIPTOR' : _DROPOUTPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.DropoutParameter)
  })
_sym_db.RegisterMessage(DropoutParameter)

DummyDataParameter = _reflection.GeneratedProtocolMessageType('DummyDataParameter', (_message.Message,), {
  'DESCRIPTOR' : _DUMMYDATAPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.DummyDataParameter)
  })
_sym_db.RegisterMessage(DummyDataParameter)

EltwiseParameter = _reflection.GeneratedProtocolMessageType('EltwiseParameter', (_message.Message,), {
  'DESCRIPTOR' : _ELTWISEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.EltwiseParameter)
  })
_sym_db.RegisterMessage(EltwiseParameter)

ExpParameter = _reflection.GeneratedProtocolMessageType('ExpParameter', (_message.Message,), {
  'DESCRIPTOR' : _EXPPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ExpParameter)
  })
_sym_db.RegisterMessage(ExpParameter)

LogParameter = _reflection.GeneratedProtocolMessageType('LogParameter', (_message.Message,), {
  'DESCRIPTOR' : _LOGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.LogParameter)
  })
_sym_db.RegisterMessage(LogParameter)

HDF5DataParameter = _reflection.GeneratedProtocolMessageType('HDF5DataParameter', (_message.Message,), {
  'DESCRIPTOR' : _HDF5DATAPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.HDF5DataParameter)
  })
_sym_db.RegisterMessage(HDF5DataParameter)

HDF5OutputParameter = _reflection.GeneratedProtocolMessageType('HDF5OutputParameter', (_message.Message,), {
  'DESCRIPTOR' : _HDF5OUTPUTPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.HDF5OutputParameter)
  })
_sym_db.RegisterMessage(HDF5OutputParameter)

HingeLossParameter = _reflection.GeneratedProtocolMessageType('HingeLossParameter', (_message.Message,), {
  'DESCRIPTOR' : _HINGELOSSPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.HingeLossParameter)
  })
_sym_db.RegisterMessage(HingeLossParameter)

ImageDataParameter = _reflection.GeneratedProtocolMessageType('ImageDataParameter', (_message.Message,), {
  'DESCRIPTOR' : _IMAGEDATAPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ImageDataParameter)
  })
_sym_db.RegisterMessage(ImageDataParameter)

InfogainLossParameter = _reflection.GeneratedProtocolMessageType('InfogainLossParameter', (_message.Message,), {
  'DESCRIPTOR' : _INFOGAINLOSSPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.InfogainLossParameter)
  })
_sym_db.RegisterMessage(InfogainLossParameter)

InnerProductParameter = _reflection.GeneratedProtocolMessageType('InnerProductParameter', (_message.Message,), {
  'DESCRIPTOR' : _INNERPRODUCTPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.InnerProductParameter)
  })
_sym_db.RegisterMessage(InnerProductParameter)

LRNParameter = _reflection.GeneratedProtocolMessageType('LRNParameter', (_message.Message,), {
  'DESCRIPTOR' : _LRNPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.LRNParameter)
  })
_sym_db.RegisterMessage(LRNParameter)

MemoryDataParameter = _reflection.GeneratedProtocolMessageType('MemoryDataParameter', (_message.Message,), {
  'DESCRIPTOR' : _MEMORYDATAPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.MemoryDataParameter)
  })
_sym_db.RegisterMessage(MemoryDataParameter)

MVNParameter = _reflection.GeneratedProtocolMessageType('MVNParameter', (_message.Message,), {
  'DESCRIPTOR' : _MVNPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.MVNParameter)
  })
_sym_db.RegisterMessage(MVNParameter)

PermuteParameter = _reflection.GeneratedProtocolMessageType('PermuteParameter', (_message.Message,), {
  'DESCRIPTOR' : _PERMUTEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PermuteParameter)
  })
_sym_db.RegisterMessage(PermuteParameter)

PoolingParameter = _reflection.GeneratedProtocolMessageType('PoolingParameter', (_message.Message,), {
  'DESCRIPTOR' : _POOLINGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PoolingParameter)
  })
_sym_db.RegisterMessage(PoolingParameter)

PoolingSpecificParameter = _reflection.GeneratedProtocolMessageType('PoolingSpecificParameter', (_message.Message,), {
  'DESCRIPTOR' : _POOLINGSPECIFICPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PoolingSpecificParameter)
  })
_sym_db.RegisterMessage(PoolingSpecificParameter)

PowerParameter = _reflection.GeneratedProtocolMessageType('PowerParameter', (_message.Message,), {
  'DESCRIPTOR' : _POWERPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PowerParameter)
  })
_sym_db.RegisterMessage(PowerParameter)

PriorBoxParameter = _reflection.GeneratedProtocolMessageType('PriorBoxParameter', (_message.Message,), {
  'DESCRIPTOR' : _PRIORBOXPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PriorBoxParameter)
  })
_sym_db.RegisterMessage(PriorBoxParameter)

PythonParameter = _reflection.GeneratedProtocolMessageType('PythonParameter', (_message.Message,), {
  'DESCRIPTOR' : _PYTHONPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PythonParameter)
  })
_sym_db.RegisterMessage(PythonParameter)

ReLUParameter = _reflection.GeneratedProtocolMessageType('ReLUParameter', (_message.Message,), {
  'DESCRIPTOR' : _RELUPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ReLUParameter)
  })
_sym_db.RegisterMessage(ReLUParameter)

ROIPoolingParameter = _reflection.GeneratedProtocolMessageType('ROIPoolingParameter', (_message.Message,), {
  'DESCRIPTOR' : _ROIPOOLINGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ROIPoolingParameter)
  })
_sym_db.RegisterMessage(ROIPoolingParameter)

ReshapeParameter = _reflection.GeneratedProtocolMessageType('ReshapeParameter', (_message.Message,), {
  'DESCRIPTOR' : _RESHAPEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ReshapeParameter)
  })
_sym_db.RegisterMessage(ReshapeParameter)

SigmoidParameter = _reflection.GeneratedProtocolMessageType('SigmoidParameter', (_message.Message,), {
  'DESCRIPTOR' : _SIGMOIDPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.SigmoidParameter)
  })
_sym_db.RegisterMessage(SigmoidParameter)

SliceParameter = _reflection.GeneratedProtocolMessageType('SliceParameter', (_message.Message,), {
  'DESCRIPTOR' : _SLICEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.SliceParameter)
  })
_sym_db.RegisterMessage(SliceParameter)

SoftmaxParameter = _reflection.GeneratedProtocolMessageType('SoftmaxParameter', (_message.Message,), {
  'DESCRIPTOR' : _SOFTMAXPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.SoftmaxParameter)
  })
_sym_db.RegisterMessage(SoftmaxParameter)

TanHParameter = _reflection.GeneratedProtocolMessageType('TanHParameter', (_message.Message,), {
  'DESCRIPTOR' : _TANHPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.TanHParameter)
  })
_sym_db.RegisterMessage(TanHParameter)

ThresholdParameter = _reflection.GeneratedProtocolMessageType('ThresholdParameter', (_message.Message,), {
  'DESCRIPTOR' : _THRESHOLDPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ThresholdParameter)
  })
_sym_db.RegisterMessage(ThresholdParameter)

WindowDataParameter = _reflection.GeneratedProtocolMessageType('WindowDataParameter', (_message.Message,), {
  'DESCRIPTOR' : _WINDOWDATAPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.WindowDataParameter)
  })
_sym_db.RegisterMessage(WindowDataParameter)

BNParameter = _reflection.GeneratedProtocolMessageType('BNParameter', (_message.Message,), {
  'DESCRIPTOR' : _BNPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.BNParameter)
  })
_sym_db.RegisterMessage(BNParameter)

CTCParameter = _reflection.GeneratedProtocolMessageType('CTCParameter', (_message.Message,), {
  'DESCRIPTOR' : _CTCPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.CTCParameter)
  })
_sym_db.RegisterMessage(CTCParameter)

PReLUParameter = _reflection.GeneratedProtocolMessageType('PReLUParameter', (_message.Message,), {
  'DESCRIPTOR' : _PRELUPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PReLUParameter)
  })
_sym_db.RegisterMessage(PReLUParameter)

CropParameter = _reflection.GeneratedProtocolMessageType('CropParameter', (_message.Message,), {
  'DESCRIPTOR' : _CROPPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.CropParameter)
  })
_sym_db.RegisterMessage(CropParameter)

AffineTransParameter = _reflection.GeneratedProtocolMessageType('AffineTransParameter', (_message.Message,), {
  'DESCRIPTOR' : _AFFINETRANSPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.AffineTransParameter)
  })
_sym_db.RegisterMessage(AffineTransParameter)

AffineTransPointParameter = _reflection.GeneratedProtocolMessageType('AffineTransPointParameter', (_message.Message,), {
  'DESCRIPTOR' : _AFFINETRANSPOINTPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.AffineTransPointParameter)
  })
_sym_db.RegisterMessage(AffineTransPointParameter)

CalcAffineMatParameter = _reflection.GeneratedProtocolMessageType('CalcAffineMatParameter', (_message.Message,), {
  'DESCRIPTOR' : _CALCAFFINEMATPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.CalcAffineMatParameter)
  })
_sym_db.RegisterMessage(CalcAffineMatParameter)

ROIParameter = _reflection.GeneratedProtocolMessageType('ROIParameter', (_message.Message,), {
  'DESCRIPTOR' : _ROIPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ROIParameter)
  })
_sym_db.RegisterMessage(ROIParameter)

InterpParameter = _reflection.GeneratedProtocolMessageType('InterpParameter', (_message.Message,), {
  'DESCRIPTOR' : _INTERPPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.InterpParameter)
  })
_sym_db.RegisterMessage(InterpParameter)

RecurrentParameter = _reflection.GeneratedProtocolMessageType('RecurrentParameter', (_message.Message,), {
  'DESCRIPTOR' : _RECURRENTPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.RecurrentParameter)
  })
_sym_db.RegisterMessage(RecurrentParameter)

FlattenParameter = _reflection.GeneratedProtocolMessageType('FlattenParameter', (_message.Message,), {
  'DESCRIPTOR' : _FLATTENPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.FlattenParameter)
  })
_sym_db.RegisterMessage(FlattenParameter)

BatchNormParameter = _reflection.GeneratedProtocolMessageType('BatchNormParameter', (_message.Message,), {
  'DESCRIPTOR' : _BATCHNORMPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.BatchNormParameter)
  })
_sym_db.RegisterMessage(BatchNormParameter)

V1LayerParameter = _reflection.GeneratedProtocolMessageType('V1LayerParameter', (_message.Message,), {
  'DESCRIPTOR' : _V1LAYERPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.V1LayerParameter)
  })
_sym_db.RegisterMessage(V1LayerParameter)

V0LayerParameter = _reflection.GeneratedProtocolMessageType('V0LayerParameter', (_message.Message,), {
  'DESCRIPTOR' : _V0LAYERPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.V0LayerParameter)
  })
_sym_db.RegisterMessage(V0LayerParameter)

ChannelShuffleParameter = _reflection.GeneratedProtocolMessageType('ChannelShuffleParameter', (_message.Message,), {
  'DESCRIPTOR' : _CHANNELSHUFFLEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ChannelShuffleParameter)
  })
_sym_db.RegisterMessage(ChannelShuffleParameter)

MovingAvgParameter = _reflection.GeneratedProtocolMessageType('MovingAvgParameter', (_message.Message,), {
  'DESCRIPTOR' : _MOVINGAVGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.MovingAvgParameter)
  })
_sym_db.RegisterMessage(MovingAvgParameter)

SlicingParameter = _reflection.GeneratedProtocolMessageType('SlicingParameter', (_message.Message,), {
  'DESCRIPTOR' : _SLICINGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.SlicingParameter)
  })
_sym_db.RegisterMessage(SlicingParameter)

BilateralSlicingParameter = _reflection.GeneratedProtocolMessageType('BilateralSlicingParameter', (_message.Message,), {
  'DESCRIPTOR' : _BILATERALSLICINGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.BilateralSlicingParameter)
  })
_sym_db.RegisterMessage(BilateralSlicingParameter)

BilateralSliceApplyParameter = _reflection.GeneratedProtocolMessageType('BilateralSliceApplyParameter', (_message.Message,), {
  'DESCRIPTOR' : _BILATERALSLICEAPPLYPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.BilateralSliceApplyParameter)
  })
_sym_db.RegisterMessage(BilateralSliceApplyParameter)

AdaptiveBilateralSliceApplyParameter = _reflection.GeneratedProtocolMessageType('AdaptiveBilateralSliceApplyParameter', (_message.Message,), {
  'DESCRIPTOR' : _ADAPTIVEBILATERALSLICEAPPLYPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.AdaptiveBilateralSliceApplyParameter)
  })
_sym_db.RegisterMessage(AdaptiveBilateralSliceApplyParameter)

PointsCurveParameter = _reflection.GeneratedProtocolMessageType('PointsCurveParameter', (_message.Message,), {
  'DESCRIPTOR' : _POINTSCURVEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PointsCurveParameter)
  })
_sym_db.RegisterMessage(PointsCurveParameter)

SubpixelDownParameter = _reflection.GeneratedProtocolMessageType('SubpixelDownParameter', (_message.Message,), {
  'DESCRIPTOR' : _SUBPIXELDOWNPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.SubpixelDownParameter)
  })
_sym_db.RegisterMessage(SubpixelDownParameter)

SubpixelUpParameter = _reflection.GeneratedProtocolMessageType('SubpixelUpParameter', (_message.Message,), {
  'DESCRIPTOR' : _SUBPIXELUPPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.SubpixelUpParameter)
  })
_sym_db.RegisterMessage(SubpixelUpParameter)

ClipParameter = _reflection.GeneratedProtocolMessageType('ClipParameter', (_message.Message,), {
  'DESCRIPTOR' : _CLIPPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ClipParameter)
  })
_sym_db.RegisterMessage(ClipParameter)

EltwiseAffineTransformParameter = _reflection.GeneratedProtocolMessageType('EltwiseAffineTransformParameter', (_message.Message,), {
  'DESCRIPTOR' : _ELTWISEAFFINETRANSFORMPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.EltwiseAffineTransformParameter)
  })
_sym_db.RegisterMessage(EltwiseAffineTransformParameter)

TileParameter = _reflection.GeneratedProtocolMessageType('TileParameter', (_message.Message,), {
  'DESCRIPTOR' : _TILEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.TileParameter)
  })
_sym_db.RegisterMessage(TileParameter)

ReduceParameter = _reflection.GeneratedProtocolMessageType('ReduceParameter', (_message.Message,), {
  'DESCRIPTOR' : _REDUCEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ReduceParameter)
  })
_sym_db.RegisterMessage(ReduceParameter)

PadParameter = _reflection.GeneratedProtocolMessageType('PadParameter', (_message.Message,), {
  'DESCRIPTOR' : _PADPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PadParameter)
  })
_sym_db.RegisterMessage(PadParameter)

ReLU6Parameter = _reflection.GeneratedProtocolMessageType('ReLU6Parameter', (_message.Message,), {
  'DESCRIPTOR' : _RELU6PARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ReLU6Parameter)
  })
_sym_db.RegisterMessage(ReLU6Parameter)

Correlation2DParameter = _reflection.GeneratedProtocolMessageType('Correlation2DParameter', (_message.Message,), {
  'DESCRIPTOR' : _CORRELATION2DPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.Correlation2DParameter)
  })
_sym_db.RegisterMessage(Correlation2DParameter)

MatMulParameter = _reflection.GeneratedProtocolMessageType('MatMulParameter', (_message.Message,), {
  'DESCRIPTOR' : _MATMULPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.MatMulParameter)
  })
_sym_db.RegisterMessage(MatMulParameter)

ParameterParameter = _reflection.GeneratedProtocolMessageType('ParameterParameter', (_message.Message,), {
  'DESCRIPTOR' : _PARAMETERPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ParameterParameter)
  })
_sym_db.RegisterMessage(ParameterParameter)

PixelShuffleParameter = _reflection.GeneratedProtocolMessageType('PixelShuffleParameter', (_message.Message,), {
  'DESCRIPTOR' : _PIXELSHUFFLEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.PixelShuffleParameter)
  })
_sym_db.RegisterMessage(PixelShuffleParameter)

ROITransformParameter = _reflection.GeneratedProtocolMessageType('ROITransformParameter', (_message.Message,), {
  'DESCRIPTOR' : _ROITRANSFORMPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ROITransformParameter)
  })
_sym_db.RegisterMessage(ROITransformParameter)

MainTransformParameter = _reflection.GeneratedProtocolMessageType('MainTransformParameter', (_message.Message,), {
  'DESCRIPTOR' : _MAINTRANSFORMPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.MainTransformParameter)
  })
_sym_db.RegisterMessage(MainTransformParameter)

InstanceNormParameter = _reflection.GeneratedProtocolMessageType('InstanceNormParameter', (_message.Message,), {
  'DESCRIPTOR' : _INSTANCENORMPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.InstanceNormParameter)
  })
_sym_db.RegisterMessage(InstanceNormParameter)

GridSampleParameter = _reflection.GeneratedProtocolMessageType('GridSampleParameter', (_message.Message,), {
  'DESCRIPTOR' : _GRIDSAMPLEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.GridSampleParameter)
  })
_sym_db.RegisterMessage(GridSampleParameter)

GridSample3DParameter = _reflection.GeneratedProtocolMessageType('GridSample3DParameter', (_message.Message,), {
  'DESCRIPTOR' : _GRIDSAMPLE3DPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.GridSample3DParameter)
  })
_sym_db.RegisterMessage(GridSample3DParameter)

ReduceL2Parameter = _reflection.GeneratedProtocolMessageType('ReduceL2Parameter', (_message.Message,), {
  'DESCRIPTOR' : _REDUCEL2PARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ReduceL2Parameter)
  })
_sym_db.RegisterMessage(ReduceL2Parameter)

VarianceParameter = _reflection.GeneratedProtocolMessageType('VarianceParameter', (_message.Message,), {
  'DESCRIPTOR' : _VARIANCEPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.VarianceParameter)
  })
_sym_db.RegisterMessage(VarianceParameter)

MeanParameter = _reflection.GeneratedProtocolMessageType('MeanParameter', (_message.Message,), {
  'DESCRIPTOR' : _MEANPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.MeanParameter)
  })
_sym_db.RegisterMessage(MeanParameter)

Interp3dParameter = _reflection.GeneratedProtocolMessageType('Interp3dParameter', (_message.Message,), {
  'DESCRIPTOR' : _INTERP3DPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.Interp3dParameter)
  })
_sym_db.RegisterMessage(Interp3dParameter)

GRUParameter = _reflection.GeneratedProtocolMessageType('GRUParameter', (_message.Message,), {
  'DESCRIPTOR' : _GRUPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.GRUParameter)
  })
_sym_db.RegisterMessage(GRUParameter)

CorrelationMigParameter = _reflection.GeneratedProtocolMessageType('CorrelationMigParameter', (_message.Message,), {
  'DESCRIPTOR' : _CORRELATIONMIGPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.CorrelationMigParameter)
  })
_sym_db.RegisterMessage(CorrelationMigParameter)

ArgSortParameter = _reflection.GeneratedProtocolMessageType('ArgSortParameter', (_message.Message,), {
  'DESCRIPTOR' : _ARGSORTPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.ArgSortParameter)
  })
_sym_db.RegisterMessage(ArgSortParameter)

CumProdParameter = _reflection.GeneratedProtocolMessageType('CumProdParameter', (_message.Message,), {
  'DESCRIPTOR' : _CUMPRODPARAMETER,
  '__module__' : 'caffe_pb2'
  # @@protoc_insertion_point(class_scope:caffe.CumProdParameter)
  })
_sym_db.RegisterMessage(CumProdParameter)


_BLOBSHAPE.fields_by_name['dim']._options = None
_BLOBPROTO.fields_by_name['data']._options = None
_BLOBPROTO.fields_by_name['diff']._options = None
# @@protoc_insertion_point(module_scope)
